using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Text;
using System.Xml;

namespace HaloMap
{
	/// <summary>
	/// Summary description for Meta.
	/// </summary>
	/// 
	

	public class TagMeta
	{
		#region Fields
		public MemoryStream MS;
		public List<TagMeta.Item> items=new List<TagMeta.Item>();
		public int itemcount;
		public int reflexivecount;
		public int headersize;
		public int size;
		public string type;
		public int offset;
		public int ident;
		public int tagnumber;
		public int magic;
		public string name;
        public int mapnumber;
		public RawDataContainer raw;
		public RawDataContainerType rawType;
		public bool parsed=false;
		public char padding;
    public bool scannedwithent=false;
		#endregion
		#region Constructor
		public TagMeta()
		{
			
		}
		#endregion
		#region Methods
		public int[] FindItemsByDescription(string name)
		{
			int[] tempint=new int[this.items.Count];
			int count=0;
			for (int x=0;x<this.items.Count;x++)
			{
				Item i=(Item)this.items[x];
                string[] tempd=i.description.Split('[');
                string ix = tempd[0];
         

                if (ix.Trim() == name)
				{
             
					tempint[count]=x;
					count++;
				}
			}
			int[] returnint=new int[count];;
			Array.Copy(tempint,0,returnint,0,count);
			return returnint;
		}
        public int FindItemsByOffset(int offset)
        {
           
 
            for (int x = 0; x < this.items.Count; x++)
            {
                Item i = (Item)this.items[x];
                if (i.offset == offset)
                {
                    
                    return x;
                }
            }
        
            return -1;
        }
		public void SortItemsByOffset()
		{
            Comparison<TagMeta.Item> compareBySalary;
            compareBySalary = new Comparison<TagMeta.Item>(new ItemSorter().Compare);

            this.items.Sort(compareBySalary);
		}
		public void ReadMetaFromMap(int tagnum,int mapnumberx,bool dontreadraw)
		{
			//set meta properties
            this.mapnumber = mapnumberx;
			this.tagnumber=tagnum;
            this.type = Maps.map[mapnumber].MetaInfo.TagType[tagnum];
            this.name = Maps.map[mapnumber].FileNames.Name[tagnum];
            this.offset = Maps.map[mapnumber].MetaInfo.Offset[tagnum];
            this.size = Maps.map[mapnumber].MetaInfo.Size[tagnum];
            this.ident = Maps.map[mapnumber].MetaInfo.Ident[tagnum];

			string temps=this.offset.ToString("X");
			char[] tempc=temps.ToCharArray();
			int	xxx=tempc.Length;
			this.padding=tempc[xxx-1];


			//read meta to memorystream

         
			this.MS=new MemoryStream(this.size);
            Maps.map[mapnumber].BR.BaseStream.Position = this.offset;
            this.MS.Write(Maps.map[mapnumber].BR.ReadBytes(this.size), 0, this.size);
     
			//checks if type has raw data
            this.rawType = Maps.map[mapnumber].Functions.Meta.CheckForRaw(this.type);
            if (dontreadraw==false)
            {
		    if (rawType != RawDataContainerType.Empty)
            {
                this.raw = Maps.map[mapnumber].Functions.Meta.ReadRaw(this.tagnumber, mapnumber, dontreadraw);
            }

	
			}
            if (this.type == "sbsp")
			{
                int h = Maps.map[mapnumber].BSP.FindBSPNumberByBSPIdent(this.ident);
                this.magic = Maps.map[mapnumber].BSP.sbsp[h].magic;

           
			}
            else if (this.type == "ltmp")
            {
                int h = Maps.map[mapnumber].BSP.FindBSPNumberByLightMapIdent(this.ident);
                this.magic = Maps.map[mapnumber].BSP.sbsp[h].magic;
            }
            else
            {
                if (Maps.map[mapnumber].HaloVersion == Map.HaloVersionEnum.HaloCE || Maps.map[mapnumber].HaloVersion == Map.HaloVersionEnum.Halo1)
                {
                    this.magic = Maps.map[mapnumber].PrimaryMagic;
                }
                else
                {
                    this.magic = Maps.map[mapnumber].SecondaryMagic;
                }
            }
			
	}

        public static void ScanMetaItems(ref TagMeta meta,bool manual, bool parsed, int mapnumber)
        {
            if (parsed == true) { meta.parsed = true; }
            meta.items.Clear();
            MetaScanner metaScanner = new MetaScanner();
            switch (Maps.map[mapnumber].HaloVersion)
            {
                case Map.HaloVersionEnum.Halo1:
                case Map.HaloVersionEnum.HaloCE:
                    meta.magic = Maps.map[mapnumber].PrimaryMagic;

                    if (Maps.map[mapnumber].MetaInfo.TagType[meta.tagnumber] == "bitm" && Maps.map[mapnumber].MetaInfo.external[meta.tagnumber] == true && Maps.map[mapnumber].HaloVersion != Map.HaloVersionEnum.Halo1) { Maps.map[mapnumber].OpenMap(MapTypes.Bitmaps); }
                    else { Maps.map[mapnumber].OpenMap(MapTypes.Internal); }
                    meta.ReadMetaFromMap(meta.tagnumber, mapnumber, false);


                    if (manual == true && meta.type != "sbsp")
                    {
                        meta.scannedwithent = true;

                        IFP.IFPHashMap.RemoveIfp(meta.type,mapnumber );

                        IFPIO io = IFP.IFPHashMap.GetIfp(meta.type, mapnumber);

                        meta.headersize = io.headerSize;

                        metaScanner.ScanWithIFP(ref io, ref meta, mapnumber);

                    }
                    else
                    {
                        metaScanner.ScanManually(ref meta, mapnumber);
                    }
                  
                    break;
                case Map.HaloVersionEnum.Halo2:
                    Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                    meta.ReadMetaFromMap(meta.tagnumber, mapnumber, false);

     
                        if (manual  == true&&meta.type!="sbsp")
                        {

                            IFP.IFPHashMap.RemoveIfp(meta.type, mapnumber);
                      
                            IFPIO io = IFP.IFPHashMap.GetIfp(meta.type,mapnumber);

                            meta.headersize = io.headerSize;

                            metaScanner.ScanWithIFP(ref io, ref meta, mapnumber);

                        }
                        else
                        {
                            metaScanner.ScanManually(ref meta, mapnumber);
                        }
                    
                 
                    break;
            }

        }


        public ArrayList RecursivelyLoadMetas(int mapnumber, bool parsed, ToolStripProgressBar pb)
		{
			ArrayList metas=new ArrayList(0);
			metas.Add(this);
            SaveRecursiveFunction(this, ref metas, mapnumber, parsed, pb, 0, 100);
			return metas;
		}

		public void SaveRecursive(string outputFilePath,int mapnumber,bool parsed,ToolStripProgressBar pb )
		{
            ArrayList metas = RecursivelyLoadMetas(mapnumber, parsed, pb);
			
			int y=this.name.LastIndexOf("\\")+1;
			FileStream FS=new FileStream(outputFilePath+"\\"+this.name.Substring(y,this.name.Length-y)+".info",FileMode.Create);
			StreamWriter SW=new StreamWriter(FS);
			for (int x=0;x<metas.Count;x++)
			{
				TagMeta m=(TagMeta)metas[x];
				
				string temp1=m.name;
				//directory to create
				int loc=temp1.LastIndexOf("\\")+1;
				string temp2=outputFilePath+"\\"+temp1.Substring(0,loc);
				Directory.CreateDirectory(temp2);
				//name of meta
				string tempname=temp1.Substring(loc,temp1.Length-loc);
				string temp3=temp2+tempname+"."+m.type;
				temp3=temp3.Replace("<","_");
				temp3=temp3.Replace(">","_");
				m.SortItemsByOffset();
				m.SaveMetaToFile(temp3);
				SW.WriteLine(temp3);
			}
				
			SW.Close();
			FS.Close();
		}
		private void SaveRecursiveFunction(TagMeta meta,ref ArrayList metas,int mapnumber,bool parsed,ToolStripProgressBar pb,float percent,float percentsize)
		{
            
			int wtf=metas.Count;
            if (meta.items.Count == 0) { return; }
            float test= meta.items.Count;
            
            float currentpercentsize = percentsize / test;
           
			for (int x=0;x<meta.items.Count;x++)
			{
				Item i=(Item) meta.items[x];
				if (i.type!=ItemType.Ident){continue;}

                float currentpercentage = percent + (currentpercentsize * x);
                pb.Value = (int)currentpercentage;
                Application.DoEvents();
				Ident id=(Ident) i;
				
				
				if (id.ident==-1){continue;}
				bool exists=false;
				
			
				
				for (int e=0;e<metas.Count;e++)
				{
					TagMeta tempmeta=(TagMeta)metas[e];
					//if (id.pointstotagtype  ==tempmeta.type&&id.pointstotagname ==tempmeta.name )
					if (id.ident==tempmeta.ident)
					{
		
						exists=true;
						break;
					}
					
				}
				if (exists==true){continue;}

                int num = Maps.map[mapnumber].Functions.Meta.FindMetaByID(id.ident, mapnumber);
				TagMeta m=new TagMeta();
				if (parsed==true){m.parsed=true;}

                m.ReadMetaFromMap(num, mapnumber, false);
                if (m.type == "ltmp" | m.type == "matg") { continue; }
				if (m.type=="phmo"|m.type=="coll"|m.type=="jmad")
				{
					m.parsed=false;
				}
				MetaScanner metaScanner=new MetaScanner();
                if (m.type != "jmad")
                {

                   IFPIO ifp = IFP.IFPHashMap.GetIfp(m.type, mapnumber);
                   m.headersize = ifp.headerSize;

                    metaScanner.ScanWithIFP(ref ifp, ref m, mapnumber);
                }
                else
                {
                    metaScanner.ScanManually(ref m, mapnumber);
                }
             
					m.SortItemsByOffset();

					metas.Add(m);
			
				Application.DoEvents();
				SaveRecursiveFunction(m,ref metas,mapnumber,parsed, pb,currentpercentage,currentpercentsize);
				
			}
					
		}
		
		public void SaveMetaToFile(string outputFileName)
		{
			
			outputFileName=outputFileName.Replace("<","_");
			outputFileName=outputFileName.Replace(">","_");
		
			//write memorysteam of meta to file
			FileStream FS=new FileStream(outputFileName,FileMode.Create);
			BinaryWriter BW=new BinaryWriter(FS);
			BW.BaseStream.Position=0;
			BW.BaseStream.Write(this.MS.ToArray(),0,this.size);
			BW.Close();
			FS.Close();
			
			XmlTextWriter xtw=new XmlTextWriter(outputFileName+".xml",Encoding.Default);
			xtw.Formatting=Formatting.Indented;
			xtw.WriteStartElement("Meta");
			xtw.WriteAttributeString("TagType",this.type);
			xtw.WriteAttributeString("TagName",this.name);
			xtw.WriteAttributeString("Offset",this.offset.ToString());
			xtw.WriteAttributeString("Size",this.size.ToString());
			xtw.WriteAttributeString("Magic",this.magic.ToString());
			xtw.WriteAttributeString("Parsed",this.parsed.ToString());
			xtw.WriteAttributeString("Date",DateTime.Today.ToShortDateString());
			xtw.WriteAttributeString("Time",DateTime.Now.ToShortTimeString());
			xtw.WriteAttributeString("EntityVersion","0.1");
			xtw.WriteAttributeString("Padding",this.padding.ToString());
			//xtw.
			for (int x=0;x<this.items.Count;x++)
			{
				Item i=(Item) this.items[x];
				switch (i.type)
				{
				case ItemType.Reflexive:
				Reflexive r=(Reflexive)i;
				xtw.WriteStartElement("Reflexive");
				xtw.WriteAttributeString("Description",r.description);
				xtw.WriteAttributeString("Offset",r.offset.ToString());
				xtw.WriteAttributeString("ChunkCount",r.chunkcount.ToString());
				xtw.WriteAttributeString("ChunkSize",r.chunksize.ToString());
				xtw.WriteAttributeString("Translation",r.translation.ToString());
				xtw.WriteAttributeString("PointsToTagType",r.pointstotagtype);
				xtw.WriteAttributeString("PointsToTagName",r.pointstotagname);
				xtw.WriteAttributeString("TagType",r.intagtype);
				xtw.WriteAttributeString("TagName", r.intagname);
				xtw.WriteEndElement();
				break;

				case ItemType.Ident:
				Ident id=(Ident)i;
				xtw.WriteStartElement("Ident");
				xtw.WriteAttributeString("Description",id.description);
				xtw.WriteAttributeString("Offset",id.offset.ToString());
				xtw.WriteAttributeString("PointsToTagType",id.pointstotagtype);
				xtw.WriteAttributeString("PointsToTagName",id.pointstotagname);
				xtw.WriteAttributeString("TagType",id.intagtype);
				xtw.WriteAttributeString("TagName", id.intagname);
				xtw.WriteEndElement();
				break;
				case ItemType.String :
				String s=(String)i;
				xtw.WriteStartElement("String");
				xtw.WriteAttributeString("Description",s.description);
				xtw.WriteAttributeString("Offset",s.offset.ToString());
				xtw.WriteAttributeString("StringName",s.name);
				xtw.WriteAttributeString("TagType",s.intagtype);
				xtw.WriteAttributeString("TagName", s.intagname);
				xtw.WriteEndElement();
				break;
				}
			}
			xtw.WriteEndElement();
			xtw.Close();
			if (this.rawType!=RawDataContainerType.Empty)
			{
				this.raw.SaveRawToFile(outputFileName,this);
			}
		}
		
		public void RelinkReferences(int mapnumber)
		{
			for (int x=0;x<this.items.Count;x++)
			{
				Item i=(Item)items[x];
				switch (i.type)
				{
					case ItemType.Ident:
						Ident id=(Ident) i;
                        id.pointstotagnumber = Maps.map[mapnumber].Functions.Meta.FindByNameAndTagType(id.pointstotagtype, id.pointstotagname, mapnumber);
                        if (id.pointstotagnumber == -1)
                        {
                            id.ident = -1;
                        }
                        else
                        {

                            id.ident = Maps.map[mapnumber].MetaInfo.Ident[id.pointstotagnumber];
                        }
						items[x]=id;
						break;
                        	case ItemType.String:
						String s=(String) i;
                       if (Array.IndexOf(Maps.map[mapnumber].Strings.Name,s.name)==-1)
                       {
                           s.name="";
                           s.id=0;
                       }
						items[x]=s;
						break;
				}
			}

		}


        public void WriteReferences()
        {
                BinaryWriter BW=new BinaryWriter(this.MS);
            
                for (int xx = 0; xx < this.items.Count; xx++)
                {

                    TagMeta.Item i = (TagMeta.Item)this.items[xx];

                    if (i.intag != this.tagnumber) { continue; }
                    switch (i.type)
                    {
                        case TagMeta.ItemType.Reflexive:
                            TagMeta.Reflexive reflex = (TagMeta.Reflexive)i;
                            int newreflex = this.offset + reflex.translation + this.magic ;


                            if (reflex.pointstotagnumber != this.tagnumber)
                            {
                                newreflex = Maps.map[mapnumber].MetaInfo.Offset[reflex.pointstotagnumber] + reflex.translation + this.magic;

                            }

                            BW.BaseStream.Position = reflex.offset;
                            BW.Write(reflex.chunkcount);
                            BW.Write(newreflex);
                            break;

                        case TagMeta.ItemType.Ident:
                            TagMeta.Ident id = (TagMeta.Ident)i;

                            BW.BaseStream.Position = id.offset;
                            BW.Write(id.ident);

                            break;

                        case TagMeta.ItemType.String:
                            TagMeta.String ss = (TagMeta.String)i;
                            byte bleh =0;
                            BW.BaseStream.Position = ss.offset;
                            BW.Write((short)ss.id);
                            BW.Write(bleh);
                            BW.Write((byte)ss.name.Length);
                            break;

                    }

                }
        }





		public void LoadMetaFromFile(string inputFileName,int mapnumber)
		{
			//write memorysteam of meta to file
			FileStream FS=new FileStream(inputFileName,FileMode.Open);
			BinaryReader BR=new BinaryReader(FS);
			this.size=(int)FS.Length;
			this.MS=new MemoryStream(this.size);
			BR.BaseStream.Position=0;
			this.MS.Write(BR.ReadBytes(this.size),0,this.size);
			BR.Close();
			FS.Close();
			


			//write idents,strings,reflexives
			XmlTextReader xtr=new XmlTextReader(inputFileName+".xml");
			xtr.WhitespaceHandling=WhitespaceHandling.None;
			
			while (xtr.Read())
			{
				//MessageBox.Show(xtr.Name);
				switch (xtr.NodeType)
				{
					case XmlNodeType.Element:
						if (xtr.Name=="Meta")
						{
							this.type=xtr.GetAttribute("TagType");
							this.name=xtr.GetAttribute("TagName");
							this.parsed=xtr.GetAttribute("Parsed")=="True"? true: false;
							this.size=Convert.ToInt32(xtr.GetAttribute("Size"));	
							this.magic=Convert.ToInt32(xtr.GetAttribute("Magic"));
							this.padding=Convert.ToChar(xtr.GetAttribute("Padding"));
							this.offset=Convert.ToInt32(xtr.GetAttribute("Offset"));	
						}
						else if (xtr.Name=="Reflexive")
						{
							Reflexive r=new Reflexive();
							r.description=xtr.GetAttribute("Description");
							r.offset=Convert.ToInt32(xtr.GetAttribute("Offset"));
							r.chunkcount=Convert.ToInt32(xtr.GetAttribute("ChunkCount"));
							r.chunksize=Convert.ToInt32(xtr.GetAttribute("ChunkSize"));
							r.translation=Convert.ToInt32(xtr.GetAttribute("Translation"));
							r.pointstotagtype=xtr.GetAttribute("PointsToTagType");
							r.pointstotagname=xtr.GetAttribute("PointsToTagName");
                            r.pointstotagnumber = Maps.map[mapnumber].Functions.Meta.FindByNameAndTagType(r.pointstotagtype, r.pointstotagname, mapnumber);
							r.intagtype=xtr.GetAttribute("TagType");
							r.intagname=xtr.GetAttribute("TagName");
                            r.intag = Maps.map[mapnumber].Functions.Meta.FindByNameAndTagType(r.intagtype, r.intagname, mapnumber);
							this.items.Add(r);
							
						}
						else if (xtr.Name=="Ident")
						{
							Ident id=new Ident();
							
							id.description=xtr.GetAttribute("Description");
							id.offset=Convert.ToInt32(xtr.GetAttribute("Offset"));
							id.pointstotagtype=xtr.GetAttribute("PointsToTagType");
							id.pointstotagname=xtr.GetAttribute("PointsToTagName");
							id.pointstotagnumber=Maps.map[mapnumber].Functions.Meta.FindByNameAndTagType(id.pointstotagtype,id.pointstotagname,mapnumber);
							id.intagtype=xtr.GetAttribute("TagType");
							id.intagname=xtr.GetAttribute("TagName");
                            id.intag = Maps.map[mapnumber].Functions.Meta.FindByNameAndTagType(id.intagtype, id.intagname, mapnumber);
							this.items.Add(id);
							
						}
						else if (xtr.Name=="String")
						{
							String s=new String();
							s.description=xtr.GetAttribute("Description");
							s.offset=Convert.ToInt32(xtr.GetAttribute("Offset"));
							s.name=xtr.GetAttribute("StringName");
							s.intagtype=xtr.GetAttribute("TagType");
							s.intagname=xtr.GetAttribute("TagName");
                            s.intag = Maps.map[mapnumber].Functions.Meta.FindByNameAndTagType(s.intagtype, s.intagname, mapnumber);
							this.items.Add(s);
							
						}
						break;	
					default:
						break;
				}
				
			}
			xtr.Close();
			///
			///check for raw
			//
			this.rawType=Maps.map[mapnumber].Functions.Meta.CheckForRaw(this.type);
			if (this.rawType!=RawDataContainerType.Empty)
			{
				this.raw=new RawDataContainer();
				this.raw=this.raw.LoadRawFromFile(inputFileName,this);
			}
			
		}




		#endregion
		#region Enumerator
		public enum ItemType
		{
			Reflexive,
			Ident,
			String
		}
		#endregion
	

		
			public class Item
			{
				#region Fields
				public int offset;
				public ItemType type;
                public int mapOffset;
				public string description;
                public int intag;
                public string intagname;
                public string intagtype;
                public int parent = -1;
                public int sibling = -1;
                public int child=-1;
				#endregion
				#region Constructor
				public Item()
				{
				}
				#endregion
				#region Methods
				#endregion
			}

	

			public class Reflexive : Item
			{
				#region Fields
				
				public int chunkcount;
				public int chunksize;
				public int translation;
				public string pointstotagname;
				public string pointstotagtype;
				public int pointstotagnumber;
			
			
				#endregion
				#region Constructor
				public Reflexive()
				{
					type=ItemType.Reflexive;
				}
				#endregion
				#region Methods
				#endregion
			}

			public class String : Item
			{
				#region Fields
	
				public int id;
				public string name;
			
				#endregion
				#region Constructor
				public String()
				{
					type=ItemType.String;
				}
				#endregion
				#region Methods
				#endregion
			}

			public class Ident : Item
			{
				#region Fields
				public int ident;
				public string pointstotagname;
				public string pointstotagtype;
				public int pointstotagnumber;
			
			
				#endregion
				#region Constructor
				public Ident()
				{
					type=ItemType.Ident;
				}
				#endregion
				#region Methods
				#endregion
			}

			public class ItemSorter:IComparer
			{
				public int Compare(object x,object y)
				{
					return ((Item)x).offset.CompareTo(((Item)y).offset);
				}
			}

		}
}
