﻿using HaloMap;
using System.Collections.Generic;
using System.IO;

namespace ScriptTools
{
    public class TagMetaSplitter
    {
        public SplitReflexive Header;
        public string name;
        public string type;
        public int tagnumber;
        public int offset;
        public RawDataContainer raw;
        public RawDataContainerType rawtype;
        public int magic;

        public TagMetaSplitter()
        {

        }

        private static string GetBaseString(string name)
        {
            string[] split = name.Split('[');
            string tempstring = split[0].Trim();
            return tempstring;
        }

        public void SplitWithIFP(ref IFPIO ifp, ref TagMeta meta, int mapnumber)
        {
            type = meta.type;
            tagnumber = meta.tagnumber;
            name = meta.name;
            offset = meta.offset;
            magic = meta.magic;
            raw = meta.raw;
            rawtype = meta.rawType;
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            if (ifp.items != null)
            {
                Maps.map[mapnumber].BR.BaseStream.Position = meta.offset;
                Header = new SplitReflexive();
                Header.offset = 0;
                Header.Chunks = new List<SplitReflexive>();
                Header.translation = 0;
                //   Header.MS = new MemoryStream(ifp.headerSize);
                //  Header.MS.Write(Maps.map[mapnumber].BR.ReadBytes(ifp.headerSize), 0, ifp.headerSize);
                Header.chunksize = ifp.headerSize;
                Header.chunkcount = 1;
                Header.splitReflexiveType = SplitReflexive.SplitReflexiveType.Container;
                Header.realtranslation = meta.offset;
                if (meta.type == "sbsp")
                {
                    int p = Maps.map[mapnumber].BSP.FindBSPNumberByBSPIdent(meta.ident);
                    CycleElements(ref Header, ifp.items, ref meta, meta.offset, mapnumber, meta.tagnumber, Maps.map[mapnumber].BSP.sbsp[p].magic);
                }
                else if (meta.type == "ltmp")
                {
                    int p = Maps.map[mapnumber].BSP.FindBSPNumberByLightMapIdent(meta.ident);
                    CycleElements(ref Header, ifp.items, ref meta, meta.offset, mapnumber, meta.tagnumber, Maps.map[mapnumber].BSP.sbsp[p].magic);
                }
                else
                {
                    CycleElements(ref Header, ifp.items, ref meta, meta.offset, mapnumber, meta.tagnumber, Maps.map[mapnumber].SecondaryMagic);
                }

            }
            Maps.map[mapnumber].CloseMap();

        }

        void CycleElements(ref SplitReflexive reflex, object[] elements, ref TagMeta meta, int offset, int mapnumber, int tagnumber, int magic)
        {
            for (int x = 0; x < reflex.chunkcount; x++)
            {
                SplitReflexive chunkreflexive = new SplitReflexive();
                Maps.map[mapnumber].BR.BaseStream.Position = reflex.realtranslation + (x * reflex.chunksize);
                chunkreflexive.MS = new MemoryStream(reflex.chunksize);
                chunkreflexive.MS.Write(Maps.map[mapnumber].BR.ReadBytes(reflex.chunksize), 0, reflex.chunksize);
                chunkreflexive.chunksize = reflex.chunksize;
                chunkreflexive.splitReflexiveType = SplitReflexive.SplitReflexiveType.Chunk;

                reflex.Chunks.Add(chunkreflexive);

                for (int xx = 0; xx < elements.Length; xx++)
                {
                    IFPIO.BaseObject tempbase = (IFPIO.BaseObject)elements[xx];
                    
                    switch (tempbase.ObjectType)
                    {
                        case IFPIO.ObjectEnum.Reflexive:
                            IFPIO.Reflexive tempreflex = (IFPIO.Reflexive)tempbase;

                            SplitReflexive r = new SplitReflexive();
                            Maps.map[mapnumber].BR.BaseStream.Position = offset + tempreflex.offset + (x * reflex.chunksize);
                            r.mapOffset = (int)Maps.map[mapnumber].BR.BaseStream.Position;
                            r.chunkcount = Maps.map[mapnumber].BR.ReadInt32();
                            if (r.chunkcount == 0) { continue; }
                            r.chunksize = tempreflex.chunkSize;
                            r.translation = Maps.map[mapnumber].BR.ReadInt32() - magic;
                            r.pointstotagnumber = Maps.map[mapnumber].Functions.Meta.FindMetaByOffset(r.translation, mapnumber);
                            r.description = tempreflex.name;
                            if (r.pointstotagnumber == -1) { continue; }
                            //  r.parent = reflex.description;// parentname;
                            //
                            
                            r.realtranslation = r.translation;
                            r.realtagnumber = r.pointstotagnumber;
                            
                            r.label = tempreflex.label;
                            
                            r.pointstotagnumber = meta.tagnumber;
                            r.pointstotagtype = meta.type;
                            r.pointstotagname = meta.name;
                            r.offset = (int)tempreflex.offset;
                            r.intag = meta.tagnumber;
                            r.intagtype = meta.type;
                            r.intagname = meta.name;
                            r.translation -= Maps.map[mapnumber].MetaInfo.Offset[r.realtagnumber];

                            r.inchunknumber = x;
                            r.splitReflexiveType = SplitReflexive.SplitReflexiveType.Container;
                            r.Chunks = new List<SplitReflexive>();
                            CycleElements(ref r, tempreflex.items, ref meta, r.realtranslation, mapnumber, r.pointstotagnumber, magic);
                            reflex.Chunks[x].ChunkResources.Add(r);
                            meta.reflexivecount++;
                            break;

                        case IFPIO.ObjectEnum.Ident:
                            IFPIO.Ident tempident = (IFPIO.Ident)tempbase;
                            SplitIdent i = new SplitIdent();
                            Maps.map[mapnumber].BR.BaseStream.Position = offset + tempident.offset + (x * reflex.chunksize);
                            i.mapOffset = (int)Maps.map[mapnumber].BR.BaseStream.Position;
                            i.ident = Maps.map[mapnumber].BR.ReadInt32();
                            if (i.ident != -1)
                            {
                                i.pointstotagnumber = Maps.map[mapnumber].Functions.Meta.FindMetaByID(i.ident, mapnumber);
                                i.pointstotagtype = Maps.map[mapnumber].MetaInfo.TagType[i.pointstotagnumber];
                                i.pointstotagname = Maps.map[mapnumber].FileNames.Name[i.pointstotagnumber];
                            }
                            else
                            {
                                continue;
                            }

                            i.mapOffset = offset + tempident.offset + (x * reflex.chunksize);

                            i.offset = tempident.offset;
                            i.inchunknumber = x;
                            i.intag = tagnumber;
                            i.intagtype = Maps.map[mapnumber].MetaInfo.TagType[i.intag];
                            i.intagname = Maps.map[mapnumber].FileNames.Name[i.intag];
                            i.description = tempident.name;

                            reflex.Chunks[x].ChunkResources.Add(i);
                            break;

                        case IFPIO.ObjectEnum.StringID:
                            IFPIO.SID tempstringid = (IFPIO.SID)tempbase;
                            SplitString si = new SplitString();
                            Maps.map[mapnumber].BR.BaseStream.Position = offset + tempstringid.offset + (x * reflex.chunksize);
                            si.mapOffset = (int)Maps.map[mapnumber].BR.BaseStream.Position;
                            si.id = Maps.map[mapnumber].BR.ReadUInt16();
                            if (si.id == 0 | si.id >= Maps.map[mapnumber].MapHeader.scriptReferenceCount) { continue; }
                            Maps.map[mapnumber].BR.ReadByte();
                            int temp = Maps.map[mapnumber].BR.ReadByte();
                            if (temp != Maps.map[mapnumber].Strings.Length[si.id]) { continue; }
                            si.mapOffset = offset + tempstringid.offset + (x * reflex.chunksize);
                            si.offset = si.mapOffset - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];

                            si.offset = tempstringid.offset;
                            si.inchunknumber = x;
                            si.name = Maps.map[mapnumber].Strings.Name[si.id];
                            si.intag = tagnumber;
                            si.intagtype = Maps.map[mapnumber].MetaInfo.TagType[si.intag];
                            si.intagname = Maps.map[mapnumber].FileNames.Name[si.intag];
                            si.description = tempstringid.name;
                            reflex.Chunks[x].ChunkResources.Add(si);
                            break;
                    }
                }
            }

            return;
        }

        public class SplitReflexive : TagMeta.Reflexive
        {
            public int inchunknumber;
            public List<TagMeta.Item> ChunkResources = new List<TagMeta.Item>();
            public List<SplitReflexive> Chunks = new List<SplitReflexive>();
            public int realtranslation;
            public int realtagnumber;
            public MemoryStream MS;
            public SplitReflexiveType splitReflexiveType;
            public string label;

            public enum SplitReflexiveType
            {
                Chunk,
                Container
            }
        }

        public class SplitString : TagMeta.String
        {
            public int inchunknumber;
        }

        public class SplitIdent : TagMeta.Ident
        {
            public int inchunknumber;
        }
    }
}