using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;

namespace HaloMap
{
    //modified 6-28-06 by TheTyckoMan
    //modified 7-23-06 by TheTyckoMan
    //modified 8-10-06 by TheTyckoMan
    public class IFPIO
    {        
        #region fields
        public List<BaseObject> ENTElements;
        public string classType;
        public int headerSize;
        public object[] items;
        public string[] revisions;
        private ArrayList alRevisions = new ArrayList(0);
        private int size;
        public string author;
        public string version;
        private string FilePath;
        #endregion
        #region methods/functions
        #region Old (recursive)
        #region Reader
        public void ReadIFP(string inputFilePath)
        {
            FileStream fs = new FileStream(inputFilePath, FileMode.Open, FileAccess.ReadWrite);
            XmlTextReader xtr = new XmlTextReader(fs);
            xtr.WhitespaceHandling = WhitespaceHandling.None;
            items = IFPReader(ref xtr);
            alRevisions.TrimToSize();
            revisions = ((string[])alRevisions.ToArray(typeof(string)));
            alRevisions.Clear();
            headerSize = size;
            xtr.Close();
            fs.Close();
            this.newReadIFP(inputFilePath);
        }
        //function calls its self recursively
        private object[] IFPReader(ref XmlTextReader xtr)
        {
            ArrayList IFPElements = new ArrayList(0);
            int offset = 0;
            bool visible;
            int dependencyOffset = -1;
            bool endElement = false;
            while (xtr.Read())
            {
                visible = xtr.GetAttribute("visible") != null ? xtr.GetAttribute("visible").ToLower() == "false" ? false : true : true;
                switch (xtr.NodeType.ToString())
                {
                    case "Element":
                        {
                            switch (xtr.Name.ToLower())
                            {
                                //<plugin class="bipd" author="Iron_Forge" version="0.3">
                                case "plugin":
                                    {
                                        author = xtr.GetAttribute("author");
                                        version = xtr.GetAttribute("version");
                                        classType = xtr.GetAttribute("class");
                                        break;
                                    }
                                //<revision author="Iron_Forge" version="0.3">Added some known values...</revision>
                                case "revision":
                                    {
                                        alRevisions.Add(xtr.GetAttribute("author"));
                                        alRevisions.Add(xtr.GetAttribute("version"));
                                        alRevisions.Add(xtr.ReadString());
                                        break;
                                    }
                                // <struct name="Predicted Resources24" offset="32" visible="true" size="8">
                                case "struct":
                                    {
                                        bool temp=true;
                                        string i=xtr.GetAttribute("HasCount");
                                        if (i!=null)
                                        {
                                            temp=bool.Parse(i);
                                        }
                                        IFPElements.Add(new Reflexive(xtr.LineNumber, offset, visible, xtr.GetAttribute("name"), xtr.GetAttribute("label"), IFPReader(ref xtr), size, temp, -1, -1));
                                        if (temp == false)
                                        {
                                            offset += 4;
                                        }
                                        else
                                        {
                                            offset += 8;
                                        }
                                        break;
                                    }
                                // <tag name="" visible="false"/>
                                case "tag":
                                    {

                                        IFPElements.Add(new TagType(offset, visible, xtr.GetAttribute("name"), xtr.LineNumber, -1, -1));
                                        offset += 4;
                                        dependencyOffset = offset;
                                        break;
                                    }
                                // <id name="" visible="false"/>
                                case "id":
                                    {
                                        bool isDependency = dependencyOffset == offset ? true : false;
                                        IFPElements.Add(new Ident(xtr.GetAttribute("name"), visible, offset, isDependency, xtr.LineNumber, -1, -1));
                                        offset += 4;
                                        break;
                                    }
                                //  <stringid name="" visible="false" />
                                case "stringid":
                                    {
                                        IFPElements.Add(new SID(xtr.GetAttribute("name"), visible, offset, xtr.LineNumber, -1, -1));
                                        offset += 4;
                                        break;
                                    }
                                case "script":
                                    {
                                        goto case "stringid";
                                    }
                                //<unused size="32" default="0" />
                                case "unused":
                                    {
                                        int sizeOfUnusedSpace = Convert.ToInt32(xtr.GetAttribute("size"));
                                        IFPElements.Add(new Unused(offset, sizeOfUnusedSpace, xtr.LineNumber, -1, -1));
                                        offset += sizeOfUnusedSpace;
                                        break;
                                    }
                                //<undefined name="Unknown" visible="false" />
                                case "undefined":
                                    {
                                        IFPElements.Add(new Unknown(offset, visible, xtr.GetAttribute("name"), xtr.LineNumber, -1, -1));
                                        offset += 4;
                                        break;
                                    }
                                //<float name="" visible="false" />
                                case "float":
                                    {
                                        IFPElements.Add(new IFPFloat(offset, visible, xtr.GetAttribute("name"), xtr.LineNumber, -1, -1));
                                        offset += 4;
                                        break;
                                    }
                                //<int name="Unknown" visible="false" />
                                case "int":
                                    {
                                        IFPElements.Add(new IFPInt(offset, ObjectEnum.Int, visible, xtr.GetAttribute("name"), makeIndex(ref xtr, "int"), xtr.LineNumber, -1, -1));
                                        offset += 4;
                                        break;
                                    }
                                //<uint name="Structure BSP Offset" />
                                case "uint":
                                    {
                                        IFPElements.Add(new IFPInt(offset, ObjectEnum.UInt, visible, xtr.GetAttribute("name"), makeIndex(ref xtr, "uint"), xtr.LineNumber, -1, -1));
                                        offset += 4;
                                        break;
                                    }
                                //<short name="Palette Chunk #" visible="true" />
                                case "short":
                                    {
                                        IFPElements.Add(new IFPInt(offset, ObjectEnum.Short, visible, xtr.GetAttribute("name"), makeIndex(ref xtr, "short"), xtr.LineNumber, -1, -1));
                                        offset += 2;
                                        break;
                                    }
                                //<ushort name="Palette Chunk #" visible="true" />
                                case "ushort":
                                    {
                                        IFPElements.Add(new IFPInt(offset, ObjectEnum.UShort, visible, xtr.GetAttribute("name"), makeIndex(ref xtr, "ushort"), xtr.LineNumber, -1, -1));
                                        offset += 2;
                                        break;
                                    }
                                //<byte name="Unknown" visible="false" />
                                case "byte":
                                    {
                                        IFPElements.Add(new IFPByte(offset, visible, xtr.GetAttribute("name"), makeIndex(ref xtr, "byte"), xtr.LineNumber, -1, -1));
                                        offset += 1;
                                        break;
                                    }
                                //<bitmask name="Spawns in (none means all)">
                                case "bitmask":
                                    {
                                        goto case "bitmask32";
                                    }
                                //<bitmask32 name="Spawns in (none means all)">
                                //added 6-9-06 start
                                case "bitmask32":
                                    {
                                        IFPElements.Add(new Bitmask(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 32, xtr.LineNumber, -1, -1));
                                        offset += 4;
                                        break;
                                    }
                                //<bitmask16 name="Spawns in (none means all)">
                                case "bitmask16":
                                    {
                                        IFPElements.Add(new Bitmask(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 16, xtr.LineNumber, -1, -1));
                                        offset += 2;
                                        break;
                                    }
                                case "bitmask8":
                                    {
                                        IFPElements.Add(new Bitmask(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 8, xtr.LineNumber, -1, -1));
                                        offset += 1;
                                        break;
                                    }
                                //<enum name="Only Use Once">

                                case "enum":
                                    {
                                        goto case "enum32";
                                    }
                                //<enum32 name="Only Use Once">
                                case "enum32":
                                    {
                                        IFPElements.Add(new IFPEnum(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 32, xtr.LineNumber, -1, -1));
                                        offset += 4;
                                        break;
                                    }
                                //<enum16 name="Only Use Once">
                                case "enum16":
                                    {
                                        IFPElements.Add(new IFPEnum(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 16, xtr.LineNumber, -1, -1));
                                        offset += 2;
                                        break;
                                    }
                                case "enum8":
                                    {
                                        IFPElements.Add(new IFPEnum(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 8, xtr.LineNumber, -1, -1));
                                        offset += 1;
                                        break;
                                    }
                                //end
                                //<string32 name="Profile Name" />
                                case "string32":
                                    {
                                        IFPElements.Add(new IFPString(xtr.GetAttribute("name"), visible, offset, 32, false, xtr.LineNumber, -1, -1));
                                        offset += 32;
                                        break;
                                    }
                                //<string256 name="Scenario Path" />
                                case "string256":
                                    {
                                        IFPElements.Add(new IFPString(xtr.GetAttribute("name"), visible, offset, 256, false, xtr.LineNumber, -1, -1));
                                        offset += 256;
                                        break;
                                    }
                                //<unicode64 name="English Name" />
                                case "unicode64":
                                    {
                                        IFPElements.Add(new IFPString(xtr.GetAttribute("name"), visible, offset, 64, true, xtr.LineNumber, -1, -1));
                                        offset += 64;
                                        break;
                                    }
                                //<unicode256 name="English Name" />
                                case "unicode256":
                                    {
                                        IFPElements.Add(new IFPString(xtr.GetAttribute("name"), visible, offset, 256, true, xtr.LineNumber, -1, -1));
                                        offset += 256;
                                        break;
                                    }
                            }
                            break;
                        }


                    case "EndElement":
                        {
                            switch (xtr.Name)
                            {

                                case "struct":
                                    {
                                        endElement = true;
                                        break;
                                    }
                                case "plugin":
                                    {
                                        goto case "struct";
                                    }
                            }
                            break;
                        }
                }
                if (endElement == true) { break; }
            }
            size = offset;
            IFPElements.TrimToSize();
            object[] tempobjectarray = IFPElements.ToArray();
            return tempobjectarray;
        }
        private object[] Options(ref XmlTextReader xtr)
        {
            ArrayList options = new ArrayList(0);
            bool endElement = false;
            while (xtr.Read())
            {
                switch (xtr.NodeType.ToString())
                {
                    case "Element":
                        {
                            switch (xtr.Name.ToLower())
                            {
                                //<option value="1" name="CTF" />
                                case "option":
                                    {
                                        options.Add(new Option(xtr.GetAttribute("name"), xtr.GetAttribute("value"),xtr.LineNumber));
                                        break;
                                    }
                            }
                            break;
                        }
                    case "EndElement":
                        {
                            switch (xtr.Name)
                            {

                                case "struct":
                                    {
                                        endElement = true;
                                        break;
                                    }
                                case "bitmask":
                                    {
                                        goto case "struct";
                                    }
                                case "bitmask32":
                                    {
                                        goto case "struct";
                                    }
                                case "bitmask16":
                                    {
                                        goto case "struct";
                                    }
                                case "bitmask8":
                                    {
                                        goto case "struct";
                                    }
                                case "enum8":
                                    {
                                        goto case "struct";
                                    }
                                case "enum32":
                                    {
                                        goto case "struct";
                                    }
                                case "enum16":
                                    {
                                        goto case "struct";
                                    }
                                case "enum":
                                    {
                                        goto case "struct";
                                    }
                                case "plugin":
                                    {
                                        goto case "struct";
                                    }

                            }
                            break;
                        }
                }
                if (endElement == true) { break; }
            }
            object[] temp = options.ToArray();
            return temp;
        }
        #endregion
        #region Writer
        public void IFPOutput(string pluginNameAndPath)
        {
            XmlTextWriter xtw = new XmlTextWriter(pluginNameAndPath, Encoding.Default);
            //try
            //{
            xtw.Formatting = Formatting.Indented;
            xtw.WriteStartElement("plugin");
            xtw.WriteAttributeString("class", this.classType);
            xtw.WriteAttributeString("author", this.author);
            xtw.WriteAttributeString("version", this.version);
            xtw.WriteAttributeString("headersize", this.headerSize.ToString());
            for (int counter = 0; counter < this.revisions.Length; counter += 3)
            {
                xtw.WriteStartElement("revision");
                xtw.WriteAttributeString("author", this.revisions[counter]);
                xtw.WriteAttributeString("version", this.revisions[counter + 1]);
                xtw.WriteString(this.revisions[counter + 2]);
                xtw.WriteEndElement();
            }
            IFPWriter(xtw, this.items);
            xtw.WriteEndElement();
            //}
            //catch
            //{
            //    MessageBox.Show("Error in writing IFP");
            //}
            //finally
            //{
            xtw.Close();
            //}
        }
        private void IFPWriter(XmlTextWriter xtw, object[] itemsToWrite)
        {
            for (int counter = 0; counter < itemsToWrite.Length; counter++)
            {
                string x = itemsToWrite[counter].GetType().ToString();
                switch (itemsToWrite[counter].GetType().ToString())
                {

                    case "entity.IFPIO+Reflexive":
                        {
                            //<struct name="Skies11" offset="8" visible="true" size="8">
                            xtw.WriteStartElement("struct");
                            xtw.WriteAttributeString("name", ((Reflexive)itemsToWrite[counter]).name);
                            xtw.WriteAttributeString("offset", ((Reflexive)itemsToWrite[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((Reflexive)itemsToWrite[counter]).visible.ToString().ToLower());
                            xtw.WriteAttributeString("size", ((Reflexive)itemsToWrite[counter]).chunkSize.ToString());
                            xtw.WriteAttributeString("label", ((Reflexive)itemsToWrite[counter]).label);
                            if (((Reflexive)itemsToWrite[counter]).HasCount == false)
                            {
                                xtw.WriteAttributeString("HasCount", ((Reflexive)itemsToWrite[counter]).HasCount.ToString());
                            }
                            IFPWriter(xtw, ((Reflexive)itemsToWrite[counter]).items);
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+SID":
                        {
                            //<stringid name="Unknown" offset="52" visible="false" />
                            xtw.WriteStartElement("stringid");
                            xtw.WriteAttributeString("name", ((SID)itemsToWrite[counter]).name);
                            xtw.WriteAttributeString("offset", ((SID)itemsToWrite[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((SID)itemsToWrite[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+TagType":
                        {
                            xtw.WriteStartElement("tag");
                           
                            xtw.WriteAttributeString("name", ((TagType)itemsToWrite[counter]).name);
                            xtw.WriteAttributeString("offset", (((TagType)itemsToWrite[counter]).offset).ToString());
                            xtw.WriteAttributeString("visible", ((TagType)itemsToWrite[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+Ident":
                        {
                            xtw.WriteStartElement("id");
                            xtw.WriteAttributeString("name", ((Ident)itemsToWrite[counter]).name);
                            xtw.WriteAttributeString("offset", (((Ident)itemsToWrite[counter]).offset).ToString());
                            xtw.WriteAttributeString("visible", ((Ident)itemsToWrite[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+IFPString":
                        {
                            switch (((IFPString)itemsToWrite[counter]).type)
                            {
                                case true:
                                    {
                                        //<unicode64 name="English Name" />
                                        if (((IFPString)itemsToWrite[counter]).size == 64) { xtw.WriteStartElement("unicode64"); }
                                        else { xtw.WriteStartElement("unicode256"); }
                                        xtw.WriteAttributeString("name", ((IFPString)itemsToWrite[counter]).name);
                                        xtw.WriteAttributeString("offset", ((IFPString)itemsToWrite[counter]).offset.ToString());
                                        xtw.WriteAttributeString("visible", ((IFPString)itemsToWrite[counter]).visible.ToString());
                                        xtw.WriteEndElement();
                                        break;
                                    }
                                case false:
                                    {
                                        if (((IFPString)itemsToWrite[counter]).size == 32) { xtw.WriteStartElement("string32"); }
                                        else { xtw.WriteStartElement("string256"); }
                                        xtw.WriteAttributeString("name", ((IFPString)itemsToWrite[counter]).name);
                                        xtw.WriteAttributeString("offset", ((IFPString)itemsToWrite[counter]).offset.ToString());
                                        xtw.WriteAttributeString("visible", ((IFPString)itemsToWrite[counter]).visible.ToString());
                                        xtw.WriteEndElement();
                                        break;
                                    }
                            }
                            break;
                        }
                    case "entity.IFPIO+Unused":
                        {
                            xtw.WriteStartElement("unused");
                            xtw.WriteAttributeString("offset", ((Unused)itemsToWrite[counter]).offset.ToString());
                            xtw.WriteAttributeString("size", ((Unused)itemsToWrite[counter]).size.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+Unknown":
                        {
                            xtw.WriteStartElement("undefined");
                            xtw.WriteAttributeString("name", ((Unknown)itemsToWrite[counter]).name);
                            xtw.WriteAttributeString("offset", ((Unknown)itemsToWrite[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((Unknown)itemsToWrite[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+IFPFloat":
                        {
                            xtw.WriteStartElement("float");
                            xtw.WriteAttributeString("name", ((IFPFloat)itemsToWrite[counter]).name);
                            xtw.WriteAttributeString("offset", ((IFPFloat)itemsToWrite[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((IFPFloat)itemsToWrite[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+IFPInt":
                        {
                            switch (((IFPInt)itemsToWrite[counter]).ObjectType)
                            {
                                case ObjectEnum.Int:
                                    {
                                        xtw.WriteStartElement("int");
                                        break;
                                    }
                                case ObjectEnum.UInt:
                                    {
                                        xtw.WriteStartElement("uint");
                                        break;
                                    }
                                case ObjectEnum.Short:
                                    {
                                        xtw.WriteStartElement("short");
                                        break;
                                    }
                                case ObjectEnum.UShort:
                                    {
                                        xtw.WriteStartElement("ushort");
                                        break;
                                    }
                            }                            
                            xtw.WriteAttributeString("name", ((IFPInt)itemsToWrite[counter]).name);
                            xtw.WriteAttributeString("offset", ((IFPInt)itemsToWrite[counter]).offset.ToString());
                            if (((IFPInt)itemsToWrite[counter]).entIndex.nulled == false)
                            {
                                xtw.WriteAttributeString("reflexiveoffset", ((IFPInt)itemsToWrite[counter]).entIndex.ReflexiveOffset.ToString());
                                xtw.WriteAttributeString("reflexivesize", ((IFPInt)itemsToWrite[counter]).entIndex.ReflexiveSize.ToString());
                                xtw.WriteAttributeString("itemoffset", ((IFPInt)itemsToWrite[counter]).entIndex.ItemOffset.ToString());
                                xtw.WriteAttributeString("itemtype", ((IFPInt)itemsToWrite[counter]).entIndex.ItemType.ToString());
                                xtw.WriteAttributeString("layer", ((IFPInt)itemsToWrite[counter]).entIndex.reflexiveLayer.ToString());
                            }
                            xtw.WriteAttributeString("visible", ((IFPInt)itemsToWrite[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+IFPByte":
                        {
                            //"reflexiveoffset"), xtr.GetAttribute("reflexivesize"), xtr.GetAttribute("itemoffset"), xtr.GetAttribute("itemtype"), xtr.GetAttribute("layer")
                            xtw.WriteStartElement("byte");
                            xtw.WriteAttributeString("name", ((IFPByte)itemsToWrite[counter]).name);
                            xtw.WriteAttributeString("offset", ((IFPByte)itemsToWrite[counter]).offset.ToString());
                            if (((IFPByte)itemsToWrite[counter]).entIndex.nulled == false)
                            {
                                xtw.WriteAttributeString("reflexiveoffset", ((IFPByte)itemsToWrite[counter]).entIndex.ReflexiveOffset.ToString());
                                xtw.WriteAttributeString("reflexivesize", ((IFPByte)itemsToWrite[counter]).entIndex.ReflexiveSize.ToString());
                                xtw.WriteAttributeString("itemoffset", ((IFPByte)itemsToWrite[counter]).entIndex.ItemOffset.ToString());
                                xtw.WriteAttributeString("itemtype", ((IFPByte)itemsToWrite[counter]).entIndex.ItemType.ToString());
                                xtw.WriteAttributeString("layer", ((IFPByte)itemsToWrite[counter]).entIndex.reflexiveLayer.ToString());
                            }
                            xtw.WriteAttributeString("visible", ((IFPByte)itemsToWrite[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    //Added 6-9-06 Start
                    case "entity.IFPIO+Bitmask":
                        {
                            switch (((Bitmask)itemsToWrite[counter]).bitmaskSize)
                            {
                                case 8:
                                    {
                                        xtw.WriteStartElement("bitmask8");
                                        break;
                                    }
                                case 16:
                                    {
                                        xtw.WriteStartElement("bitmask16");
                                        break;
                                    }
                                case 32:
                                    {
                                        xtw.WriteStartElement("bitmask32");
                                        break;
                                    }
                            }
                            xtw.WriteAttributeString("name", ((Bitmask)itemsToWrite[counter]).name);
                            xtw.WriteAttributeString("offset", ((Bitmask)itemsToWrite[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((Bitmask)itemsToWrite[counter]).visible.ToString());
                            if (((Bitmask)itemsToWrite[counter]).options != null)
                            {
                                for (int counter2 = 0; counter2 < ((Bitmask)itemsToWrite[counter]).options.Length; counter2++)
                                {
                                    xtw.WriteStartElement("option");
                                    xtw.WriteAttributeString("name", ((Option)((Bitmask)itemsToWrite[counter]).options[counter2]).name);
                                    xtw.WriteAttributeString("value", ((Option)((Bitmask)itemsToWrite[counter]).options[counter2]).value.ToString());
                                    xtw.WriteEndElement();
                                }
                            }
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+IFPEnum":
                        {
                            switch (((IFPEnum)itemsToWrite[counter]).enumSize)
                            {
                                case 8:
                                    {
                                        xtw.WriteStartElement("enum8");
                                        break;
                                    }
                                case 16:
                                    {
                                        xtw.WriteStartElement("enum16");
                                        break;
                                    }
                                case 32:
                                    {
                                        xtw.WriteStartElement("enum32");
                                        break;
                                    }
                            }
                            xtw.WriteAttributeString("name", ((IFPEnum)itemsToWrite[counter]).name);
                            xtw.WriteAttributeString("offset", ((IFPEnum)itemsToWrite[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((IFPEnum)itemsToWrite[counter]).visible.ToString());
                            if (((IFPEnum)itemsToWrite[counter]).options != null)
                            {
                                for (int counter2 = 0; counter2 < ((IFPEnum)itemsToWrite[counter]).options.Length; counter2++)
                                {
                                    xtw.WriteStartElement("option");
                                    xtw.WriteAttributeString("name", ((Option)((IFPEnum)itemsToWrite[counter]).options[counter2]).name);
                                    xtw.WriteAttributeString("value", ((Option)((IFPEnum)itemsToWrite[counter]).options[counter2]).value.ToString());
                                    xtw.WriteEndElement();
                                }
                            }
                            xtw.WriteEndElement();
                            break;
                        }
                    //End
                }
            }
        }
        #endregion
        #endregion
        #region New (Child/Sibling)
        #region Reader
        public void newReadIFP(string inputFilePath)
        {
            FilePath = inputFilePath;
            ENTElements = new List<BaseObject>(0);
            FileStream fs = new FileStream(inputFilePath, FileMode.Open, FileAccess.ReadWrite);
            XmlTextReader xtr = new XmlTextReader(fs);
            xtr.WhitespaceHandling = WhitespaceHandling.None;
            NewIFPReader(ref xtr,-1,false);
            alRevisions.TrimToSize();
            revisions = ((string[])alRevisions.ToArray(typeof(string)));
            alRevisions.Clear();
            headerSize = size;
            xtr.Close();
            fs.Close();
        }
        //function calls its self recursively
        private void NewIFPReader(ref XmlTextReader xtr,int parent,bool nested)
        {            
            int offset = 0;
            bool visible;
            int dependencyOffset = -1;
            bool endElement = false;
            int sibling = -1;
            int tempsibint = 0;
            while (xtr.Read())
            {
                visible = xtr.GetAttribute("visible") != null ? xtr.GetAttribute("visible").ToLower() == "false" ? false : true : true;
                tempsibint = ENTElements.Count;                
                switch (xtr.NodeType.ToString())
                {
                    case "Element":
                        {
                            if (sibling > -1 && ENTElements.Count > 0)
                                ENTElements[sibling].siblingNext = tempsibint;
                            switch (xtr.Name.ToLower())
                            {
                                //<plugin class="bipd" author="Iron_Forge" version="0.3">
                                case "plugin":
                                    {
                                        author = xtr.GetAttribute("author");
                                        version = xtr.GetAttribute("version");
                                        classType = xtr.GetAttribute("class");
                                        break;
                                    }
                                //<revision author="Iron_Forge" version="0.3">Added some known values...</revision>
                                case "revision":
                                    {
                                        alRevisions.Add(xtr.GetAttribute("author"));
                                        alRevisions.Add(xtr.GetAttribute("version"));
                                        alRevisions.Add(xtr.ReadString());
                                        break;
                                    }
                                // <struct name="Predicted Resources24" offset="32" visible="true" size="8">
                                case "struct":
                                    {                                        
                                        bool temp = true;
                                        string i = xtr.GetAttribute("HasCount");
                                        if (i != null)
                                        {
                                            temp = bool.Parse(i);
                                        }
                                        ENTElements.Add(new Reflexive(xtr.LineNumber, offset, visible, xtr.GetAttribute("name"), xtr.GetAttribute("label"), null, size, temp, parent, sibling));
                                        NewIFPReader(ref xtr, tempsibint, true);
                                        if (tempsibint != this.ENTElements.Count - 1)
                                        {
                                            this.ENTElements[this.ENTElements.Count - 1].siblingNext = -1;
                                            ENTElements[tempsibint].child = tempsibint + 1;
                                            ((Reflexive)ENTElements[tempsibint]).chunkSize = size;
                                        }
                                        if (temp == false)
                                        {
                                            offset += 4;
                                        }
                                        else
                                        {
                                            offset += 8;
                                        }                                        
                                        break;
                                    }
                                // <tag name="" visible="false"/>
                                case "tag":
                                    {
                                        ENTElements.Add(new TagType(offset, visible, xtr.GetAttribute("name"), xtr.LineNumber,parent, sibling));
                                        offset += 4;
                                        dependencyOffset = offset;
                                        break;
                                    }
                                // <id name="" visible="false"/>
                                case "id":
                                    {
                                        bool isDependency = dependencyOffset == offset ? true : false;
                                        ENTElements.Add(new Ident(xtr.GetAttribute("name"), visible, offset, isDependency, xtr.LineNumber,parent, sibling));
                                        offset += 4;
                                        break;
                                    }
                                //  <stringid name="" visible="false" />
                                case "stringid":
                                    {
                                        ENTElements.Add(new SID(xtr.GetAttribute("name"), visible, offset, xtr.LineNumber, parent, sibling));
                                        offset += 4;
                                        break;
                                    }
                                case "script":
                                    {
                                        goto case "stringid";
                                    }
                                //<unused size="32" default="0" />
                                case "unused":
                                    {
                                        int sizeOfUnusedSpace = Convert.ToInt32(xtr.GetAttribute("size"));
                                        ENTElements.Add(new Unused(offset, sizeOfUnusedSpace, xtr.LineNumber, parent, sibling));
                                        offset += sizeOfUnusedSpace;
                                        break;
                                    }
                                //<undefined name="Unknown" visible="false" />
                                case "undefined":
                                    {
                                        ENTElements.Add(new Unknown(offset, visible, xtr.GetAttribute("name"), xtr.LineNumber, parent, sibling));
                                        offset += 4;
                                        break;
                                    }
                                //<float name="" visible="false" />
                                case "float":
                                    {
                                        ENTElements.Add(new IFPFloat(offset, visible, xtr.GetAttribute("name"), xtr.LineNumber, parent, sibling));
                                        offset += 4;
                                        break;
                                    }
                                //<int name="Unknown" visible="false" />
                                case "int":
                                    {
                                        ENTElements.Add(new IFPInt(offset, ObjectEnum.Int, visible, xtr.GetAttribute("name"), makeIndex(ref xtr, "int"), xtr.LineNumber, parent, sibling));
                                        offset += 4;
                                        break;
                                    }
                                //<uint name="Structure BSP Offset" />
                                case "uint":
                                    {
                                        ENTElements.Add(new IFPInt(offset, ObjectEnum.UInt, visible, xtr.GetAttribute("name"), makeIndex(ref xtr, "uint"), xtr.LineNumber, parent, sibling));
                                        offset += 4;
                                        break;
                                    }
                                //<short name="Palette Chunk #" visible="true" />
                                case "short":
                                    {
                                        ENTElements.Add(new IFPInt(offset, ObjectEnum.Short, visible, xtr.GetAttribute("name"), makeIndex(ref xtr, "short"), xtr.LineNumber, parent, sibling));
                                        offset += 2;
                                        break;
                                    }
                                //<ushort name="Palette Chunk #" visible="true" />
                                case "ushort":
                                    {
                                        ENTElements.Add(new IFPInt(offset, ObjectEnum.UShort, visible, xtr.GetAttribute("name"), makeIndex(ref xtr, "ushort"), xtr.LineNumber, parent, sibling));
                                        offset += 2;
                                        break;
                                    }
                                //<byte name="Unknown" visible="false" />
                                case "byte":
                                    {
                                        ENTElements.Add(new IFPByte(offset, visible, xtr.GetAttribute("name"), makeIndex(ref xtr, "byte"), xtr.LineNumber, parent, sibling));
                                        offset += 1;
                                        break;
                                    }
                                //<bitmask name="Spawns in (none means all)">
                                case "bitmask":
                                    {
                                        goto case "bitmask32";
                                    }
                                //<bitmask32 name="Spawns in (none means all)">
                                //added 6-9-06 start
                                case "bitmask32":
                                    {
                                        ENTElements.Add(new Bitmask(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 32, xtr.LineNumber, parent, sibling));
                                        offset += 4;
                                        break;
                                    }
                                //<bitmask16 name="Spawns in (none means all)">
                                case "bitmask16":
                                    {
                                        ENTElements.Add(new Bitmask(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 16, xtr.LineNumber, parent, sibling));
                                        offset += 2;
                                        break;
                                    }
                                case "bitmask8":
                                    {
                                        ENTElements.Add(new Bitmask(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 8, xtr.LineNumber, parent, sibling));
                                        offset += 1;
                                        break;
                                    }
                                //<enum name="Only Use Once">

                                case "enum":
                                    {
                                        goto case "enum32";
                                    }
                                //<enum32 name="Only Use Once">
                                case "enum32":
                                    {
                                        ENTElements.Add(new IFPEnum(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 32, xtr.LineNumber, parent, sibling));
                                        offset += 4;
                                        break;
                                    }
                                //<enum16 name="Only Use Once">
                                case "enum16":
                                    {
                                        ENTElements.Add(new IFPEnum(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 16, xtr.LineNumber, parent, sibling));
                                        offset += 2;
                                        break;
                                    }
                                case "enum8":
                                    {
                                        ENTElements.Add(new IFPEnum(offset, visible, xtr.GetAttribute("name"), xtr.IsEmptyElement == false ? Options(ref xtr) : null, 8, xtr.LineNumber, parent, sibling));
                                        offset += 1;
                                        break;
                                    }
                                //end
                                //<string32 name="Profile Name" />
                                case "string32":
                                    {
                                        ENTElements.Add(new IFPString(xtr.GetAttribute("name"), visible, offset, 32, false, xtr.LineNumber, parent, sibling));
                                        offset += 32;
                                        break;
                                    }
                                //<string256 name="Scenario Path" />
                                case "string256":
                                    {
                                        ENTElements.Add(new IFPString(xtr.GetAttribute("name"), visible, offset, 256, false, xtr.LineNumber, parent, sibling));
                                        offset += 256;
                                        break;
                                    }
                                //<unicode64 name="English Name" />
                                case "unicode64":
                                    {
                                        ENTElements.Add(new IFPString(xtr.GetAttribute("name"), visible, offset, 64, true, xtr.LineNumber, parent, sibling));
                                        offset += 64;
                                        break;
                                    }
                                //<unicode256 name="English Name" />
                                case "unicode256":
                                    {
                                        ENTElements.Add(new IFPString(xtr.GetAttribute("name"), visible, offset, 256, true, xtr.LineNumber, parent, sibling));
                                        offset += 256;
                                        break;
                                    }
                            }
                            break;
                        }


                    case "EndElement":
                        {
                            switch (xtr.Name)
                            {

                                case "struct":
                                    {
                                        endElement = true;
                                        break;
                                    }
                                case "plugin":
                                    {
                                        goto case "struct";
                                    }
                            }
                            break;
                        }
                }
                sibling = ENTElements.Count < 1 ? -1 : tempsibint;
                if (endElement == true) { break; }
            }
            size = offset;
            ENTElements.TrimExcess();
        }
#endregion
        #region Writer
        public void EntOutput(string pluginNameAndPath,bool reWritePlugin)
        {
            if (reWritePlugin == true)
                pluginNameAndPath = FilePath;
            XmlTextWriter xtw = new XmlTextWriter(pluginNameAndPath, Encoding.Default);
            xtw.Formatting = Formatting.Indented;
            xtw.WriteStartElement("plugin");
            xtw.WriteAttributeString("class", this.classType);
            xtw.WriteAttributeString("author", this.author);
            xtw.WriteAttributeString("version", this.version);
            xtw.WriteAttributeString("headersize", this.headerSize.ToString());
            for (int counter = 0; counter < this.revisions.Length; counter += 3)
            {
                xtw.WriteStartElement("revision");
                xtw.WriteAttributeString("author", this.revisions[counter]);
                xtw.WriteAttributeString("version", this.revisions[counter + 1]);
                xtw.WriteString(this.revisions[counter + 2]);
                xtw.WriteEndElement();
            }
            EntWriter(xtw,0);
            xtw.WriteEndElement();
            xtw.Close();
        }
        private void EntWriter(XmlTextWriter xtw, int counter)
        {
            while (counter != -1)
            {
                switch (this.ENTElements[counter].GetType().ToString())
                {

                    case "entity.IFPIO+Reflexive":
                        {
                            //<struct name="Skies11" offset="8" visible="true" size="8">
                            xtw.WriteStartElement("struct");
                            xtw.WriteAttributeString("name", ((Reflexive)this.ENTElements[counter]).name);
                            xtw.WriteAttributeString("offset", ((Reflexive)this.ENTElements[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((Reflexive)this.ENTElements[counter]).visible.ToString().ToLower());
                            xtw.WriteAttributeString("size", ((Reflexive)this.ENTElements[counter]).chunkSize.ToString());
                            xtw.WriteAttributeString("label", ((Reflexive)this.ENTElements[counter]).label);
                            if (((Reflexive)this.ENTElements[counter]).HasCount == false)
                            {
                                xtw.WriteAttributeString("HasCount", ((Reflexive)this.ENTElements[counter]).HasCount.ToString());
                            }
                            EntWriter(xtw, this.ENTElements[counter].child);
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+SID":
                        {
                            //<stringid name="Unknown" offset="52" visible="false" />
                            xtw.WriteStartElement("stringid");
                            xtw.WriteAttributeString("name", ((SID)this.ENTElements[counter]).name);
                            xtw.WriteAttributeString("offset", ((SID)this.ENTElements[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((SID)this.ENTElements[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+TagType":
                        {
                            xtw.WriteStartElement("tag");

                            xtw.WriteAttributeString("name", ((TagType)this.ENTElements[counter]).name);
                            xtw.WriteAttributeString("offset", (((TagType)this.ENTElements[counter]).offset).ToString());
                            xtw.WriteAttributeString("visible", ((TagType)this.ENTElements[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+Ident":
                        {
                            xtw.WriteStartElement("id");
                            xtw.WriteAttributeString("name", ((Ident)this.ENTElements[counter]).name);
                            xtw.WriteAttributeString("offset", (((Ident)this.ENTElements[counter]).offset).ToString());
                            xtw.WriteAttributeString("visible", ((Ident)this.ENTElements[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+IFPString":
                        {
                            switch (((IFPString)this.ENTElements[counter]).type)
                            {
                                case true:
                                    {
                                        //<unicode64 name="English Name" />
                                        if (((IFPString)this.ENTElements[counter]).size == 64) { xtw.WriteStartElement("unicode64"); }
                                        else { xtw.WriteStartElement("unicode256"); }
                                        xtw.WriteAttributeString("name", ((IFPString)this.ENTElements[counter]).name);
                                        xtw.WriteAttributeString("offset", ((IFPString)this.ENTElements[counter]).offset.ToString());
                                        xtw.WriteAttributeString("visible", ((IFPString)this.ENTElements[counter]).visible.ToString());
                                        xtw.WriteEndElement();
                                        break;
                                    }
                                case false:
                                    {
                                        if (((IFPString)this.ENTElements[counter]).size == 32) { xtw.WriteStartElement("string32"); }
                                        else { xtw.WriteStartElement("string256"); }
                                        xtw.WriteAttributeString("name", ((IFPString)this.ENTElements[counter]).name);
                                        xtw.WriteAttributeString("offset", ((IFPString)this.ENTElements[counter]).offset.ToString());
                                        xtw.WriteAttributeString("visible", ((IFPString)this.ENTElements[counter]).visible.ToString());
                                        xtw.WriteEndElement();
                                        break;
                                    }
                            }
                            break;
                        }
                    case "entity.IFPIO+Unused":
                        {
                            xtw.WriteStartElement("unused");
                            xtw.WriteAttributeString("offset", ((Unused)this.ENTElements[counter]).offset.ToString());
                            xtw.WriteAttributeString("size", ((Unused)this.ENTElements[counter]).size.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+Unknown":
                        {
                            xtw.WriteStartElement("undefined");
                            xtw.WriteAttributeString("name", ((Unknown)this.ENTElements[counter]).name);
                            xtw.WriteAttributeString("offset", ((Unknown)this.ENTElements[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((Unknown)this.ENTElements[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+IFPFloat":
                        {
                            xtw.WriteStartElement("float");
                            xtw.WriteAttributeString("name", ((IFPFloat)this.ENTElements[counter]).name);
                            xtw.WriteAttributeString("offset", ((IFPFloat)this.ENTElements[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((IFPFloat)this.ENTElements[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+IFPInt":
                        {
                            switch (((IFPInt)ENTElements[counter]).ObjectType)
                            {
                                case ObjectEnum.Int:
                                    {
                                        xtw.WriteStartElement("int");
                                        break;
                                    }
                                case ObjectEnum.UInt:
                                    {
                                        xtw.WriteStartElement("uint");
                                        break;
                                    }
                                case ObjectEnum.Short:
                                    {
                                        xtw.WriteStartElement("short");
                                        break;
                                    }
                                case ObjectEnum.UShort:
                                    {
                                        xtw.WriteStartElement("ushort");
                                        break;
                                    }
                            }  
                            xtw.WriteAttributeString("name", ((IFPInt)this.ENTElements[counter]).name);
                            xtw.WriteAttributeString("offset", ((IFPInt)this.ENTElements[counter]).offset.ToString());
                            if (((IFPInt)this.ENTElements[counter]).entIndex.nulled == false)
                            {
                                xtw.WriteAttributeString("reflexiveoffset", ((IFPInt)this.ENTElements[counter]).entIndex.ReflexiveOffset.ToString());
                                xtw.WriteAttributeString("reflexivesize", ((IFPInt)this.ENTElements[counter]).entIndex.ReflexiveSize.ToString());
                                xtw.WriteAttributeString("itemoffset", ((IFPInt)this.ENTElements[counter]).entIndex.ItemOffset.ToString());
                                xtw.WriteAttributeString("itemtype", ((IFPInt)this.ENTElements[counter]).entIndex.ItemType.ToString());
                                xtw.WriteAttributeString("layer", ((IFPInt)this.ENTElements[counter]).entIndex.reflexiveLayer.ToString());
                            }
                            xtw.WriteAttributeString("visible", ((IFPInt)this.ENTElements[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+IFPByte":
                        {
                            //"reflexiveoffset"), xtr.GetAttribute("reflexivesize"), xtr.GetAttribute("itemoffset"), xtr.GetAttribute("itemtype"), xtr.GetAttribute("layer")
                            xtw.WriteStartElement("byte");
                            xtw.WriteAttributeString("name", ((IFPByte)this.ENTElements[counter]).name);
                            xtw.WriteAttributeString("offset", ((IFPByte)this.ENTElements[counter]).offset.ToString());
                            if (((IFPByte)this.ENTElements[counter]).entIndex.nulled == false)
                            {
                                xtw.WriteAttributeString("reflexiveoffset", ((IFPByte)this.ENTElements[counter]).entIndex.ReflexiveOffset.ToString());
                                xtw.WriteAttributeString("reflexivesize", ((IFPByte)this.ENTElements[counter]).entIndex.ReflexiveSize.ToString());
                                xtw.WriteAttributeString("itemoffset", ((IFPByte)this.ENTElements[counter]).entIndex.ItemOffset.ToString());
                                xtw.WriteAttributeString("itemtype", ((IFPByte)this.ENTElements[counter]).entIndex.ItemType.ToString());
                                xtw.WriteAttributeString("layer", ((IFPByte)this.ENTElements[counter]).entIndex.reflexiveLayer.ToString());
                            }
                            xtw.WriteAttributeString("visible", ((IFPByte)this.ENTElements[counter]).visible.ToString());
                            xtw.WriteEndElement();
                            break;
                        }
                    //Added 6-9-06 Start
                    case "entity.IFPIO+Bitmask":
                        {
                            switch (((Bitmask)this.ENTElements[counter]).bitmaskSize)
                            {
                                case 8:
                                    {
                                        xtw.WriteStartElement("bitmask8");
                                        break;
                                    }
                                case 16:
                                    {
                                        xtw.WriteStartElement("bitmask16");
                                        break;
                                    }
                                case 32:
                                    {
                                        xtw.WriteStartElement("bitmask32");
                                        break;
                                    }
                            }
                            xtw.WriteAttributeString("name", ((Bitmask)this.ENTElements[counter]).name);
                            xtw.WriteAttributeString("offset", ((Bitmask)this.ENTElements[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((Bitmask)this.ENTElements[counter]).visible.ToString());
                            if (((Bitmask)this.ENTElements[counter]).options != null)
                            {
                                for (int counter2 = 0; counter2 < ((Bitmask)this.ENTElements[counter]).options.Length; counter2++)
                                {
                                    xtw.WriteStartElement("option");
                                    xtw.WriteAttributeString("name", ((Option)((Bitmask)this.ENTElements[counter]).options[counter2]).name);
                                    xtw.WriteAttributeString("value", ((Option)((Bitmask)this.ENTElements[counter]).options[counter2]).value.ToString());
                                    xtw.WriteEndElement();
                                }
                            }
                            xtw.WriteEndElement();
                            break;
                        }
                    case "entity.IFPIO+IFPEnum":
                        {
                            switch (((IFPEnum)this.ENTElements[counter]).enumSize)
                            {
                                case 8:
                                    {
                                        xtw.WriteStartElement("enum8");
                                        break;
                                    }
                                case 16:
                                    {
                                        xtw.WriteStartElement("enum16");
                                        break;
                                    }
                                case 32:
                                    {
                                        xtw.WriteStartElement("enum32");
                                        break;
                                    }
                            }
                            xtw.WriteAttributeString("name", ((IFPEnum)this.ENTElements[counter]).name);
                            xtw.WriteAttributeString("offset", ((IFPEnum)this.ENTElements[counter]).offset.ToString());
                            xtw.WriteAttributeString("visible", ((IFPEnum)this.ENTElements[counter]).visible.ToString());
                            if (((IFPEnum)this.ENTElements[counter]).options != null)
                            {
                                for (int counter2 = 0; counter2 < ((IFPEnum)this.ENTElements[counter]).options.Length; counter2++)
                                {
                                    xtw.WriteStartElement("option");
                                    xtw.WriteAttributeString("name", ((Option)((IFPEnum)this.ENTElements[counter]).options[counter2]).name);
                                    xtw.WriteAttributeString("value", ((Option)((IFPEnum)this.ENTElements[counter]).options[counter2]).value.ToString());
                                    xtw.WriteEndElement();
                                }
                            }
                            xtw.WriteEndElement();
                            break;
                        }
                    //End
                }
                counter = this.ENTElements[counter].siblingNext;
            }

        }
        #endregion
        #endregion
        private Index makeIndex(ref XmlTextReader xtr,string type)
        {
            return new Index(xtr.GetAttribute("reflexiveoffset"), xtr.GetAttribute("reflexivesize"), xtr.GetAttribute("itemoffset"), xtr.GetAttribute("itemtype"), xtr.GetAttribute("layer"), xtr.LineNumber);
        }
        #endregion
        #region enums
        public enum ObjectEnum
        {
            Reflexive,
            StringID,
            TagType,
            Ident,
            String32,
            UnicodeString64,
            String256,
            UnicodeString256,
            Unused,
            Unknown,
            Float,
            Int,
            UInt,
            Short,
            UShort,
            Byte,            
            Bitmask,
            Enum,
            Option,
            Index
        }
        #endregion
        #region classes

        public class BaseObject
        {
            #region fields
            public string name;
            public int child = -1;
            public int siblingNext = -1;
            public int siblingPrevious = -1;
            public int parent = -1;
            public int offset;          
            public bool visible;
            public int lineNumber;
            public ObjectEnum ObjectType;
            #endregion
        }
        public class Reflexive : BaseObject
        {
            #region fields
            
            public int totalSize;
            public int chunkSize;
            public bool HasCount = true;
            
            public string label;
            public object[] items;
            #endregion
            #region Constructor
            public Reflexive(int entlineNumber, int ifpOffset, bool ifpVisible, string ifpName, string ifpLabel, object[] ifpitems, int ifpChunkSize, bool hascount, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.ObjectType = ObjectEnum.Reflexive;
                this.offset = ifpOffset;
                this.visible = ifpVisible;
                this.chunkSize = ifpChunkSize;
                this.name = ifpName;
                this.label = ifpLabel;
                this.items = ifpitems;
                this.HasCount = hascount;
            }
            #endregion
        }
        public class SID : BaseObject
        {
            #region fields
            #endregion
            #region Constructor
            public SID(string ifpName, bool ifpVisible, int ifpOffset, int entlineNumber, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.ObjectType = ObjectEnum.StringID;
                this.offset = ifpOffset;
                this.name = ifpName;
                this.visible = ifpVisible;
            }
            #endregion
        }
        public class Ident : BaseObject
        {
            #region fields
            public bool hasTagType;
            #endregion
            #region Constructor
            public Ident(string ifpName, bool ifpVisible, int ifpOffset, bool ifpHasTagType, int entlineNumber, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.ObjectType = ObjectEnum.Ident;
                this.offset = ifpOffset;
                this.name = ifpName;
                this.visible = ifpVisible;
                this.hasTagType = ifpHasTagType;
            }
            #endregion
        }
        public class IFPString : BaseObject
        {
            #region fields
            public int size;
            public bool type;//true means unicode, false means normal string
            #endregion
            #region Constructor
            public IFPString(string ifpName, bool ifpVisible, int ifpOffset, int ifpSize, bool ifpType, int entlineNumber, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                switch (ifpType)
                {
                    case true:
                        {
                            //<unicode64 name="English Name" />
                            if (ifpSize == 64) { this.ObjectType = ObjectEnum.UnicodeString64; }
                            else if (ifpSize == 256) { this.ObjectType = ObjectEnum.UnicodeString256; }
                            break;
                        }
                    case false:
                        {
                            if (ifpSize == 32) { this.ObjectType = ObjectEnum.String32; }
                            else if (ifpSize == 256) { this.ObjectType = ObjectEnum.String256; }
                            break;
                        }
                }
                this.offset = ifpOffset;
                this.name = ifpName;
                this.visible = ifpVisible;
                this.size = ifpSize;
                this.type = ifpType;
            }
            #endregion
        }
        public class Unused : BaseObject
        {
            #region fields
            public int size;
            #endregion
            #region Constructor
            public Unused(int ifpOffset, int ifpSize, int entlineNumber, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.ObjectType = ObjectEnum.Unused;
                this.offset = ifpOffset;
                this.size = ifpSize;
            }
            #endregion
        }
        public class Unknown : BaseObject
        {
            #region fields
            #endregion
            #region Constructor
            public Unknown(int ifpoffset, bool ifpvisible, string ifpname, int entlineNumber, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.ObjectType = ObjectEnum.Unknown;
                this.offset = ifpoffset;
                this.visible = ifpvisible;
                this.name = ifpname;
            }
            #endregion
        }
        public class TagType : BaseObject
        {
            #region fields
            #endregion
            #region Constructor
            public TagType(int ifpoffset, bool ifpvisible, string ifpname, int entlineNumber, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.ObjectType = ObjectEnum.TagType;
                this.offset = ifpoffset;
                this.visible = ifpvisible;
                this.name = ifpname;
            }
            #endregion
        }
        public class IFPFloat : BaseObject
        {
            #region fields
            #endregion
            #region Constructor
            public IFPFloat(int ifpoffset, bool ifpvisible, string ifpname, int entlineNumber, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.ObjectType = ObjectEnum.Float;
                this.offset = ifpoffset;
                this.visible = ifpvisible;
                this.name = ifpname;
            }
            #endregion
        }
        public class IFPInt : BaseObject
        {
            #region fields
            public Index entIndex;
            #endregion
            #region Constructor
            public IFPInt(int ifpoffset,ObjectEnum iType, bool ifpvisible, string ifpname, Index iIndex, int entlineNumber, int entparent, int entPrevSibling)
            {
                
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.entIndex = iIndex;
                this.ObjectType = iType;
                this.offset = ifpoffset;
                this.visible = ifpvisible;
                this.name = ifpname;
            }
            #endregion
        }
        public class IFPByte : BaseObject
        {
            #region fields
            public Index entIndex;
            #endregion
            #region Constructor
            public IFPByte(int ifpoffset, bool ifpvisible, string ifpname, Index iIndex, int entlineNumber, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.entIndex = iIndex;
                this.ObjectType = ObjectEnum.Byte;
                this.offset = ifpoffset;
                this.visible = ifpvisible;
                this.name = ifpname;
            }
            #endregion
        }
        //added 6-8-06
        //Start
        public class Bitmask : BaseObject
        {
            #region fields
            public object[] options;
            public int bitmaskSize;
            #endregion
            #region Constructor
            public Bitmask(int ifpoffset, bool ifpvisible, string ifpname, object[] ifpoptions, int ifpsize, int entlineNumber, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.ObjectType = ObjectEnum.Bitmask;
                this.offset = ifpoffset;
                this.name = ifpname;
                this.options = ifpoptions;
                this.visible = ifpvisible;
                this.bitmaskSize = ifpsize;
            }
            #endregion
        }
        public class IFPEnum : BaseObject
        {
            #region fields
            public object[] options;
            public int enumSize;
            #endregion
            #region Constructor
            public IFPEnum(int ifpoffset, bool ifpvisible, string ifpname, object[] ifpoptions, int ifpsize, int entlineNumber, int entparent, int entPrevSibling)
            {
                this.siblingPrevious = entPrevSibling;
                this.parent = entparent;
                this.lineNumber = entlineNumber;
                this.ObjectType = ObjectEnum.Enum;
                this.offset = ifpoffset;
                this.name = ifpname;
                this.options = ifpoptions;
                this.visible = ifpvisible;
                this.enumSize = ifpsize;
            }
            #endregion
        }
        //End
        public class Option: BaseObject
        {            
            #region fields
            public int value;
            #endregion
            #region Constructor
            public Option(string ifpname, string ifpvalue, int entlineNumber)
            {
                this.lineNumber = entlineNumber;
                this.ObjectType = ObjectEnum.Option;
                this.name = ifpname;
                if (ifpvalue.Length >= 2)
                {
                    string tempx = ifpvalue.Substring(0, 2);
                    string tempx2 = ifpvalue.Substring(2, ifpvalue.Length - 2);
                    this.value = tempx != "0x" ? Convert.ToInt32(ifpvalue) : int.Parse(tempx2, System.Globalization.NumberStyles.HexNumber);
                }
                else
                {
                    this.value = Convert.ToInt32(ifpvalue);
                }
            }
            #endregion
        }
        public class Index
        { 
            #region fields
            public bool nulled = true;
            public int ReflexiveOffset;
            public int ReflexiveSize;
            public int ItemOffset;
            public string ItemType;
            public int lineNumber;
            public string reflexiveLayer;
            #endregion
            #region Constructor
            public Index(string ifpReflexiveOffset, string ifpReflexiveSize, string ifpItemOffset, string ENTItemType,string Layer,int linenumber)
            {
                this.reflexiveLayer = Layer;
                lineNumber = linenumber;
                if (ifpReflexiveOffset == null | ifpItemOffset == null | ifpReflexiveSize == null)
                    return;
                if (ifpReflexiveOffset.Contains("0x") | ifpItemOffset.Contains("0x") | ifpReflexiveSize.Contains("0x"))
                {
                    System.Windows.Forms.MessageBox.Show("Please use Decimal numbers and not hex numbers in your indicies. You used hex on line " + linenumber.ToString());
                    return;
                }
                this.nulled = false;
                try
                {
                    this.ReflexiveOffset = Convert.ToInt32(ifpReflexiveOffset);
                    this.ReflexiveSize = Convert.ToInt32(ifpReflexiveSize);
                    this.ItemOffset = Convert.ToInt32(ifpItemOffset);
                    this.ItemType = ENTItemType;
                }
                catch
                {
                    this.nulled = true;
                    return;
                }                
            }
            #endregion       
        }
        #endregion
    }
}
