using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using HaloMap;
using System.IO;
using ScriptTools.ConstructorFiles;

namespace ScriptTools
{
    public partial class Constructor : Form
    {
        public enum sortType { id, alpha, ret};
        public static sortType sort;

        public int scriptNumber;
        public int toDecompile;

        public Constructor()
        {
            InitializeComponent();
            toDecompile = -1;
        }

        private void Constructor_Load(object sender, EventArgs e)
        {
            sort = sortType.id;
            scriptTree.TreeViewNodeSorter = new treeSorter();
            sortTree();
            if (toDecompile >= 0)
                decompile(toDecompile);
            else
            {
                ConstructBox box = new ConstructBox(2, 1);
                box.loadValues(4);
                scriptPanel.Controls.Add(box);
            }
        }

        public class treeSorter : System.Collections.IComparer
        {
            public int Compare(object x, object y)
            {
                TreeNode xx = (TreeNode)x;
                TreeNode yy = (TreeNode)y;
                if (Constructor.sort == sortType.alpha)
                {
                    if (xx.Text.StartsWith("("))
                        return -1;
                    else if (yy.Text.StartsWith("("))
                        return 1;
                    return string.Compare(xx.Text, yy.Text);
                }
                else
                {
                    if (int.Parse(xx.Name) > int.Parse(yy.Name))
                        return 1;
                    else if (int.Parse(xx.Name) < int.Parse(yy.Name))
                        return -1;
                    else
                        return 0;
                }
            }
        }

        public void sortTree()
        {
            scriptTree.Nodes.Clear();
            scriptTree.TreeViewNodeSorter = null;
            scriptTree.Nodes.Add("-3", "(STATIC)");
            scriptTree.Nodes.Add("-2", "(GLOBAL)");
            scriptTree.Nodes.Add("-1", "(VALUE)");
            scriptTree.Nodes.Add("0", "begin");
            scriptTree.Nodes.Add("1", "begin_random");
            for (int i = 0; i < ScriptDatabaseInfo.Entries.Count; i++)
            {
                switch (sort)
                {
                    case sortType.id:
                    case sortType.alpha:
                        if (scriptTree.Nodes[ScriptDatabaseInfo.Entries[i].Identity.ToString()] == null)
                        {
                            scriptTree.Nodes.Add(ScriptDatabaseInfo.Entries[i].Identity.ToString(), ScriptDatabaseInfo.Entries[i].Name);
                        }
                        break;
                    case sortType.ret:
                        if (scriptTree.Nodes[ScriptDatabaseInfo.Entries[i].ReturnType.ToString()] == null)
                        {
                            scriptTree.Nodes.Add(ScriptDatabaseInfo.Entries[i].ReturnType.ToString(), ScriptValueTypes.Names[ScriptDatabaseInfo.Entries[i].ReturnType]);
                        }
                        if (scriptTree.Nodes[ScriptDatabaseInfo.Entries[i].ReturnType.ToString()].Nodes[ScriptDatabaseInfo.Entries[i].Identity.ToString()] == null)
                        {
                            scriptTree.Nodes[ScriptDatabaseInfo.Entries[i].ReturnType.ToString()].Nodes.Add(ScriptDatabaseInfo.Entries[i].Identity.ToString(), ScriptDatabaseInfo.Entries[i].Name);
                        }
                        break;
                }
            }
            scriptTree.TreeViewNodeSorter = new treeSorter();
            scriptTree.Sort();
        }

        public void decompile(int number)
        {
            scriptNumber = number;
            Script sinfo = ScriptListForm.scripts[number];
            scriptType.Text = ScriptTypes.Names[sinfo.Type];
            scriptName.Text = sinfo.Name;

            ScriptExpression chunk = new ScriptExpression(sinfo.ChunkNumber);
            ConstructBox box = new ConstructBox(2, 1);
            scriptPanel.Controls.Add(box);
            int line = 2;
            int indent = 1;
            List<ScriptExpression> syntax = new List<ScriptExpression>();
            List<ConstructBox> boxes = new List<ConstructBox>();
            List<int> parents = new List<int>();

            while (true)
            {
                //chunk immediately following syntax - skip it, go onto the next one
                if (chunk.ValueType == 2)
                {
                    if (chunk.Sibling != null) //this command has arguments
                    {
                        chunk = new ScriptExpression(chunk.Sibling.ChunkNumber);
                        box = box.child;
                    }
                    else //argument-less command like (players)
                    {
                        //send focus back to the (
                        chunk = syntax[parents[parents.Count - 1]];
                        box = boxes[parents[parents.Count - 1]];
                    }
                }

                if (box == null) //we don't have anywhere to put it!
                    break;

                //syntax chunk
                if (chunk.ExpressionType == 8 || chunk.ExpressionType == 10)
                {
                    if (chunk.Child.Followed) //second time - follow sibling
                    {
                        //we're no longer inside this command
                        parents.RemoveAt(parents.Count - 1);
                        if (chunk.Sibling != null)
                        {
                            chunk = new ScriptExpression(chunk.Sibling.ChunkNumber);
                            if (box.sibling != null)
                                box = box.sibling;
                            //we should have a sibling; make one
                            else
                            {
                                if (box.closeP != null)
                                    line = box.closeP.Top / 25 + 1;
                                else
                                    line = box.Top / 25 + 1;
                                indent = box.indent - 1;
                                if (chunk.ExpressionType == 10)
                                    indent++;
                                box.sibling = new ConstructBox(line, indent);
                                box = box.sibling;
                                //shift
                                foreach (Control temp in scriptPanel.Controls)
                                {
                                    if (temp.GetType().Name != "ConstructBox")
                                        continue;
                                    ConstructBox tempbox = (ConstructBox)temp;
                                    if (tempbox.line > box.line)
                                        tempbox.line++;
                                    else if (tempbox.closeP != null && tempbox.openP.Top < box.Top && tempbox.closeP.Top >= box.Top)
                                        tempbox.closeP.Top += 25;
                                }
                                scriptClose.Top += 25;
                                scriptPanel.Controls.Add(box);
                            }
                        }
                        else //go back up another level
                        {
                            if (parents.Count == 0) //no more levels to go up!
                                break;

                            chunk = syntax[parents[parents.Count - 1]];
                            box = boxes[parents[parents.Count - 1]];
                        }
                    }
                    else //first time - follow child
                    {
                        //load command, and build children
                        box.valueType = chunk.ValueType;
                        if (chunk.ExpressionType == 8)
                            box.makeCommand(chunk.Identity);
                        else
                        {
                            box.makeCommand(-3);
                            ScriptExpression tempchunk = new ScriptExpression(chunk.Child.ChunkNumber);
                            box.Text = tempchunk.String;
                        }
                        //load chunk and box into arrays
                        parents.Add(syntax.Count);
                        syntax.Add(chunk);
                        boxes.Add(box);
                        //set flag for when we finish the children
                        chunk.Child.Followed = true;
                        //find our new chunk
                        chunk = new ScriptExpression(chunk.Child.ChunkNumber);
                        //the box will be found in the value type 2 check
                    }
                }
                else if (chunk.ExpressionType == 9) //basic value
                {
                    //load values
                    box.loadValues(chunk.ValueType);
                    switch (chunk.ValueType)
                    {
                        case 5:
                            box.Text = chunk.Bool.ToString();
                            break;
                        case 6:
                            box.Text = chunk.Real.ToString();
                            break;
                        case 7:
                            box.Text = chunk.Int.ToString();
                            break;
                        case 8:
                            box.Text = chunk.Long.ToString();
                            break;
                        default:
                            box.Text = chunk.String;
                            break;
                    }

                    //add to arrays
                    syntax.Add(chunk);
                    boxes.Add(box);

                    //load next chunk/box
                    if (chunk.Sibling != null)
                    {
                        chunk = new ScriptExpression(chunk.Sibling.ChunkNumber);
                        if (box.sibling != null)
                            box = box.sibling;
                        else
                        {
                            if (box.closeP != null)
                                line = box.closeP.Top / 25;
                            else
                                line = box.Top / 25 + 1;
                            indent = box.indent;
                            box.sibling = new ConstructBox(line, indent);
                            box = box.sibling;
                            //shift
                            foreach (Control temp in scriptPanel.Controls)
                            {
                                if (temp.GetType().Name != "ConstructBox")
                                    continue;
                                ConstructBox tempbox = (ConstructBox)temp;
                                if (tempbox.line > box.line)
                                    tempbox.line++;
                                else if (tempbox.closeP != null && tempbox.openP.Top < box.Top && tempbox.closeP.Top >= box.Top)
                                    tempbox.closeP.Top += 25;
                            }
                            scriptClose.Top += 25;
                            scriptPanel.Controls.Add(box);
                        }
                    }
                    else
                    {
                        if (parents.Count == 0)
                            break;

                        chunk = syntax[parents[parents.Count - 1]];
                        box = boxes[parents[parents.Count - 1]];
                    }
                }
                else if (chunk.ExpressionType == 13)
                {
                    //load values
                    if ((chunk.ValueType == 7 || chunk.ValueType == 19 || chunk.ValueType == 31) && chunk.String.StartsWith("ai_"))
                    {
                        box.loadValues(chunk.ValueType);
                        box.Text = chunk.String;
                    }
                    else
                    {
                        box.loadValues(-2);
                        box.Text = chunk.String;
                    }

                    //add to arrays
                    syntax.Add(chunk);
                    boxes.Add(box);

                    //load next chunk/box
                    if (chunk.Sibling != null)
                    {
                        chunk = new ScriptExpression(chunk.Sibling.ChunkNumber);
                        if (box.sibling != null)
                            box = box.sibling;
                        else
                        {
                            if (box.closeP != null)
                                line = box.closeP.Top / 25;
                            else
                                line = box.Top / 25 + 1;
                            indent = box.indent;
                            box.sibling = new ConstructBox(line, indent);
                            box = box.sibling;
                            //shift
                            foreach (Control temp in scriptPanel.Controls)
                            {
                                if (temp.GetType().Name != "ConstructBox")
                                    continue;
                                ConstructBox tempbox = (ConstructBox)temp;
                                if (tempbox.line > box.line)
                                    tempbox.line++;
                                else if (tempbox.closeP != null && tempbox.openP.Top < box.Top && tempbox.closeP.Top >= box.Top)
                                    tempbox.closeP.Top += 25;
                            }
                            scriptClose.Top += 25;
                            scriptPanel.Controls.Add(box);
                        }
                    }
                    else
                    {
                        if (parents.Count == 0)
                            break;

                        chunk = syntax[parents[parents.Count - 1]];
                        box = boxes[parents[parents.Count - 1]];
                    }
                }
            }
        }

        public scriptInfo compile(int startChunk)
        {
            ConstructBox start = null;
            scriptInfo sinfo = new scriptInfo();
            List<ScriptExpression> syntax = new List<ScriptExpression>();
            List<ConstructBox> boxes = new List<ConstructBox>();
            List<int> parents = new List<int>();

            sinfo.type = type.Script;
            string scriptname = scriptName.Text;
            for (int i = 0; i < scriptname.Length && i < 32; i++)
            {
                sinfo.header[i] = (byte)scriptname[i];
            }
            if (scriptType.SelectedIndex < 0)
            {
                MessageBox.Show("Invalid Script Type!");
                return null;
            }
            sinfo.header[32] = (byte)scriptType.SelectedIndex;

            foreach (Control ctrl in scriptPanel.Controls)
            {
                if (ctrl.GetType().Name == "ConstructBox")
                {
                    start = (ConstructBox)ctrl;
                    break;
                }
            }

            ScriptExpression chunk;
            ConstructBox box = start;
            while (true)
            {
                chunk = new ScriptExpression();
                chunk.ChunkNumber = startChunk + syntax.Count;
                chunk.ExpressionIndex = 58227 + (uint)chunk.ChunkNumber;
                if (chunk.ExpressionIndex > 65536)
                    chunk.ExpressionIndex -= 65536;

                dropdownValue value = null;
                if (box.SelectedIndex >= 0)
                    value = (dropdownValue)box.Items[box.SelectedIndex];

                switch (box.identity)
                {
                    case -3: //static
                        chunk.ExpressionType = 10;
                        int identity = BitConverter.ToInt16(value.hex, 0);
                        chunk.Identity = identity;
                        chunk.Child = new ScriptPointer(startChunk + syntax.Count + 1);
                        chunk.ValueType = box.valueType;
                        syntax.Add(chunk);
                        chunk = new ScriptExpression();
                        chunk.ExpressionType = 9;
                        chunk.ValueType = 2;
                        chunk.Identity = identity;
                        chunk.String = box.Text;
                        chunk.StringOffset = ScriptEditorForm.getStringPointer(box.Text);
                        break;
                    case -2: //global
                        chunk.ExpressionType = 13;
                        chunk.ValueType = box.valueType;
                        chunk.Identity = box.valueType;
                        chunk.Hex = value.hex;
                        chunk.String = box.Text;
                        chunk.StringOffset = ScriptEditorForm.getStringPointer(box.Text);
                        break;
                    case -1: //value
                        chunk.ExpressionType = 9;
                        chunk.ValueType = box.valueType;
                        chunk.Identity = box.valueType;
                        if (box.SelectedIndex >= 0)
                        {
                            chunk.Hex = value.hex;
                            chunk.String = box.Text;
                            chunk.StringOffset = ScriptEditorForm.getStringPointer(box.Text);
                            if (box.Text.StartsWith("ai_"))
                                chunk.ExpressionType = 13; //command script constants
                        }
                        else
                        {
                            try
                            {
                                switch (box.valueType)
                                {
                                    case 4: //so anything can go in a begin statement
                                        short blah;
                                        if (bool.TryParse(box.Text, out chunk.Bool))
                                        {
                                            byte boolean = BitConverter.GetBytes(chunk.Bool)[0];
                                            chunk.Hex = new byte[] { boolean, 255, 255, 255 };
                                            box.valueType = 5;
                                            chunk.ValueType = 5;
                                        }
                                        else if (short.TryParse(box.Text, out blah))
                                        {
                                            chunk.Int = (int)blah;
                                            chunk.Hex = BitConverter.GetBytes(blah);
                                            chunk.Hex[2] = 255;
                                            chunk.Hex[3] = 255;
                                            box.valueType = 7;
                                            chunk.ValueType = 7;
                                        }
                                        else if (int.TryParse(box.Text, out chunk.Int))
                                        {
                                            chunk.Long = (long)chunk.Int;
                                            chunk.Hex = BitConverter.GetBytes(chunk.Int);
                                            chunk.ValueType = 8;
                                            box.valueType = 8;
                                        }
                                        else if (float.TryParse(box.Text, out chunk.Real))
                                        {
                                            chunk.Hex = BitConverter.GetBytes(chunk.Real);
                                            box.valueType = 6;
                                            chunk.ValueType = 6;
                                        }
                                        else
                                        {
                                            DialogResult test = MessageBox.Show("Unknown void entry: " + box.Text + "\nContinue compiling?", "Unknown", MessageBoxButtons.YesNo);
                                            if (test == DialogResult.No)
                                                return null;
                                        }
                                        break;
                                    case 5:
                                        chunk.Bool = bool.Parse(box.Text);
                                        chunk.Hex = new byte[] { BitConverter.GetBytes(chunk.Bool)[0], 255, 255, 255 };
                                        break;
                                    case 6:
                                        chunk.Real = float.Parse(box.Text);
                                        chunk.Hex = BitConverter.GetBytes(chunk.Real);
                                        break;
                                    case 7:
                                        chunk.Int = int.Parse(box.Text);
                                        chunk.Hex = BitConverter.GetBytes(chunk.Int);
                                        chunk.Hex[2] = 255;
                                        chunk.Hex[3] = 255;
                                        break;
                                    case 8:
                                        chunk.Long = long.Parse(box.Text);
                                        chunk.Hex = BitConverter.GetBytes((int)chunk.Long);
                                        break;
                                    case 9:
                                        chunk.String = box.Text;
                                        chunk.StringOffset = ScriptEditorForm.getStringPointer(box.Text);
                                        break;
                                    default:
                                        MessageBox.Show("Invalid entry: " + box.Text + "\nPlease select from the dropdown");
                                        return null;
                                }
                            }
                            catch
                            {
                                MessageBox.Show("Invalid entry: " + box.Text + "\nUnable to parse number");
                                return null;
                            }
                        }
                        break;
                    default: //script
                        chunk.ExpressionType = 8;
                        chunk.ValueType = box.valueType;
                        chunk.Identity = box.identity;
                        chunk.Child = new ScriptPointer(startChunk + syntax.Count + 1);
                        syntax.Add(chunk);
                        chunk = new ScriptExpression();
                        chunk.ChunkNumber = syntax[syntax.Count - 1].Child.ChunkNumber;
                        chunk.ExpressionIndex = syntax[syntax.Count - 1].Child.ExpressionIndex;
                        chunk.ExpressionType = 9;
                        chunk.ValueType = 2;
                        chunk.Identity = box.identity;
                        if (box.identity < 2)
                        {
                            chunk.String = box.identity == 0 ? "begin" : "begin_random";
                            chunk.StringOffset = ScriptEditorForm.getStringPointer(chunk.String);
                        }
                        else
                        {
                            chunk.String = ScriptDatabaseInfo.Entries[box.scriptNum].Name;
                            chunk.StringOffset = ScriptEditorForm.getStringPointer(chunk.String);
                        }
                        break;
                }
                boxes.Add(box);
                if (box.child != null)
                {
                    box = box.child;
                    chunk.Sibling = new ScriptPointer(chunk.ChunkNumber + 1);
                }
                else if (box.sibling != null)
                {
                    box.followed = true;
                    if (box.identity >= 0) //it may be the box's sibling, but does it belong to the ( or the name?
                        syntax[syntax.Count - 1].Sibling = new ScriptPointer(chunk.ChunkNumber + 1);
                    else
                        chunk.Sibling = new ScriptPointer(chunk.ChunkNumber + 1);
                    box = box.sibling;
                }
                else
                {
                    if (boxes.Count > 1)
                    {
                        int good = -1;
                        int count = box.identity >= 0 ? 3 : 2; //fixed for commands like (players)
                        for (int i = 2; i <= boxes.Count; i++)
                        {
                            if (boxes[boxes.Count - i].sibling != null && !boxes[boxes.Count - i].followed)
                            {
                                good = i;
                                boxes[boxes.Count - i].followed = true;
                                box = boxes[boxes.Count - i].sibling;
                                syntax[syntax.Count - count].Sibling = new ScriptPointer(startChunk + syntax.Count + 1);
                                break;
                            }
                            count += boxes[boxes.Count - i].identity >= 0 ? 2 : 1;
                        }
                        if (good < 0)
                        {
                            syntax.Add(chunk);
                            break; //done
                        }
                    }
                    else
                    {
                        syntax.Add(chunk);
                        break;
                    }
                }
                syntax.Add(chunk);
            }

            for (int i = 0; i < syntax.Count; i++)
            {
                chunk = syntax[i];
                byte[] hex = new byte[20];

                byte[] expindex = BitConverter.GetBytes(chunk.ExpressionIndex);
                if (i == 0)
                {
                    byte[] chunknum = BitConverter.GetBytes(chunk.ChunkNumber);
                    sinfo.header[36] = chunknum[0];
                    sinfo.header[37] = chunknum[1];
                    sinfo.header[38] = expindex[0];
                    sinfo.header[39] = expindex[1];
                }
                hex[0] = expindex[0];
                hex[1] = expindex[1];

                byte[] identity = BitConverter.GetBytes(chunk.Identity);
                hex[2] = identity[0];
                hex[3] = identity[1];

                byte[] valueType = BitConverter.GetBytes(chunk.ValueType);
                hex[4] = valueType[0];
                hex[5] = valueType[1];

                byte[] expType = BitConverter.GetBytes(chunk.ExpressionType);
                hex[6] = expType[0];
                hex[7] = expType[1];

                if (chunk.Sibling == null)
                {
                    hex[8] = 255;
                    hex[9] = 255;
                    hex[10] = 255;
                    hex[11] = 255;
                }
                else
                {
                    byte[] schunkn = BitConverter.GetBytes(chunk.Sibling.ChunkNumber);
                    hex[8] = schunkn[0];
                    hex[9] = schunkn[1];

                    byte[] schunke = BitConverter.GetBytes(chunk.Sibling.ExpressionIndex);
                    hex[10] = schunke[0];
                    hex[11] = schunke[1];
                }

                byte[] stringPointer = BitConverter.GetBytes(chunk.StringOffset);
                hex[12] = stringPointer[0];
                hex[13] = stringPointer[1];
                hex[14] = stringPointer[2];
                hex[15] = stringPointer[3];

                if (chunk.Child != null) //never converts a child pointer to hex
                {
                    byte[] cchunkn = BitConverter.GetBytes(chunk.Child.ChunkNumber);
                    hex[16] = cchunkn[0];
                    hex[17] = cchunkn[1];

                    byte[] cchunke = BitConverter.GetBytes(chunk.Child.ExpressionIndex);
                    hex[18] = cchunke[0];
                    hex[19] = cchunke[1];
                }
                else
                {
                    hex[16] = chunk.Hex[0];
                    hex[17] = chunk.Hex[1];
                    hex[18] = chunk.Hex[2];
                    hex[19] = chunk.Hex[3];
                }

                for (int j = 0; j < 20; j++)
                {
                    sinfo.syntax.Add(hex[j]);
                }
            }
            return sinfo;
        }

        private void identityToolStripMenuItem_Click(object sender, EventArgs e)
        {
            identityToolStripMenuItem.Checked = true;
            alphabeticalToolStripMenuItem.Checked = false;
            returnTypeToolStripMenuItem.Checked = false;
            sort = sortType.id;
            sortTree();
        }

        private void alphabeticalToolStripMenuItem_Click(object sender, EventArgs e)
        {
            identityToolStripMenuItem.Checked = false;
            alphabeticalToolStripMenuItem.Checked = true;
            returnTypeToolStripMenuItem.Checked = false;
            sort = sortType.alpha;
            sortTree();
        }

        private void returnTypeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            identityToolStripMenuItem.Checked = false;
            alphabeticalToolStripMenuItem.Checked = false;
            returnTypeToolStripMenuItem.Checked = true;
            sort = sortType.ret;
            sortTree();
        }

        private void scriptTree_ItemDrag(object sender, ItemDragEventArgs e)
        {
            if (e.Button != MouseButtons.Left)
                return;

            scriptPanel.DoDragDrop(e.Item, DragDropEffects.Copy);
        }

        public void constructBox_DragEnter(object sender, DragEventArgs e)
        {
            e.Effect = DragDropEffects.Copy;
        }

        public void constructBox_DragDrop(object sender, DragEventArgs e)
        {
            TreeNode node = (TreeNode)e.Data.GetData("System.Windows.Forms.TreeNode");
            int identity = int.Parse(node.Name);

            ConstructBox box = (ConstructBox)sender;
            box.makeCommand(identity);
        }

        private void compileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int[] size = ScriptEditorForm.getsize(scriptNumber);
            scriptInfo info = compile(size[0]);
            if (info == null)
            {
                MessageBox.Show("Compile failed!", "Error!");
                return;
            }
            ScriptEditorForm.writeScript(scriptNumber, size[0], size[1], size[2], info);
            MessageBox.Show("Compile successful!");
        }

        private void scriptTree_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            if (e.Button != MouseButtons.Right || int.Parse(e.Node.Name) < 2 || ScriptValueTypes.GetIndex(e.Node.Text) > 0)
                return;

            ScriptInfoBox box = new ScriptInfoBox(int.Parse(e.Node.Name));
            box.ShowDialog();
        }
    }

    public class ConstructBox : System.Windows.Forms.ComboBox
    {
        public int identity;
        public int scriptNum;
        public int valueType;
        public int paramCount;

        public ConstructBox child;
        public ConstructBox sibling;
        public Label openP;
        public Label closeP;
        private int indentation;
        private int lineNum;

        public bool followed; //>.< for compiling...

        public int indent
        {
            get { return indentation; }
            set
            {
                indentation = value;
                this.Left = value * 30 + 30;
                if (openP != null && closeP != null)
                {
                    openP.Left = value * 30;
                    closeP.Left = value * 30;
                }
            }
        }

        public int line
        {
            get { return lineNum; }
            set
            {
                int dif = value - lineNum;
                lineNum = value;
                this.Top = value * 25;
                if (openP != null && closeP != null)
                {
                    openP.Top += dif * 25;
                    closeP.Top += dif * 25;
                }
            }
        }

        public ConstructBox(int line, int indent)
        {
            identity = -1;
            lineNum = line;
            this.Top = line * 25;
            indentation = indent;
            this.Left = (indent + 1) * 30;
            this.Width = 200;
            this.DisplayMember = "name";
            AllowDrop = true;
            this.DragEnter += new DragEventHandler(constructBox_DragEnter);
            this.DragDrop += new DragEventHandler(constructBox_DragDrop);
            this.MouseDown += new MouseEventHandler(ConstructBox_MouseClick);
        }

        public void ConstructBox_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button != MouseButtons.Right)
                return;

            ConstructBox box = (ConstructBox)sender;
            ConstructBoxInfo info = new ConstructBoxInfo();
            info.LoadValue(box.valueType);
            info.btnOK.Click += new EventHandler(InfoOK_Click);
            info.ShowDialog();
        }

        public void InfoOK_Click(object sender, EventArgs e)
        {
            ConstructBoxInfo info = (ConstructBoxInfo)((Button)sender).Parent;
            int type = ScriptTools.ScriptValueTypes.GetIndex(info.cboTypes.Text);
            this.valueType = type;
            if (this.identity == -1)
                this.loadValues(type); //reload value choices
            else if (this.identity < -1)
                this.loadValues(this.identity);
            else
            {
                string name = "";
                if (this.identity < 2)
                    name = identity == 0 ? "begin" : "begin_random";
                else
                    name = ScriptDatabaseInfo.Entries[scriptNum].Name;
                this.Items.Clear();
                name += " - " + ScriptTools.ScriptValueTypes.Names[type];
                this.Items.Add(new dropdownValue(name));
                this.Text = name;
            }
            info.Close();
        }

        public void constructBox_DragEnter(object sender, DragEventArgs e)
        {
            e.Effect = DragDropEffects.Copy;
        }

        public void constructBox_DragDrop(object sender, DragEventArgs e)
        {
            TreeNode node = (TreeNode)e.Data.GetData("System.Windows.Forms.TreeNode");
            int identity = int.Parse(node.Name);

            ConstructBox box = (ConstructBox)sender;
            box.makeCommand(identity);
        }

        public void loadValues(int type)
        {
            scriptNum = -1;

            this.Items.Clear();
            this.DropDownStyle = ComboBoxStyle.DropDown;
            Map map = Maps.map[0];
            TagMeta meta = new TagMeta();
            BinaryReader br = map.BR;
            if (type >= 4)
            {
                this.Text = "<" + ScriptTools.ScriptValueTypes.Names[type] + ">";
            }
            if (type == -2)
            {
                for (int i = 0; i < ScriptListForm.globals.Length; i++)
                {
                    if (this.valueType == 0 || this.valueType == ScriptListForm.globals[i].Type)
                        this.Items.Add(new dropdownValue(ScriptListForm.globals[i].Name, i));
                }
            }
            else if (type == -3)
            {
                for (int i = 0; i < ScriptListForm.scripts.Length; i++)
                {
                    if (ScriptListForm.scripts[i].Type == 3 && (this.valueType == 0 || ScriptListForm.scripts[i].ReturnType == this.valueType))
                        this.Items.Add(new dropdownValue(ScriptListForm.scripts[i].Name, i));
                }
            }
            else if (type > 4 && type < 9) //numbers
            {
                switch (type)
                {
                    case 5:
                        this.Items.Add(new dropdownValue("True", 1, 255, 255, 255));
                        this.Items.Add(new dropdownValue("False", 0, 255, 255, 255));
                        this.DropDownStyle = ComboBoxStyle.DropDownList;
                        break;
                    case 7:
                        this.Items.Add(new dropdownValue("ai_combat_status_alert", 39, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_combat_status_active", 40, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_combat_status_uninspected", 41, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_combat_status_definite", 42, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_combat_status_certain", 43, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_combat_status_visible", 44, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_combat_status_clear_los", 45, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_combat_status_dangerous", 46, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_combat_status_idle", 47, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_movement_patrol", 47, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_movement_combat", 49, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_movement_flee", 50, 131, 255, 255));
                        break;
                    default: break;
                }
            }
            else if (type == 16 || type == 13) //stringids
            {
                int off = 0;
                int size = 0;
                meta.ReadMetaFromMap(3, 0, false);
                if (type == 13)
                {
                    off = 264;
                    size = 68;
                }
                else
                {
                    off = 496;
                    size = 36;
                }
                br.BaseStream.Position = meta.offset + off;
                int count = br.ReadInt32();
                br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                for (int i = 0; i < count; i++)
                {
                    byte[] blah = new byte[4];
                    for (int y = 0; y < 2; y++)
                    {
                        blah[y] = br.ReadByte();
                    }
                    int num = BitConverter.ToInt32(blah, 0);
                    this.Items.Add(new dropdownValue(map.Strings.Name[num], i));
                    br.BaseStream.Position += size - 2;
                }
            }
            else if ((type > 13 && type < 19) || (type > 21 && type < 25) || type == 27 || type == 31 || type > 49)
            {
                int tagnum = 3;
                int off = 0;
                int off2 = 0;
                int size = 0;
                switch (type)
                {
                    case 14:
                        off = 480;
                        off2 = 4;
                        size = 56;
                        break;
                    case 15:
                        off = 488;
                        off2 = 4;
                        size = 64;
                        break;
                    case 18:
                        off = 160;
                        size = 40;
                        break;
                    case 22:
                        off = 84;
                        size = 32;
                        for (int x = 0; x < map.FileNames.Name.Length; x++)
                        {
                            if (map.FileNames.Name[x] == "ai\\all")
                            {
                                tagnum = x;
                                break;
                            }
                        }
                        break;
                    case 23:
                        off = 576;
                        size = 124;
                        break;
                    case 24:
                        off = 248;
                        size = 68;
                        break;
                    case 27:
                        off = 292;
                        size = 104;
                        for (int x = 0; x < map.FileNames.Name.Length; x++)
                        {
                            if (map.MetaInfo.TagType[x] == "hudg")
                            {
                                tagnum = x;
                                break;
                            }
                        }
                        break;
                    case 31:
                        this.Items.Add(new dropdownValue("ai_current_actor", 14, 131, 255, 255));
                        off = 72;
                        size = 36;
                        break;
                    case 50:
                    case 51:
                    case 52:
                    case 53:
                    case 54:
                    case 55:
                    case 56:
                        off = 72;
                        size = 36;
                        break;
                }
                meta.ReadMetaFromMap(tagnum, 0, false);
                br.BaseStream.Position = meta.offset + off;
                int count = br.ReadInt32();
                try { br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic; }
                catch { MessageBox.Show("There are no " + ScriptTools.ScriptValueTypes.Names[type] + " in this map."); }
                for (int i = 0; i < count; i++)
                {
                    br.BaseStream.Position += off2;
                    string blah = "";
                    for (int j = 0; j < 32; j++)
                    {
                        char moo = br.ReadChar();
                        if (moo == 0)
                        {
                            br.BaseStream.Position += 31 - j;
                            break;
                        }
                        blah += moo.ToString();
                    }
                    if (type < 32)
                    {
                        this.Items.Add(new dropdownValue(blah, i));
                    }
                    else
                    {
                        this.Items.Add(new dropdownValue(blah, (short)i));
                    }
                    br.BaseStream.Position += size - 32 - off2;
                }
            }
            else if (type > 31 && type < 42)
            {
                for (int i = 0; i < map.MetaInfo.Ident.Length; i++)
                {
                    if (type == 32 && (map.MetaInfo.TagType[i] == "snd!" || map.MetaInfo.TagType[i] == "ssce"))
                        this.Items.Add(new dropdownValue(map.FileNames.Name[i], i));
                    else if (type == 33 && map.MetaInfo.TagType[i] == "effe")
                        this.Items.Add(new dropdownValue(map.FileNames.Name[i], i));
                    else if (type == 34 && map.MetaInfo.TagType[i] == "jpt!")
                        this.Items.Add(new dropdownValue(map.FileNames.Name[i], i));
                    else if (type == 35 && map.MetaInfo.TagType[i] == "lsnd")
                        this.Items.Add(new dropdownValue(map.FileNames.Name[i], i));
                    else if (type == 36 && map.MetaInfo.TagType[i] == "jmad")
                        this.Items.Add(new dropdownValue(map.FileNames.Name[i], i));
                    else if (type == 38 && map.MetaInfo.TagType[i] == "bipd")
                        this.Items.Add(new dropdownValue(map.FileNames.Name[i], i));
                    else if (type == 38 && map.MetaInfo.TagType[i] == "weap")
                        this.Items.Add(new dropdownValue(map.FileNames.Name[i] + ".weapon", i));
                    else if (type == 41 && (map.MetaInfo.TagType[i] == "mode" || map.MetaInfo.TagType[i] == "sbsp"))
                        this.Items.Add(new dropdownValue(map.FileNames.Name[i], i));
                }
            }
            else
            {
                switch (type)
                {
                    case 9:
                        for (int i = 0; i < ScriptListForm.stringTable.Length; i++)
                        {
                            this.Items.Add(new dropdownValue(ScriptListForm.stringTable[i]));
                        }
                        break;
                    case 10:
                    case 21:
                        for (int i = 0; i < ScriptListForm.scripts.Length; i++)
                        {
                            if ((type == 21 && ScriptListForm.scripts[i].Type == 5) || (type == 10 && (ScriptListForm.scripts[i].Type == 1 || ScriptListForm.scripts[i].Type == 3)))
                                this.Items.Add(new dropdownValue(ScriptListForm.scripts[i].Name, i));
                        }
                        break;
                    case 11:
                        for (int i = 0; i < map.Strings.Name.Length; i++)
                        {
                            dropdownValue value = new dropdownValue(map.Strings.Name[i]);
                            BitConverter.GetBytes((short)i).CopyTo(value.hex, 0);
                            value.hex[3] = (byte)map.Strings.Length[i];
                            this.Items.Add(value);
                        }
                        break;
                    case 19:
                        this.Items.Add(new dropdownValue("ai_current_actor", 14, 131, 255, 255));
                        this.Items.Add(new dropdownValue("ai_current_squad", 13, 131, 255, 255));
                        meta.ReadMetaFromMap(3, 0, false);
                        br.BaseStream.Position = meta.offset + 344; //platoons first
                        int count = br.ReadInt32();
                        try { br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic; }
                        catch { }
                        for (int i = 0; i < count; i++)
                        {
                            string blah = "";
                            for (int j = 0; j < 32; j++)
                            {
                                char moo = br.ReadChar();
                                if (moo == 0)
                                {
                                    br.BaseStream.Position += 35 - j;
                                    break;
                                }
                                blah += moo.ToString();
                            }
                            this.Items.Add(new dropdownValue(blah, (byte)i, 0, 0, 64));
                        }
                        br.BaseStream.Position = meta.offset + 352;
                        count = br.ReadInt32();
                        long outoffset = br.ReadUInt32() - map.SecondaryMagic;
                        try { br.BaseStream.Position = outoffset; }
                        catch { MessageBox.Show("There are no ai squads in this map."); }
                        for (int i = 0; i < count; i++)
                        {
                            br.BaseStream.Position = outoffset + i * 116;
                            string blah = "";
                            for (int j = 0; j < 32; j++)
                            {
                                char moo = br.ReadChar();
                                if (moo == 0)
                                {
                                    br.BaseStream.Position += 31 - j;
                                    break;
                                }
                                blah += moo.ToString();
                            }
                            this.Items.Add(new dropdownValue(blah, (byte)i, 0, 0, 0));
                            br.BaseStream.Position += 40;
                            int innercount = br.ReadInt32();
                            if (innercount > 0)
                                br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                            for (int x = 0; x < innercount; x++)
                            {
                                byte[] bleh = new byte[4];
                                for (int y = 0; y < 2; y++)
                                {
                                    bleh[y] = br.ReadByte();
                                }
                                int num = BitConverter.ToInt32(bleh, 0);
                                if (num <= map.Strings.Name.Length)
                                    this.Items.Add(new dropdownValue(blah + "/" + map.Strings.Name[num], (byte)x, 0, (byte)i, 192));
                                br.BaseStream.Position += 98;
                            }
                        }
                        break;
                    case 26:
                        meta.ReadMetaFromMap(3, 0, false);
                        br.BaseStream.Position = meta.offset + 528;
                        count = br.ReadInt32();
                        try { br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic; }
                        catch { MessageBox.Show("There are no " + ScriptTools.ScriptValueTypes.Names[type] + " in this map."); }
                        for (int i = 0; i < count; i++)
                        {
                            br.BaseStream.Position += 20;
                            int ident = br.ReadInt32();
                            for (int x = 0; x < map.MetaInfo.Ident.Length; x++)
                            {
                                if (map.MetaInfo.Ident[x] == ident)
                                {
                                    string filename = map.FileNames.Name[x];
                                    string[] temp = filename.Split(new char[] { '\\' });
                                    filename = temp[temp.Length - 1];
                                    this.Items.Add(new dropdownValue(filename, ident));
                                    break;
                                }
                            }
                            br.BaseStream.Position += 44;
                        }
                        break;
                    case 28:
                        meta.ReadMetaFromMap(3, 0, false);
                        br.BaseStream.Position = meta.offset + 476;
                        br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                        count = br.ReadInt32();
                        outoffset = br.ReadUInt32() - map.SecondaryMagic;
                        try { br.BaseStream.Position = outoffset; }
                        catch { MessageBox.Show("There are no positions in this map."); }
                        for (int i = 0; i < count; i++)
                        {
                            br.BaseStream.Position = outoffset + i * 48;
                            string blah = "";
                            for (int j = 0; j < 32; j++)
                            {
                                char moo = br.ReadChar();
                                if (moo == 0)
                                {
                                    br.BaseStream.Position += 31 - j;
                                    break;
                                }
                                blah += moo.ToString();
                            }
                            this.Items.Add(new dropdownValue(blah, (byte)i, 0, 0, 192));
                            int innercount = br.ReadInt32();
                            if (innercount > 0)
                                br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                            for (int x = 0; x < innercount; x++)
                            {
                                string bleh = "";
                                for (int j = 0; j < 32; j++)
                                {
                                    char moo = br.ReadChar();
                                    if (moo == 0)
                                    {
                                        br.BaseStream.Position += 31 - j;
                                        break;
                                    }
                                    bleh += moo.ToString();
                                }
                                this.Items.Add(new dropdownValue(blah + "/" + bleh, (byte)x, 0, (byte)i, 0));
                                br.BaseStream.Position += 28;
                            }
                        }
                        break;
                    case 44:
                        this.Items.Add(new dropdownValue("easy", (short)0));
                        this.Items.Add(new dropdownValue("normal", (short)1));
                        this.Items.Add(new dropdownValue("heroic", (short)2));
                        this.Items.Add(new dropdownValue("legendary", (short)3));
                        this.DropDownStyle = ComboBoxStyle.DropDownList;
                        break;
                    case 45:
                        this.Items.Add(new dropdownValue("player", (short)1));
                        this.Items.Add(new dropdownValue("human", (short)2));
                        this.Items.Add(new dropdownValue("covenant", (short)3));
                        this.Items.Add(new dropdownValue("flood", (short)4));
                        this.Items.Add(new dropdownValue("sentinel", (short)5));
                        this.Items.Add(new dropdownValue("heretic", (short)6));
                        this.Items.Add(new dropdownValue("prophet", (short)7));
                        this.DropDownStyle = ComboBoxStyle.DropDownList;
                        break;
                    case 48:
                        this.Items.Add(new dropdownValue("destroyed", (short)4));
                        this.DropDownStyle = ComboBoxStyle.DropDownList;
                        break;

                }
            }
            valueType = type;
            for (int i = 0; i < this.Items.Count; i++)
            {
                if (this.Items[i].ToString().Length * 4 > this.Width)
                    this.Width = this.Items[i].ToString().Length * 4;
            }
        }

        public void makeCommand(int identity)
        {
            if (this.identity == identity && identity < 0)
                return;

            this.Width = 200;

            Panel panel = (Panel)this.Parent;
            int horiz = panel.HorizontalScroll.Value;
            int vert = panel.VerticalScroll.Value;
            panel.HorizontalScroll.Value = 0;
            panel.VerticalScroll.Value = 0;

            if (this.identity < 0 && identity >= 0) //making a value into a command
            {
                #region makecommand
                this.identity = identity;

                paramCount = 0;
                string name = "";
                int num = scriptNum; //default: -1
                if (identity < 2) //bleh >.< don't want to do this
                {
                    paramCount = 1;
                    //valueType = 0;
                    scriptNum = -1;
                    name = identity == 0 ? "begin" : "begin_random";
                    name += " - " + ScriptTools.ScriptValueTypes.Names[valueType];
                }
                else
                {
                    //starts at 0 if it doesn't have a number; if we're changing
                    //parameters/return then it will already be set
                    if (ScriptDatabaseInfo.Entries[num + 1].Identity != identity)
                        num = -1;
                    for (int i = scriptNum + 1; i < ScriptDatabaseInfo.Entries.Count; i++)
                    {
                        if (ScriptDatabaseInfo.Entries[i].Identity == identity)
                        {
                            num = i;
                            break;
                        }
                    }
                    scriptNum = num;
                    if (((Constructor)this.Parent.Parent.Parent.Parent).toDecompile < 0 && valueType > 4 && valueType != ScriptDatabaseInfo.Entries[num].ReturnType)
                    {
                        int num2 = num;
                        for (int i = num; i < ScriptDatabaseInfo.Entries.Count; i++)
                        {
                            if (ScriptDatabaseInfo.Entries[i].Identity != identity)
                                break;
                            if (ScriptDatabaseInfo.Entries[i].ReturnType == valueType)
                            {
                                num2 = i;
                                break;
                            }
                        }
                        if (num2 == num)
                        {
                            DialogResult result = MessageBox.Show("Return type does not match value.\nProceed anyway?", "", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                            if (result == DialogResult.No)
                            {
                                panel.AutoScrollPosition = new Point(horiz, vert);
                                return;
                            }
                        }
                        num = num2;
                        scriptNum = num;
                    }
                    if (valueType < 5)
                        valueType = ScriptDatabaseInfo.Entries[num].ReturnType;
                    paramCount = ScriptDatabaseInfo.Entries[num].Parameters.Count;
                    name = ScriptDatabaseInfo.Entries[num].Name + " - " + ScriptTools.ScriptValueTypes.Names[valueType];
                }

                openP = new Label();
                openP.Text = "(";
                this.Parent.Controls.Add(openP);
                closeP = new Label();
                closeP.Text = ")";
                this.Parent.Controls.Add(closeP);
                openP.Top = this.Top;
                closeP.Top = this.Top + 25;
                indent++;

                ConstructBox box = null;
                closeP.Top += 25 * paramCount;

                Constructor main = (Constructor)this.Parent.Parent.Parent.Parent;
                foreach (Control tempbox in panel.Controls)
                {
                    if (tempbox.GetType().Name != "ConstructBox")
                        continue;
                    ConstructBox thisbox = (ConstructBox)tempbox;
                    if (thisbox.line > this.line)
                        thisbox.line += paramCount + 1;
                    else if (thisbox.closeP != null && thisbox != this && thisbox.closeP.Top > this.Top)
                        thisbox.closeP.Top += (paramCount + 1) * 25;
                }

                for (int i = 0; i < paramCount; i++)
                {
                    ConstructBox oldbox = box;
                    box = new ConstructBox(line + i + 1, indent + 1);
                    if (identity < 2)
                        box.loadValues(4);
                    else
                        box.loadValues(ScriptDatabaseInfo.Entries[num].Parameters[i]);
                    if (oldbox == null)
                        this.child = box;
                    else
                        oldbox.sibling = box;
                    this.Parent.Controls.Add(box);
                }

                main.scriptClose.Top += 25 * (paramCount + 1);

                this.Items.Clear();
                this.Items.Add(new dropdownValue(name));
                this.Text = name;
                #endregion
            }

            else if (identity < 0 && this.identity >= 0) //making a command into a value
            {
                #region removecommand
                this.identity = identity;
                int num = scriptNum;
                scriptNum = -1;

                this.Parent.Controls.Remove(openP);
                this.Parent.Controls.Remove(closeP);
                openP = null;
                closeP = null;
                indent--;

                ConstructBox box = null;
                ConstructBox parent = null;
                List<ConstructBox> boxes = new List<ConstructBox>();
                while(true)
                {
                    if (box == null)
                        box = this.child;
                    else
                        box = box.sibling;
                    if (box == null)
                        break;

                    boxes.Add(box);
                }
                int count = boxes.Count;
                for (int i = boxes.Count; i > 0; i--)
                {
                    boxes[i - 1].makeCommand(-1); //removes their children too
                    this.Parent.Controls.Remove(boxes[i - 1]);
                    boxes[i - 1].sibling = null;
                    boxes[i - 1] = null;
                }
                this.child = null;

                Constructor main = (Constructor)this.Parent.Parent.Parent.Parent;
                foreach (Control tempbox in panel.Controls)
                {
                    if (tempbox.GetType().Name != "ConstructBox")
                        continue;
                    ConstructBox thisbox = (ConstructBox)tempbox;
                    if (thisbox.line > this.line)
                        thisbox.line -= count + 1;
                    else if (thisbox.closeP != null && thisbox != this && thisbox.closeP.Top > this.Top)
                    {
                        thisbox.closeP.Top -= (count + 1) * 25;
                        //the direct parent is the last one this command effects
                        parent = thisbox;
                    }
                }

                main.scriptClose.Top -= 25 * (count + 1);

                if (parent == null || (parent.identity < 2 && parent.identity >= 0))
                {
                    loadValues(valueType);
                }
                else
                {
                    int pnum = parent.scriptNum;
                    for (int i = 0; i < ScriptDatabaseInfo.Entries[pnum].Parameters.Count; i++)
                    {
                        if (i == 0)
                            box = parent.child;
                        else
                            box = box.sibling;
                        if (box == this)
                        {
                            if (identity == -1)
                                loadValues(ScriptDatabaseInfo.Entries[pnum].Parameters[i]);
                            else
                            {
                                valueType = ScriptDatabaseInfo.Entries[pnum].Parameters[i];
                                loadValues(identity);
                                valueType = ScriptDatabaseInfo.Entries[pnum].Parameters[i];
                            }
                            break;
                        }
                    }
                }

                this.Text = "";
                #endregion
            }

            else if (this.identity == identity) //changing the types
            {
                if (identity < 2)
                {
                    #region beginadd
                    //just add another box
                    int line = this.closeP.Top / 25;
                    ConstructBox box = new ConstructBox(line, indent + 1);
                    box.loadValues(4);
                    ConstructBox temp = null;
                    while (true)
                    {
                        if (temp == null)
                            temp = this.child;
                        else if (temp.sibling != null)
                            temp = temp.sibling;
                        else
                        {
                            temp.sibling = box;
                            break;
                        }
                    }
                    this.Parent.Controls.Add(box);
                    paramCount++;

                    foreach (Control tempbox in this.Parent.Controls)
                    {
                        if (tempbox.GetType().Name != "ConstructBox")
                            continue;
                        ConstructBox thisbox = (ConstructBox)tempbox;
                        if (thisbox.Top > this.closeP.Top)
                            thisbox.line++;
                        else if (thisbox.closeP != null && thisbox != this && thisbox.Top < this.Top && thisbox.closeP.Top > this.Top)
                        {
                            thisbox.closeP.Top += 25;
                        }
                    }

                    this.closeP.Top += 25;
                    ((Constructor)this.Parent.Parent.Parent.Parent).scriptClose.Top += 25;
                    #endregion
                }
                else
                {
                    int num = scriptNum;
                    this.makeCommand(-1);
                    if (ScriptDatabaseInfo.Entries[num + 1].Identity == identity)
                        scriptNum = num;
                    else
                        scriptNum = -1;
                    this.makeCommand(identity);
                }
            }

            else if (this.identity < 0 && (identity == -2 || identity == -3))
            {
                this.identity = identity;
                int type = valueType;
                loadValues(identity);
                valueType = type;
            }

            else if (this.identity < -1 && identity == -1)
            {
                loadValues(valueType);
                this.identity = -1;
            }

            else //changing command
            {
                this.makeCommand(-1); //clear everything first
                this.makeCommand(identity); //redo it
            }

            panel.AutoScrollPosition = new Point(horiz, vert);
            for (int i = 0; i < this.Items.Count; i++)
            {
                if (this.Items[i].ToString().Length * 4 > this.Width)
                    this.Width = this.Items[i].ToString().Length * 4;
            }
        }

        public override string ToString()
        {
            if (this.identity >= 0)
                return this.Text;
            else
                return this.Text + " (" + ScriptTools.ScriptValueTypes.Names[this.valueType] + ")";
        }
    }

    public class dropdownValue
    {
        public string name;
        public byte[] hex;

        public dropdownValue(string name)
        {
            hex = new byte[4];
            this.name = name;
        }

        public dropdownValue(string name, int index)
        {
            hex = new byte[4];
            BitConverter.GetBytes(index).CopyTo(hex, 0);
            this.name = name;
        }

        public dropdownValue(string name, short index)
        {
            hex = new byte[4];
            BitConverter.GetBytes(index).CopyTo(hex, 0);
            this.name = name;
        }

        public dropdownValue(string name, byte a, byte b, byte c, byte d)
        {
            hex = new byte[] { a, b, c, d };
            this.name = name;
        }

        public override string ToString()
        {
            return name;
        }
    }
}