using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using HaloMap;
using System.Collections;

namespace ScriptTools
{
    public partial class ScriptEditorForm : Form
    {
        public int scriptNum;

        public ScriptEditorForm()
        {
            InitializeComponent();
        }

        public static string Decompile(Script sinfo)
        {
            return decomp("script", ScriptTypes.Names[sinfo.Type], sinfo.Name, sinfo.ChunkNumber, true);
        }

        public static string decomp(string scriptType, string typetext, string name, int chunkNumber, bool database)
        {
            Map map = Maps.map[0];
            BinaryReader br = map.BR;

            ScriptExpression chunk;
            List<ScriptExpression> chunks = new List<ScriptExpression>();
            List<int> parents = new List<int>();
            string output = "";
            output += "(" + scriptType + " " + typetext + " " + name + " ";
            string indent = " ";
            chunk = new ScriptExpression(chunkNumber);
            chunks.Add(chunk);
            do
            {
                if (chunk.ExpressionType != 8 && chunk.ValueType >= 0 && chunk.ValueType < 57 && (ScriptValueTypes.Names[chunk.ValueType] == "unknown" || chunk.ValueType == 12) && database)
                {
                    br.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + chunk.ChunkNumber * 20 + 16;
                    int[] value = new int[4];
                    for (int i = 0; i < 4; i++)
                    {
                        value[i] = br.ReadByte();
                    }
                    ScriptDatabase.AddValue(chunk.String, chunk.ValueType, chunk.Identity, value);
                }
                if (parents.Count > 0 && chunks[parents[parents.Count - 1]].ExpressionType == 8 &&
                    chunks[parents[parents.Count - 1]].Identity > 1 && parents[parents.Count - 1] != chunks.Count - 2
                    && !(chunk.Child != null && chunk.Child.Followed) && database)
                {
                    chunks[parents[parents.Count - 1]].DatabaseInfo.Parameters.Add(chunk.ValueType);
                }
                if (chunk.ExpressionType == 8 || (chunk.ExpressionType == 10 && chunk.ValueType != 2)) //syntax chunk "("
                {
                    if (chunk.Child.Followed) //second time, follow sibling
                    {
                        if (chunk.Sibling == null)
                        {
                            chunk = chunks[parents[parents.Count - 1]];
                            parents.Remove(parents[parents.Count - 1]);
                            if (chunk.Identity < 2)
                            {
                                indent = indent.Substring(1, indent.Length - 1);
                                output += "\r\n" + indent;
                            }
                            else if(chunk.DatabaseInfo != null)
                                ScriptDatabaseInfo.Add(chunk.DatabaseInfo);
                            output += ")";
                        }
                        else
                        {
                            output += " ";
                            chunk = new ScriptExpression(chunk.Sibling.ChunkNumber);
                            chunks.Add(chunk);
                        }
                    }
                    else //first time, follow child
                    {
                        parents.Add(chunks.Count - 1);
                        if (chunk.Identity < 2 || (chunks.Count > 1 && chunks[chunks.Count - 2].Identity < 2))
                        {
                            indent += " ";
                            output += "\r\n" + indent;
                        }
                        else if (chunk.ValueType == 4)
                            output += "\r\n" + indent;
                        output += "(";
                        chunk.Child.Followed = true;
                        chunk = new ScriptExpression(chunk.Child.ChunkNumber);
                        chunks.Add(chunk);
                        if (chunks[parents[parents.Count - 1]].ExpressionType == 8)
                            chunks[parents[parents.Count - 1]].DatabaseInfo.Name = chunk.String;
                    }
                }
                else
                {
                    switch (chunk.ExpressionType)
                    {
                        case 9:
                            switch (chunk.ValueType)
                            {
                                case 4:
                                    output += "void";
                                    break;
                                case 5:
                                    output += chunk.Bool.ToString();
                                    break;
                                case 6:
                                    output += chunk.Real.ToString();
                                    break;
                                case 7:
                                    output += chunk.Int.ToString();
                                    break;
                                case 8:
                                    output += chunk.Long.ToString();
                                    break;
                                case 9:
                                    output += "\"" + chunk.String + "\"";
                                    break;
                                case 11:
                                    output += chunk.String.Replace(' ', '_');
                                    break;
                                default:
                                    output += chunk.String;
                                    break;
                            }
                            break;
                        case 10:
                            //static script
                            if (chunk.ValueType == 4)
                                output += "\r\n" + indent;
                            output += "(" + chunk.String + ")";
                            break;
                        case 13:
                            //global
                            output += chunk.String;
                            break;
                        default:
                            //unknown
                            output += "<UNKNOWN:" + chunk.ExpressionType + ", " + chunk.String + ">";
                            break;
                    }
                    if (chunk.Sibling == null)
                    {
                        if (parents.Count > 0)
                        {
                            chunk = chunks[parents[parents.Count - 1]];
                            parents.Remove(parents[parents.Count - 1]);
                            output += ")";
                            if (chunk.Identity > 1 && chunk.DatabaseInfo != null)
                                ScriptDatabaseInfo.Add(chunk.DatabaseInfo);
                        }
                        if (chunk.Identity < 2)
                        {
                            indent = indent.Substring(1, indent.Length - 1);
                            output += "\r\n" + indent;
                        }
                    }
                    else
                    {
                        output += " ";
                        chunk = new ScriptExpression(chunk.Sibling.ChunkNumber);
                        chunks.Add(chunk);
                    }
                }
            } while (parents.Count > 0);
            if (scriptType == "script")
                output += "\r\n)";
            else
                output += ")";

            return output;
        }

        public static string DecompileGlobal(ScriptGlobal gbl)
        {
            return decomp("global", ScriptValueTypes.Names[gbl.Type], gbl.Name, gbl.ChunkNumber, true);
        }

        public void decompile(Script sinfo)
        {
            textBox1.Text = ScriptEditorForm.Decompile(sinfo);

            var dbFile = new FileInfo(Path.Combine(Application.StartupPath, "db.txt"));

            if (dbFile.Exists)
                dbFile.Delete();

            ScriptDatabase.Save(dbFile);
            ScriptDatabaseInfo.Save(dbFile);
        }

        public static scriptInfo comp(string script, int startChunk)
        {
            scriptInfo sinfo = new scriptInfo();
            List<ScriptExpression> syntax = new List<ScriptExpression>();
            List<int> parents = new List<int>();
            List<int> scriptNumber = new List<int>();

            int pos1 = script.IndexOf('(') + 1;
            int pos2 = script.IndexOf(' ', pos1);
            if (script.Substring(pos1, pos2 - pos1).ToLower() == "script")
                sinfo.type = type.Script;
            else if (script.Substring(pos1, pos2 - pos1).ToLower() == "global")
                sinfo.type = type.Global;
            else
            {
                MessageBox.Show("Error!  Invalid header!");
                return null;
            }
            pos1 = pos2 + 1;
            pos2 = script.IndexOf(' ', pos1);
            string headertype = script.Substring(pos1, pos2 - pos1);
            if (sinfo.type == type.Global)
            {
                byte[] rtype = BitConverter.GetBytes(ScriptValueTypes.GetIndex(headertype));
                sinfo.header[32] = rtype[0];
                sinfo.header[33] = rtype[1];
            }
            else
            {
                byte[] stype = BitConverter.GetBytes(ScriptTypes.GetIndex(headertype));
                sinfo.header[32] = stype[0];
                sinfo.header[33] = stype[1];
            }
            byte[] chunknumber = BitConverter.GetBytes(startChunk);
            sinfo.header[36] = chunknumber[0];
            sinfo.header[37] = chunknumber[1];
            byte[] rootexp = BitConverter.GetBytes(58227 + startChunk);
            sinfo.header[38] = rootexp[0];
            sinfo.header[39] = rootexp[1];
            pos1 = pos2 + 1;
            pos2 = script.IndexOfAny(new char[] { ' ', '\r', '\t', '(' }, pos1);
            string scriptname = script.Substring(pos1, pos2 - pos1);
            char[] name = script.ToCharArray(pos1, pos2 - pos1);
            pos1 = pos2 + 1;
            pos2 = script.IndexOfAny(new char[] { '(' }, pos1);
            if (pos2 == -1)
                pos2 = pos1;
            for (int i = 0; i < (name.Length < 32 ? name.Length : 32); i++)
            {
                sinfo.header[i] = (byte)name[i];
            }
            while (true) //break out later
            {
                ScriptExpression chunk = new ScriptExpression();
                ScriptExpression parent = null;
                int j;
                CharEnumerator enumer;

                if (parents.Count > 0)
                    chunk.Parent = parents[parents.Count - 1];

                chunk.ChunkNumber = startChunk + syntax.Count;
                chunk.ExpressionIndex = (uint)(startChunk + syntax.Count + 58227);

                if(script.Substring(pos2, 1) == "(")
                {
                    if (syntax.Count > 0)
                    {
                        if (syntax[chunk.Parent].Children.Count > 0)
                            syntax[syntax[chunk.Parent].Children[syntax[chunk.Parent].Children.Count - 1]].Sibling = new ScriptPointer(syntax.Count + startChunk);
                        else
                            syntax[chunk.Parent + 1].Sibling = new ScriptPointer(syntax.Count + startChunk);
                    }
                    parents.Add(syntax.Count);
                    scriptNumber.Add(0); //increment it as we find ones that don't work
                    chunk.ExpressionType = 8; //assume it until proven false 
                    parent = chunk;
                    chunk = new ScriptExpression();
                    chunk.Parent = syntax.Count;
                    parent.Child = new ScriptPointer(syntax.Count + startChunk + 1);
                    syntax.Add(parent);
                    chunk.ChunkNumber = startChunk + syntax.Count;
                    chunk.ExpressionIndex = (uint)(startChunk + syntax.Count + 58227);
                    scriptNumber[chunk.Parent] = 0;
                    chunk.ExpressionType = 9;
                    chunk.ValueType = 2;
                    pos2++;
                }
                else if (script.Substring(pos2, 1) == ")")
                {
                    //tons of post-processing stuff to determine return + value types
                    //based on the number/kinds of parameters

                    if (parents.Count > 0)
                    {
                        chunk = syntax[parents[parents.Count - 1]]; //temp
                        if (parents.Count > 1)
                        {
                            int i = scriptNumber[parents[parents.Count - 1]];
                            int id = ScriptDatabaseInfo.Entries[i].Identity;
                            //at this point, all parameters match this db entry
                            //let's make sure its return type fits

                            if (chunk.ExpressionType == 10)
                            {
                                if (ScriptDatabaseInfo.Entries[scriptNumber[chunk.Parent]].Parameters.Count > syntax[chunk.Parent].Children.Count && (syntax[chunk.Parent].Identity < 2 || ScriptDatabaseInfo.Entries[scriptNumber[chunk.Parent]].Parameters[syntax[chunk.Parent].Children.Count] == chunk.ValueType))
                                {
                                    syntax[chunk.Parent].Children.Add(parents[parents.Count - 1]);
                                }
                                else
                                {
                                    int id2 = ScriptDatabaseInfo.Entries[scriptNumber[chunk.Parent]].Identity;
                                    for (int x = scriptNumber[chunk.Parent]; ScriptDatabaseInfo.Entries[x].Identity == id2; x++)
                                    {
                                        if (ScriptDatabaseInfo.Entries[x].Parameters.Count <= syntax[chunk.Parent].Children.Count)
                                            continue;

                                        if (ScriptDatabaseInfo.Entries[x].Parameters[syntax[chunk.Parent].Children.Count] == ScriptDatabaseInfo.Entries[i].ReturnType)
                                        {
                                            scriptNumber[chunk.Parent] = x;
                                            break;
                                        }
                                    }
                                    if (id2 != ScriptDatabaseInfo.Entries[scriptNumber[chunk.Parent]].Identity)
                                    {
                                        MessageBox.Show("Invalid Parameters\n" + syntax[chunk.Parent + 1].String, "Error!");
                                        return null;
                                    }
                                }
                            }
                            else if (syntax[chunk.Parent].Identity > 1)
                            {
                                if (syntax[chunk.Parent].Children.Count >= ScriptDatabaseInfo.Entries[scriptNumber[chunk.Parent]].Parameters.Count)
                                {
                                    scriptNumber[chunk.Parent]++;
                                    continue; //in an infinite while, nothing's happened yet... makes life easier
                                }
                                int param = ScriptDatabaseInfo.Entries[scriptNumber[chunk.Parent]].Parameters[syntax[chunk.Parent].Children.Count];
                                int id2 = ScriptDatabaseInfo.Entries[scriptNumber[chunk.Parent]].Identity;
                                if (param == ScriptDatabaseInfo.Entries[i].ReturnType)
                                {
                                    chunk.ValueType = ScriptDatabaseInfo.Entries[i].ReturnType;
                                    syntax[chunk.Parent].Children.Add(parents[parents.Count - 1]);
                                }
                                else
                                {
                                    int a = scriptNumber[chunk.Parent];
                                    for (int x = i; ScriptDatabaseInfo.Entries[x].Identity == id; x++) //see if it has one that matches
                                    {
                                        if (ScriptDatabaseInfo.Entries[x].ReturnType == param)
                                        {
                                            scriptNumber[parents[parents.Count - 1]] = x;
                                            break;
                                        }
                                    }
                                    if (scriptNumber[parents[parents.Count - 1]] == i) //hasn't changed
                                    {
                                        for (int x = scriptNumber[chunk.Parent]; ScriptDatabaseInfo.Entries[x].Identity == id2; x++)
                                        {
                                            if (ScriptDatabaseInfo.Entries[x].Parameters.Count <= syntax[chunk.Parent].Children.Count)
                                                continue;

                                            if (ScriptDatabaseInfo.Entries[x].Parameters[syntax[chunk.Parent].Children.Count] == ScriptDatabaseInfo.Entries[i].ReturnType)
                                            {
                                                scriptNumber[chunk.Parent] = x;
                                                break;
                                            }
                                        }
                                        if (scriptNumber[chunk.Parent] == a)
                                        {
                                            MessageBox.Show("Parameter Mismatch\n" + syntax[chunk.Parent + 1].String, "Error!");
                                            return null;
                                        }
                                    }
                                    continue;
                                }
                            }
                            else
                            {
                                chunk.ValueType = ScriptDatabaseInfo.Entries[i].ReturnType;
                                syntax[chunk.Parent].Children.Add(parents[parents.Count - 1]);
                            }
                        }
                        if (chunk.Identity < 2 && chunk.ExpressionType == 8)
                        {
                            chunk.ValueType = syntax[chunk.Children[chunk.Children.Count - 1]].ValueType;
                        }
                        parents.RemoveAt(parents.Count - 1);
                    }
                    else
                    {
                        break; //we're done!!
                    }

                    j = 1;
                    enumer = script.Substring(pos2 + 1).GetEnumerator();
                    enumer.MoveNext();
                    while (enumer.Current == ' ' || enumer.Current == '\r' || enumer.Current == '\n' || enumer.Current == '\t')
                    {
                        j++;
                        enumer.MoveNext();
                    }
                    pos2 += j;
                    continue;
                }
                else
                {
                    if (syntax.Count > 0)
                    {
                        if (syntax[chunk.Parent].Children.Count > 0)
                            syntax[syntax[chunk.Parent].Children[syntax[chunk.Parent].Children.Count - 1]].Sibling = new ScriptPointer(syntax.Count + startChunk);
                        else
                            syntax[chunk.Parent + 1].Sibling = new ScriptPointer(syntax.Count + startChunk);
                    }
                    chunk.ExpressionType = 9;
                }

                string command;
                if (script.Substring(pos2, 1) == "\"")
                {
                    pos1 = pos2 + 1;
                    pos2 = script.IndexOf('"', pos1);
                    command = script.Substring(pos1, pos2 - pos1);
                    pos2++;
                }
                else
                {
                    pos1 = pos2;
                    pos2 = script.IndexOfAny(new char[] { ' ', '\r', '\t', '(', ')' }, pos1);
                    command = script.Substring(pos1, pos2 - pos1);
                }
                float test;
                bool test2;
                if (command != "(" && !float.TryParse(command, out test) && !bool.TryParse(command, out test2))
                {
                    chunk.StringOffset = getStringPointer(command);
                }
                j = 0;
                enumer = script.Substring(pos2).GetEnumerator();
                enumer.MoveNext();
                while (enumer.Current == ' ' || enumer.Current == '\r' || enumer.Current == '\n' || enumer.Current == '\t')
                {
                    j++;
                    enumer.MoveNext();
                }
                pos2 += j;
                if (parent != null) //find out identities
                {
                    int identity = -1;
                    chunk.ValueType = 2;
                    chunk.String = command;
                    if (command == "begin")
                    {
                        identity = 0;
                        chunk.Identity = 0;
                        parent.Identity = 0;
                    }
                    else if (command == "begin_random")
                    {
                        identity = 1;
                        chunk.Identity = 1;
                        parent.Identity = 1;
                    }
                    else
                    {
                        for (int i = 0; i < ScriptListForm.scripts.Length; i++)
                        {
                            if (ScriptListForm.scripts[i].Name == command)
                            {
                                identity = i;
                                chunk.Identity = i;
                                parent.Identity = i;
                                parent.ExpressionType = 10; //static script
                                parent.ValueType = ScriptListForm.scripts[i].ReturnType;
                                break;
                            }
                        }
                    }
                    if (identity < 0) //not a static script
                    {
                        for (int i = 0; i < ScriptDatabaseInfo.Entries.Count; i++)
                        {
                            if (ScriptDatabaseInfo.Entries[i].Name == command)
                            {
                                identity = i;
                                chunk.Identity = ScriptDatabaseInfo.Entries[i].Identity;
                                parent.Identity = ScriptDatabaseInfo.Entries[i].Identity;
                                scriptNumber[chunk.Parent] = i;
                                break;
                            }
                        }
                    }
                    if (identity < 0)
                    {
                        MessageBox.Show("Unknown Script:\n" + command, "Error");
                        return null;
                    }
                }
                else //parameter... let's find out what
                {
                    Map map = Maps.map[0];
                    BinaryReader br = map.BR;
                    if (sinfo.type == type.Script || syntax.Count > 0)
                    {
                        if (syntax[chunk.Parent].Identity < 2) //begin/begin_random
                        {
                            syntax[chunk.Parent].Children.Add(syntax.Count);
                        }
                        else
                        {
                            for (int i = scriptNumber[chunk.Parent]; i < ScriptDatabaseInfo.Entries.Count; i++)
                            {
                                if (ScriptDatabaseInfo.Entries[i].Name != syntax[chunk.Parent + 1].String)
                                {
                                    string sname = syntax[chunk.Parent + 1].String;
                                    MessageBox.Show("Invalid Parameters\n" + command + "\n" + sname);
                                    return null;
                                }

                                if (ScriptDatabaseInfo.Entries[i].Parameters.Count <= syntax[chunk.Parent].Children.Count)
                                    continue;

                                int param = ScriptDatabaseInfo.Entries[i].Parameters[syntax[chunk.Parent].Children.Count];
                                chunk.ValueType = param;
                                chunk.Identity = param;
                                for (int a = 0; a < ScriptListForm.globals.Length; a++)
                                {
                                    if (ScriptListForm.globals[a].Name == command)
                                    {
                                        chunk.ExpressionType = 13;
                                        chunk.Int = a;
                                        chunk.Hex = BitConverter.GetBytes(a);
                                        chunk.ValueType = ScriptListForm.globals[a].Type;
                                        chunk.String = command;
                                        break;
                                    }
                                }
                                if (chunk.ExpressionType == 13) //found a global
                                    break;
                                else
                                {
                                    bool good = getvalue(ref chunk, param, command);
                                    if (chunk == null)
                                    {
                                        MessageBox.Show("Problem with the script!");
                                        return null;
                                    }
                                    if (!good)
                                    {
                                        scriptNumber[chunk.Parent]++;
                                        continue;
                                    }
                                }
                                break; //only continue if explicitly told
                            }
                            syntax[chunk.Parent].Children.Add(syntax.Count);
                        }
                    }
                    else
                    {
                        int param = (int)sinfo.header[32];
                        for (int a = 0; a < ScriptListForm.globals.Length; a++)
                        {
                            if (ScriptListForm.globals[a].Name == command)
                            {
                                chunk.ExpressionType = 13;
                                chunk.Int = a;
                                chunk.Hex = BitConverter.GetBytes(a);
                                chunk.ValueType = ScriptListForm.globals[a].Type;
                                chunk.String = command;
                                break;
                            }
                        }
                        if (chunk.ExpressionType != 13)
                        {
                            bool good = getvalue(ref chunk, param, command);
                            if (!good)
                            {
                                MessageBox.Show("Invalid Global value!");
                                break;
                            }
                        }
                        chunk.ValueType = param;
                        chunk.Identity = param;
                    }
                }
                syntax.Add(chunk);
                scriptNumber.Add(0);
            }

            if (sinfo.type == type.Script)
                BitConverter.GetBytes((short)syntax[0].ValueType).CopyTo(sinfo.header, 34);

            //and now to convert these syntax chunks into hex...

            for (int i = 0; i < syntax.Count; i++)
            {
                ScriptExpression chunk = syntax[i];
                byte[] hex = new byte[20];

                byte[] expindex = BitConverter.GetBytes(chunk.ExpressionIndex);
                hex[0] = expindex[0];
                hex[1] = expindex[1];

                byte[] identity = BitConverter.GetBytes(chunk.Identity);
                hex[2] = identity[0];
                hex[3] = identity[1];

                byte[] valueType = BitConverter.GetBytes(chunk.ValueType);
                hex[4] = valueType[0];
                hex[5] = valueType[1];

                byte[] expType = BitConverter.GetBytes(chunk.ExpressionType);
                hex[6] = expType[0];
                hex[7] = expType[1];

                if (chunk.Sibling == null)
                {
                    hex[8] = 255;
                    hex[9] = 255;
                    hex[10] = 255;
                    hex[11] = 255;
                }
                else
                {
                    byte[] schunkn = BitConverter.GetBytes(chunk.Sibling.ChunkNumber);
                    hex[8] = schunkn[0];
                    hex[9] = schunkn[1];

                    byte[] schunke = BitConverter.GetBytes(chunk.Sibling.ExpressionIndex);
                    hex[10] = schunke[0];
                    hex[11] = schunke[1];
                }

                byte[] stringPointer = BitConverter.GetBytes(chunk.StringOffset);
                hex[12] = stringPointer[0];
                hex[13] = stringPointer[1];
                hex[14] = stringPointer[2];
                hex[15] = stringPointer[3];

                if (chunk.Child != null) //never converts a child pointer to hex
                {
                    byte[] cchunkn = BitConverter.GetBytes(chunk.Child.ChunkNumber);
                    hex[16] = cchunkn[0];
                    hex[17] = cchunkn[1];

                    byte[] cchunke = BitConverter.GetBytes(chunk.Child.ExpressionIndex);
                    hex[18] = cchunke[0];
                    hex[19] = cchunke[1];
                }
                else
                {
                    hex[16] = chunk.Hex[0];
                    hex[17] = chunk.Hex[1];
                    hex[18] = chunk.Hex[2];
                    hex[19] = chunk.Hex[3];
                }

                for (int j = 0; j < 20; j++)
                {
                    sinfo.syntax.Add(hex[j]);
                }
            }

            if (sinfo.type == type.Script)
            {
                sinfo.sinfo = new Script();
                sinfo.sinfo.ChunkNumber = startChunk;
                sinfo.sinfo.ExpressionIndex = (uint)(58227 + startChunk);
                sinfo.sinfo.Name = scriptname;
                sinfo.sinfo.ReturnType = syntax[0].ValueType;
                sinfo.sinfo.Type = ScriptTypes.GetIndex(headertype);
            }
            else
            {
                sinfo.ginfo = new ScriptGlobal();
                sinfo.ginfo.ChunkNumber = startChunk;
                sinfo.ginfo.ExpressionIndex = (uint)(58227 + startChunk);
                sinfo.ginfo.Name = scriptname;
                sinfo.ginfo.Type = ScriptValueTypes.GetIndex(headertype);
            }

            return sinfo;
        }

        public static bool getvalue(ref ScriptExpression chunk, int param, string command)
        {
            Map map = Maps.map[0];
            BinaryReader br = map.BR;
            if (command == "destroyed") //so it doesn't crash >_>
            {
                chunk.Hex = new byte[] { 4, 0, 255, 255 };
                chunk.Identity = 48;
                chunk.ValueType = 48;
                return true;
            }
            if (command.StartsWith("ai_combat_status_"))
            {
                chunk.ExpressionType = 13;
                chunk.ValueType = 7;
                chunk.Identity = 7;
                switch (command.Substring(17))
                {
                    case "alert":
                        chunk.Hex = new byte[] { 39, 131, 255, 255 };
                        break;
                    case "active":
                        chunk.Hex = new byte[] { 40, 131, 255, 255 };
                        break;
                    case "uninspected":
                        chunk.Hex = new byte[] { 41, 131, 255, 255 };
                        break;
                    case "definite":
                        chunk.Hex = new byte[] { 42, 131, 255, 255 };
                        break;
                    case "certain":
                        chunk.Hex = new byte[] { 43, 131, 255, 255 };
                        break;
                    case "visible":
                        chunk.Hex = new byte[] { 44, 131, 255, 255 };
                        break;
                    case "clear_los":
                        chunk.Hex = new byte[] { 45, 131, 255, 255 };
                        break;
                    case "dangerous":
                        chunk.Hex = new byte[] { 46, 131, 255, 255 };
                        break;
                    case "idle":
                        chunk.Hex = new byte[] { 47, 131, 255, 255 };
                        break;
                }
            }
            else if (command.StartsWith("ai_movement_"))
            {
                chunk.ExpressionType = 13;
                chunk.ValueType = 7;
                chunk.Identity = 7;
                switch (command.Substring(12))
                {
                    case "patrol":
                        chunk.Hex = new byte[] { 47, 131, 255, 255 };
                        break;
                    case "combat":
                        chunk.Hex = new byte[] { 49, 131, 255, 255 };
                        break;
                    case "flee":
                        chunk.Hex = new byte[] { 50, 131, 255, 255 };
                        break;
                }
            }
            else if (param > 2 && param < 9)
            {
                try //numbers only
                {
                    switch (param)
                    {
                        case 4:
                            chunk.Hex = new byte[] { 255, 255, 255, 255 };
                            break;
                        case 5:
                            chunk.Bool = bool.Parse(command);
                            chunk.Hex = new byte[] { (byte)(chunk.Bool ? 1 : 0), 255, 255, 255 };
                            break;
                        case 6:
                            chunk.Real = float.Parse(command);
                            chunk.Hex = BitConverter.GetBytes(chunk.Real);
                            break;
                        case 7:
                            chunk.Int = int.Parse(command);
                            byte[] b = BitConverter.GetBytes(chunk.Int);
                            chunk.Hex = new byte[] { b[0], b[1], 255, 255 };
                            break;
                        case 8:
                            chunk.Long = long.Parse(command);
                            chunk.Hex = BitConverter.GetBytes(int.Parse(command));
                            break;
                    }
                }
                catch
                {
                    return false;
                }
            }
            else if (param == 2 || (param > 8 && param < 14) || param == 16 || param == 19 || param == 21 || param == 26 || param == 28 || (param > 30 && param < 42) || param == 44 || param == 45) //odd ones
            {
                TagMeta meta = new TagMeta();
                meta.ReadMetaFromMap(3, 0, false);
                chunk.String = command;
                switch (param)
                {
                    case 2:
                        chunk.Hex = new byte[] { 0, 0, 0, 0 };
                        for (int x = 0; x < ScriptDatabaseInfo.Entries.Count; x++)
                        {
                            if (ScriptDatabaseInfo.Entries[x].Name == command)
                            {
                                chunk.Identity = ScriptDatabaseInfo.Entries[x].Identity;
                                break;
                            }
                        }
                        if (chunk.Identity == 2)
                        {
                            return false;
                        }
                        break;
                    case 10:
                    case 21:
                        for (int x = 0; x < ScriptListForm.scripts.Length; x++)
                        {
                            if (ScriptListForm.scripts[x].Name == command)
                            {
                                chunk.Hex = BitConverter.GetBytes(x);
                                chunk.Int = x;
                            }
                        }
                        if (chunk.Int <= 0)
                        {
                            return false;
                        }
                        break;
                    case 9:
                        //strings... do nothing, just need a table ref
                        break;
                    case 11:
                        for (int x = 0; x < map.Strings.Name.Length; x++)
                        {
                            if (map.Strings.Name[x] == command)
                            {
                                byte[] b = BitConverter.GetBytes(x);
                                chunk.Hex = new byte[] { b[0], b[1], 0, (byte)map.Strings.Length[x] };
                                break;
                            }
                        }
                        if (chunk.Hex[3] == 0)
                        {
                            return false;
                        }
                        break;
                    case 12:
                        //vehicle seats... tba when i know how...
                        break;
                    case 13:
                        br.BaseStream.Position = meta.offset + 264;
                        int volcount = br.ReadInt32();
                        br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                        for (int x = 0; x < volcount; x++)
                        {
                            byte[] blah = new byte[4];
                            for (int y = 0; y < 2; y++)
                            {
                                blah[y] = br.ReadByte();
                            }
                            int num = BitConverter.ToInt32(blah, 0);
                            if (command == map.Strings.Name[num])
                            {
                                chunk.Int = x;
                                chunk.Hex = BitConverter.GetBytes(x);
                                break;
                            }
                            br.BaseStream.Position += 66;
                        }
                        if (chunk.Int <= 0)
                        {
                            return false;
                        }
                        break;
                    case 16: //cutscene titles
                        br.BaseStream.Position = meta.offset + 496;
                        int count = br.ReadInt32();
                        br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                        bool good = false;
                        for (int x = 0; x < count; x++)
                        {
                            byte[] blah = new byte[4];
                            for (int y = 0; y < 2; y++)
                            {
                                blah[y] = br.ReadByte();
                            }
                            int num = BitConverter.ToInt32(blah, 0);
                            if (command == map.Strings.Name[num])
                            {
                                chunk.Int = x;
                                chunk.Hex = BitConverter.GetBytes(x);
                                good = true;
                                break;
                            }
                            br.BaseStream.Position += 34;
                        }
                        if (!good)
                        {
                            return false;
                        }
                        break;
                    case 19: //AI...
                        if (command == "ai_current_actor") //constant
                        {
                            chunk.Hex = new byte[] { 30, 131, 255, 255 };
                            chunk.ExpressionType = 13;
                        }
                        else if (command == "ai_current_squad")
                        {
                            chunk.Hex = new byte[] { 29, 131, 255, 255 };
                            chunk.ExpressionType = 13;
                        }
                        else
                        {
                            string[] str = null;
                            if (command.IndexOf('/') > 0)
                            {
                                str = command.Split(new char[] { '/' });
                                command = str[0];
                            }
                            br.BaseStream.Position = meta.offset + 352;
                            int acount = br.ReadInt32();
                            br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                            for (int x = 0; x < acount; x++)
                            {
                                string blah = "";
                                for (int a = 0; a < 32; a++)
                                {
                                    Byte bt = br.ReadByte();
                                    if (bt == 0)
                                    {
                                        br.BaseStream.Position += (31 - a);
                                        break; //null separator
                                    }
                                    blah += BitConverter.ToChar(new byte[] { bt, 0 }, 0);
                                }
                                if (blah == command)
                                {
                                    if (str != null)
                                    {
                                        br.BaseStream.Position += 40;
                                        count = br.ReadInt32();
                                        br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                                        for (int y = 0; y < count; y++)
                                        {
                                            blah = "";
                                            for (int b = 0; b < 32; b++)
                                            {
                                                Byte bt = br.ReadByte();
                                                if (bt == 0)
                                                {
                                                    br.BaseStream.Position += (31 - b);
                                                    break; //null separator
                                                }
                                                blah += BitConverter.ToChar(new byte[] { bt, 0 }, 0);
                                            }
                                            if (blah == str[1])
                                            {
                                                byte[] a = BitConverter.GetBytes(x);
                                                byte[] b = BitConverter.GetBytes(y);
                                                chunk.Hex = new byte[] { a[0], a[1], b[0], 192 };
                                                break;
                                            }
                                            br.BaseStream.Position += 68;
                                        }
                                        break;
                                    }
                                    else
                                    {
                                        chunk.Hex = BitConverter.GetBytes(x);
                                        break;
                                    }
                                }
                                br.BaseStream.Position += 84;
                            }
                            //oh boy... it's a platoon, change reflexives
                            br.BaseStream.Position = meta.offset + 344;
                            count = br.ReadInt32();
                            br.BaseStream.Position = br.ReadUInt32() - (uint)map.SecondaryMagic;
                            for (int x = 0; x < count; x++)
                            {
                                string blah = "";
                                for (int a = 0; a < 32; a++)
                                {
                                    Byte bt = br.ReadByte();
                                    if (bt == 0)
                                    {
                                        br.BaseStream.Position += (35 - a);
                                        break; //null separator
                                    }
                                    blah += BitConverter.ToChar(new byte[] { bt, 0 }, 0);
                                }
                                if (blah == command)
                                {
                                    byte[] tmp = BitConverter.GetBytes(x);
                                    chunk.Hex = new byte[] { tmp[0], 0, 0, 64 };
                                    break;
                                }
                            }
                        }
                        if (chunk.Hex.Equals(new byte[] { 0, 0, 0, 0 }))
                        {
                            return false;
                        }
                        break;
                    case 26:
                        int ident = 0; //first locate its ident
                        good = false;
                        if (command.Contains("\\"))
                        {
                            for (int x = 0; x < map.FileNames.Name.Length; x++)
                            {
                                if (map.FileNames.Name[x] == command)
                                {
                                    ident = map.MetaInfo.Ident[x];
                                    break;
                                }
                            }
                            break;
                        }
                        else //can be referenced without path, just name
                        {
                            for (int x = 0; x < map.MetaInfo.TagType.Length; x++)
                            {
                                if (map.MetaInfo.TagType[x] == "sbsp")
                                {
                                    string[] str = map.FileNames.Name[x].Split(new char[] { '\\' });
                                    if (str[str.Length - 1] == command)
                                    {
                                        ident = map.MetaInfo.Ident[x];
                                        break;
                                    }
                                }
                            }
                        }
                        br.BaseStream.Position = meta.offset + 528;
                        int ccount = br.ReadInt32();
                        br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                        for (int x = 0; x < ccount; x++) //find out which chunk points to it
                        {
                            br.BaseStream.Position += 20;
                            if (br.ReadInt32() == ident)
                            {
                                chunk.Int = x;
                                chunk.Hex = BitConverter.GetBytes(x);
                                good = true;
                                break;
                            }
                            br.BaseStream.Position += 44;
                        }
                        if (!good)
                        {
                            return false;
                        }
                        break;
                    case 28:
                        string[] cmd = command.Split(new char[] { '/' });
                        br.BaseStream.Position = meta.offset + 476;
                        br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                        int outercount = br.ReadInt32();
                        br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                        for (int x = 0; x < outercount; x++)
                        {
                            string bleh = "";
                            for (int y = 0; y < 32; y++)
                            {
                                char meh = br.ReadChar();
                                if (meh == 0)
                                {
                                    br.BaseStream.Position += 31 - y;
                                    break;
                                }
                                bleh += meh.ToString();
                            }
                            if (bleh == cmd[0]) //we found our outer!
                            {
                                if (cmd.Length == 1)
                                {
                                    byte[] a = BitConverter.GetBytes(x);
                                    chunk.Hex = new byte[] { 255, 255, a[0], 0 };
                                    break;
                                }
                                int innercount = br.ReadInt32();
                                br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                                for (int y = 0; y < innercount; y++)
                                {
                                    string grr = "";
                                    for (int z = 0; z < 32; z++)
                                    {
                                        char moo = br.ReadChar();
                                        if (moo == 0)
                                        {
                                            br.BaseStream.Position += 31 - z;
                                            break;
                                        }
                                        grr += moo.ToString();
                                    }
                                    if (grr == cmd[1])
                                    {
                                        byte[] a = BitConverter.GetBytes(x);
                                        byte[] b = BitConverter.GetBytes(y);
                                        chunk.Hex = new byte[] { b[0], 0, a[0], 0 };
                                        break;
                                    }
                                    br.BaseStream.Position += 28;
                                }
                                break;
                            }
                            br.BaseStream.Position += 16;
                        }
                        break;
                    case 31:
                        if (command == "ai_current_actor")
                        {
                            chunk.Hex = new byte[] { 30, 131, 255, 255 };
                            chunk.ExpressionType = 13;
                        }
                        else
                        {
                            br.BaseStream.Position = meta.offset + 76;
                            int listcount = br.ReadInt32();
                            br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                            for (int x = 0; x < listcount; x++)
                            {
                                string mmm = "";
                                for (int y = 0; y < 32; y++)
                                {
                                    char aaa = br.ReadChar();
                                    if (aaa == 0)
                                    {
                                        br.BaseStream.Position += 31 - y;
                                        break;
                                    }
                                    mmm += aaa.ToString();
                                }
                                if (mmm == command)
                                {
                                    byte[] a = BitConverter.GetBytes(x);
                                    chunk.Hex = new byte[] { a[0], a[1], 255, 255 };
                                    break;
                                }
                                br.BaseStream.Position += 4;
                            }
                        }
                        break;
                    case 32:
                    case 33:
                    case 34:
                    case 35:
                    case 36:
                    case 38:
                    case 41:
                        if (command.IndexOf('.') > 0) //i.e. ".weapon"
                            command = command.Substring(0, command.IndexOf('.'));
                        for (int a = 0; a < map.MetaInfo.Ident.Length; a++)
                        {
                            if (map.FileNames.Name[a] == command)
                            {
                                byte[] hex = BitConverter.GetBytes(a);
                                if (param == 32 && (map.MetaInfo.TagType[a] == "snd!" || map.MetaInfo.TagType[a] == "ssce"))
                                    chunk.Hex = hex;
                                else if (param == 33 && map.MetaInfo.TagType[a] == "effe")
                                    chunk.Hex = hex;
                                else if (param == 34 && map.MetaInfo.TagType[a] == "jpt!")
                                    chunk.Hex = hex;
                                else if (param == 35 && map.MetaInfo.TagType[a] == "lsnd")
                                    chunk.Hex = hex;
                                else if (param == 36 && map.MetaInfo.TagType[a] == "jmad")
                                    chunk.Hex = hex;
                                else if (param == 38 && (map.MetaInfo.TagType[a] == "bipd" || map.MetaInfo.TagType[a] == "weap"))
                                    chunk.Hex = hex;
                                else if (param == 41 && (map.MetaInfo.TagType[a] == "mode" || map.MetaInfo.TagType[a] == "sbsp"))
                                    chunk.Hex = hex;
                            }
                        }
                        if (chunk.Hex.Equals(new byte[] { 0, 0, 0, 0 }))
                        {
                            return false;
                        }
                        break;
                    case 44:
                        switch (command)
                        {
                            case "easy":
                                chunk.Int = 0;
                                chunk.Hex = new byte[] { 0, 0, 255, 255 };
                                break;
                            case "normal":
                                chunk.Int = 1;
                                chunk.Hex = new byte[] { 1, 0, 255, 255 };
                                break;
                            case "heroic":
                                chunk.Int = 2;
                                chunk.Hex = new byte[] { 2, 0, 255, 255 };
                                break;
                            case "legendary":
                                chunk.Int = 3;
                                chunk.Hex = new byte[] { 3, 0, 255, 255 };
                                break;
                            default: return false;
                        }
                        break;
                    case 45:
                        switch (command)
                        {
                            case "none":
                                chunk.Hex = new byte[] { 0, 0, 255, 255 };
                                break;
                            case "player":
                                chunk.Hex = new byte[] { 1, 0, 255, 255 };
                                break;
                            case "human":
                                chunk.Hex = new byte[] { 2, 0, 255, 255 };
                                break;
                            case "covenant":
                                chunk.Hex = new byte[] { 3, 0, 255, 255 };
                                break;
                            case "flood":
                                chunk.Hex = new byte[] { 4, 0, 255, 255 };
                                break;
                            case "sentinel":
                                chunk.Hex = new byte[] { 5, 0, 255, 255 };
                                break;
                            case "heretic":
                                chunk.Hex = new byte[] { 6, 0, 255, 255 };
                                break;
                            case "prophet":
                                chunk.Hex = new byte[] { 7, 0, 255, 255 };
                                break;
                            default: return false;
                        }
                        break;
                }
            }
            else //indices, all scan reflexives
            {
                chunk.String = command;
                int off = 0;
                int off2 = 0;
                int size = 0;
                int tagnum = 3;
                TagMeta meta = new TagMeta();
                switch (param)
                {
                    case 14:
                        off = 480;
                        off2 = 4;
                        size = 56;
                        break;
                    case 15:
                        off = 488;
                        off2 = 4;
                        size = 64;
                        break;
                    case 18:
                        off = 160;
                        size = 40;
                        break;
                    case 22:
                        off = 84;
                        size = 32;
                        for (int x = 0; x < map.FileNames.Name.Length; x++)
                        {
                            if (map.FileNames.Name[x] == "ai\\all")
                            {
                                tagnum = x;
                                break;
                            }
                        }
                        break;
                    case 23:
                        off = 576;
                        size = 124;
                        break;
                    case 24:
                        off = 248;
                        size = 68;
                        break;
                    case 27:
                        off = 292;
                        size = 104;
                        for (int x = 0; x < map.FileNames.Name.Length; x++)
                        {
                            if (map.MetaInfo.TagType[x] == "hudg")
                            {
                                tagnum = x;
                                break;
                            }
                        }
                        break;
                    case 48:
                        chunk.Hex = new byte[] { 4, 0, 255, 255 };
                        return false;
                    case 50:
                    case 51:
                    case 52:
                    case 53:
                    case 54:
                    case 55:
                    case 56:
                        off = 72;
                        size = 36;
                        break;
                }
                meta.ReadMetaFromMap(tagnum, 0, false);
                br.BaseStream.Position = meta.offset + off;
                int ccount = br.ReadInt32();
                br.BaseStream.Position = br.ReadUInt32() - map.SecondaryMagic;
                for (int x = 0; x < ccount; x++)
                {
                    string bleh = "";
                    for (int y = 0; y < 32; y++)
                    {
                        char meh = br.ReadChar();
                        if (meh == 0)
                        {
                            br.BaseStream.Position += 31 - y;
                            break;
                        }
                        bleh += meh.ToString();
                    }
                    if (bleh == command)
                    {
                        chunk.Int = x;
                        chunk.Hex = BitConverter.GetBytes(x);
                        break;
                    }
                    br.BaseStream.Position += (size - 32 - off2);
                }
            }
            return true;
        }

        public static long getStringPointer(string str)
        {
            int length = 0;
            for (int i = 0; i < ScriptListForm.stringTable.Length; i++)
            {
                if (ScriptListForm.stringTable[i] == str)
                {
                    return length;
                }
                length += ScriptListForm.stringTable[i].Length + 1; //1 for null
            }

            BinaryWriter bw = Maps.map[0].BW;
            bw.BaseStream.Position = ScriptOffsetInfo.ScriptStringTableEnd;
            ScriptOffsetInfo.ScriptStringTableEnd += str.Length + 1;
            for (int i = 0; i < str.Length; i++)
            {
                bw.Write((byte)str[i]);
            }
            bw.Write((byte)0);
            string[] temp = ScriptListForm.stringTable;
            ScriptListForm.stringTable = new string[temp.Length + 1];
            temp.CopyTo(ScriptListForm.stringTable, 0);
            ScriptListForm.stringTable[temp.Length] = str;
            return length - 2;
        }

        public static void writeScript(int scriptNum, int lowestNumber, int highestNumber, int sizepadding, scriptInfo info)
        {
            BinaryReader br = Maps.map[0].BR;
            int count = highestNumber - lowestNumber + 1;
            int sizeDif = info.syntax.Count - (count * 20);
            BinaryWriter bw = Maps.map[0].BW;
            byte[] padding = new byte[] { 0, 0, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186 };
            if (sizeDif <= 0 || sizepadding >= 80 + sizeDif) //no difference, write it!
            {
                bw.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + lowestNumber * 20;
                byte[] data = new byte[info.syntax.Count];
                info.syntax.CopyTo(data);
                bw.Write(data); //we should be ready to write...

                bw.BaseStream.Position = ScriptOffsetInfo.ScriptsOffset + scriptNum * 40;
                bw.Write(info.header);
                ScriptListForm.scripts[scriptNum].ChunkNumber = (int)BitConverter.ToInt16(info.header, 36);
                ScriptListForm.scripts[scriptNum].ExpressionIndex = (uint)BitConverter.ToInt16(info.header, 38);
                string name = "";
                for (int i = 0; i < 32; i++)
                {
                    char a = (char)info.header[i];
                    if (a == 0)
                        break;
                    name += a.ToString();
                }
                ScriptListForm.scripts[scriptNum].Name = name;
                if (sizeDif < 0)
                {
                    for (int i = lowestNumber + info.syntax.Count / 20; i <= highestNumber; i++)
                    {
                        bw.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + i * 20;
                        bw.Write(padding);
                    }
                }
            }

            else
            {
                int startchunk = highestNumber + 1;
                int lastscript = ScriptListForm.scripts[ScriptListForm.scripts.Length - 1].ChunkNumber;
                int lastchunk = lastscript;
                TagMeta meta = new TagMeta();
                meta.ReadMetaFromMap(3, 0, false);
                br.BaseStream.Position = meta.offset + 568;
                int syntaxcount = br.ReadInt32();
                for (int i = lastscript; i < syntaxcount; i++)
                {
                    br.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + i * 20;
                    short expindex = br.ReadInt16();
                    if (expindex == 0)
                    {
                        lastchunk = i;
                        break;
                    }
                }
                int sizetoread = lastchunk - startchunk;
                for (int i = lastchunk; i >= startchunk; i--)
                {
                    br.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + i * 20;
                    byte[] chunk = br.ReadBytes(20);
                    ushort expindex = BitConverter.ToUInt16(chunk, 0);
                    if (expindex == 0) //padding
                    {
                        bw.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + i * 20 + sizeDif;
                        bw.Write(chunk); //move the padding, no updates
                        continue;
                    }
                    try
                    {
                        expindex += (ushort)(sizeDif / 20);
                    }
                    catch
                    {
                        expindex = (ushort)(expindex + sizeDif / 20 - 65536);
                    }
                    BitConverter.GetBytes(expindex).CopyTo(chunk, 0);
                    short siblingchunk = BitConverter.ToInt16(chunk, 8);
                    if (siblingchunk != -1)
                    {
                        siblingchunk += (short)(sizeDif / 20);
                        BitConverter.GetBytes(siblingchunk).CopyTo(chunk, 8);
                        ushort siblingexp = BitConverter.ToUInt16(chunk, 10);
                        try
                        {
                            siblingexp += (ushort)(sizeDif / 20);
                        }
                        catch
                        {
                            siblingexp = (ushort)(siblingexp + sizeDif / 20 - 65536);
                        }
                        BitConverter.GetBytes(siblingexp).CopyTo(chunk, 10);
                    }
                    short exptype = chunk[6];
                    if (exptype == 8 || exptype == 10) //child pointer
                    {
                        short childchunk = BitConverter.ToInt16(chunk, 16);
                        childchunk += (short)(sizeDif / 20);
                        BitConverter.GetBytes(childchunk).CopyTo(chunk, 16);
                        ushort childexp = BitConverter.ToUInt16(chunk, 10);
                        try
                        {
                            childexp += (ushort)(sizeDif / 20);
                        }
                        catch
                        {
                            childexp = (ushort)(childexp + sizeDif / 20 - 65536);
                        }
                        BitConverter.GetBytes(childexp).CopyTo(chunk, 18);
                    }
                    bw.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + i * 20 + sizeDif;
                    bw.Write(chunk); //chunk moved!
                }

                for (int i = scriptNum + 1; i < ScriptListForm.scripts.Length; i++)
                {
                    if (ScriptListForm.scripts[i].ChunkNumber < startchunk)
                        continue;

                    int loc = ScriptOffsetInfo.ScriptsOffset + i * 40 + 36;
                    bw.BaseStream.Position = loc;
                    int chNum = ScriptListForm.scripts[i].ChunkNumber + sizeDif / 20;
                    ScriptListForm.scripts[i].ChunkNumber = chNum;
                    ushort expIn = (ushort)ScriptListForm.scripts[i].ExpressionIndex;
                    try
                    {
                        expIn += (ushort)(sizeDif / 20);
                    }
                    catch
                    {
                        expIn -= (ushort)(expIn + sizeDif / 20 - 65536);
                    }
                    ScriptListForm.scripts[i].ExpressionIndex = (uint)expIn;
                    byte[] temp1 = BitConverter.GetBytes(chNum);
                    byte[] temp2 = BitConverter.GetBytes(expIn);
                    byte[] temp3 = new byte[] { temp1[0], temp1[1], temp2[0], temp2[1] };
                    bw.Write(temp3);
                }

                for (int i = 0; i < ScriptListForm.globals.Length; i++)
                {
                    if (ScriptListForm.globals[i].ChunkNumber < startchunk)
                        continue;

                    int loc = ScriptOffsetInfo.GlobalsStart + i * 40 + 36;
                    bw.BaseStream.Position = loc;
                    int chNum = ScriptListForm.globals[i].ChunkNumber + sizeDif / 20;
                    ScriptListForm.globals[i].ChunkNumber = chNum;
                    ushort expIn = (ushort)ScriptListForm.scripts[i].ExpressionIndex;
                    try
                    {
                        expIn += (ushort)(sizeDif / 20);
                    }
                    catch
                    {
                        expIn -= (ushort)(expIn + sizeDif / 20 - 65536);
                    }
                    ScriptListForm.globals[i].ExpressionIndex = (uint)expIn;
                    byte[] temp1 = BitConverter.GetBytes(chNum);
                    byte[] temp2 = BitConverter.GetBytes(expIn);
                    byte[] temp3 = new byte[] { temp1[0], temp1[1], temp2[0], temp2[1] };
                    bw.Write(temp3);
                }

                bw.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + lowestNumber * 20;
                byte[] data = new byte[info.syntax.Count];
                info.syntax.CopyTo(data);
                bw.Write(data);

                bw.BaseStream.Position = ScriptOffsetInfo.ScriptsOffset + scriptNum * 40;
                bw.Write(info.header);
                ScriptListForm.scripts[scriptNum].ChunkNumber = (int)BitConverter.ToInt16(info.header, 36);
                ScriptListForm.scripts[scriptNum].ExpressionIndex = (uint)BitConverter.ToInt16(info.header, 38);
                string name = "";
                for (int i = 0; i < 32; i++)
                {
                    char a = (char)info.header[i];
                    if (a == 0)
                        break;
                    name += a.ToString();
                }
                ScriptListForm.scripts[scriptNum].Name = name;
            }
        }

        public static int[] getsize(int scriptNum)
        {
            int lowestNumber = ScriptListForm.scripts[scriptNum].ChunkNumber; //begin's at the end...
            int highestNumber = lowestNumber;
            int curNumber = lowestNumber; //start here
            int openP = 0; //so we know when to stop
            int count = 0; //how big it was before
            List<int> parents = new List<int>();
            BinaryReader br = Maps.map[0].BR;
            while (true)
            {
                count++;
                if (curNumber < lowestNumber)
                    lowestNumber = curNumber;
                if (curNumber > highestNumber)
                    highestNumber = curNumber;
                br.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + curNumber * 20;
                br.BaseStream.Position += 6;
                int exptype = (int)br.ReadInt16();
                if (exptype == 8 || exptype == 10)
                {
                    openP++;
                    br.BaseStream.Position += 8;
                    parents.Add(curNumber);
                    curNumber = (int)br.ReadInt16();
                    continue;
                }
                int pointer = (int)br.ReadInt16();
                if (pointer == -1)
                {
                    openP--;
                    if (openP == 0) //finished!
                    {
                        break;
                    }
                    for (int i = parents.Count; i > 0; i--)
                    {
                        br.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + (parents[i - 1] * 20) + 8;
                        pointer = (int)br.ReadInt16();
                        if (pointer != -1)
                        {
                            curNumber = pointer;
                            break;
                        }
                        openP--;
                        parents.RemoveAt(i - 1);
                        if (openP == 0)
                        {
                            break;
                        }
                    }
                    if (openP == 0) //we need it 3 times... ugh
                    {
                        break;
                    }
                    parents.RemoveAt(parents.Count - 1);
                }
                else
                    curNumber = pointer;
            }
            int sizepadding = 0; //if you compile a small script, then change it to a larger one, it doesn't reshift
            for (int i = highestNumber + 1; true; i++)
            {
                br.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + i * 20;
                int expindex = (int)br.ReadInt16();
                if (expindex != 0)
                    break;
                sizepadding += 20;
            }

            int[] array = new int[] { lowestNumber, highestNumber, sizepadding };
            return array;
        }

        private void button1_Click(object sender, EventArgs e)
        {
            int[] size = getsize(scriptNum);
            scriptInfo info = comp(textBox1.Text, size[0]);
            if (info == null)
                return;

            if (info.type != type.Script)
            {
                MessageBox.Show("Don't compile globals here!");
                return;
            }

            writeScript(scriptNum, size[0], size[1], size[2], info);

            MessageBox.Show("Script Compiled!", "Success!");
        }
    }

    public enum type { Script, Global };

    public class scriptInfo
    {
        public List<byte> syntax;
        public byte[] header;
        public type type;
        public Script sinfo;
        public ScriptGlobal ginfo;

        public scriptInfo()
        {
            header = new byte[40];
            syntax = new List<byte>();
        }
    }
}