﻿using HaloMap;
using System;
using System.Collections.Generic;
using System.IO;

namespace ScriptTools
{
    public class ScriptExpression
    {
        public int Parent; //for compiling
        public List<int> Children;

        public int ChunkNumber;
        public uint ExpressionIndex;
        public int ExpressionType;
        public int ValueType;
        public int Identity;
        public ScriptPointer Sibling;
        public long StringOffset;
        public string String;

        public bool Bool;
        public byte Byte;
        public int Int;
        public long Long;
        public float Real;
        public ScriptPointer Child;
        public byte[] Hex;

        public ScriptDatabaseInfo DatabaseInfo;

        public ScriptExpression() //for compiling
        {
            Hex = new byte[4];
            Children = new List<int>();
        }

        public ScriptExpression(int chunk) //for decompiling
        {
            Map map = Maps.map[0];
            BinaryReader br = map.BR;

            int syntaxOffset = ScriptOffsetInfo.ScriptSyntaxDataStart + chunk * 20;
            br.BaseStream.Position = syntaxOffset;

            ChunkNumber = chunk;
            ExpressionIndex = br.ReadUInt16();
            Identity = br.ReadInt16();
            ValueType = br.ReadInt16();
            ExpressionType = br.ReadInt16();

            int siblingChunk = br.ReadInt16();
            if (siblingChunk > 0)
                Sibling = new ScriptPointer(siblingChunk);
            else
                Sibling = null;

            br.BaseStream.Position += 2;

            StringOffset = br.ReadInt32(); //hold on to this, we'll get it later
            if (ExpressionType != 9)
            {
                switch (ExpressionType)
                {
                    case 8:
                        Child = new ScriptPointer(br.ReadInt16());
                        break;
                    case 10:
                        if (ValueType != 2)
                            Child = new ScriptPointer(br.ReadInt16());
                        else
                            Int = br.ReadInt16();
                        break;
                    default:
                        Int = br.ReadInt16();
                        break;
                }
            }
            else
            {
                switch (ValueType)
                {
                    case 5:
                        Bool = br.ReadBoolean();
                        break;
                    case 6:
                        Real = br.ReadSingle();
                        break;
                    case 8:
                        Long = br.ReadInt32();
                        break;
                    case 9:
                        break;
                    case 11:
                        Int = br.ReadInt16();
                        break;
                    case 12:
                        Hex = new byte[4];
                        for (int i = 0; i < 4; i++)
                        {
                            Hex[i] = br.ReadByte();
                        }
                        break;
                    default:
                        Int = br.ReadInt16();
                        break;
                }
            }

            //get the string
            if (ExpressionType == 9 && StringOffset >= 0)
            {
                String = "";
                br.BaseStream.Position = (long)ScriptOffsetInfo.ScriptStringTableStart + StringOffset;
                Byte bt = new Byte();
                do
                {
                    bt = br.ReadByte();
                    if (bt == 0)
                        break;
                    String += BitConverter.ToChar(new byte[] { bt, 0 }, 0);
                } while (bt != 0);
            }
            else if (ExpressionType == 10)
            {
                br.BaseStream.Position = (long)ScriptOffsetInfo.ScriptsOffset + Int * 40;
                Byte bt = new Byte();
                do
                {
                    bt = br.ReadByte();
                    if (bt == 0)
                        break;
                    String += BitConverter.ToChar(new byte[] { bt, 0 }, 0);
                } while (bt != 0);
            }
            else if (ExpressionType == 13)
            {
                if (Int > 0)
                    br.BaseStream.Position = (long)ScriptOffsetInfo.GlobalsStart + Int * 40;
                else
                    br.BaseStream.Position = (long)ScriptOffsetInfo.ScriptStringTableStart + StringOffset;
                Byte bt = new Byte();
                do
                {
                    bt = br.ReadByte();
                    if (bt == 0)
                        break;
                    String += BitConverter.ToChar(new byte[] { bt, 0 }, 0);
                } while (bt != 0);
            }

            if (ExpressionType == 8) //syntax chunk
            {
                DatabaseInfo = new ScriptDatabaseInfo();
                DatabaseInfo.Identity = Identity;
                DatabaseInfo.ReturnType = ValueType;
            }
        }
    }
}