using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.IO;
using HaloMap;
using System.Net;

namespace ScriptTools
{
    public partial class ScriptListForm : Form
    {
        public Map map;

        public static Script[] scripts;
        public static ScriptGlobal[] globals;
        public static string[] stringTable;
        public static int SortType;

        public ScriptListForm()
        {
            InitializeComponent();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DialogResult result = openMap.ShowDialog();
            if (result != DialogResult.OK)
                return;
            if (map != null && map.isOpen)
            {
                map.CloseMap();
                Maps.Remove(0);
            }
            string filename = openMap.FileName;
            int mapnum = Maps.Add(filename);
            Maps.map[mapnum].OpenMap(MapTypes.Internal);
            map = Maps.map[mapnum];
            Text = "Script Tools - " + filename;
            TagMeta meta = new TagMeta();
            meta.ReadMetaFromMap(3, mapnum, false);
            if (meta.type != "scnr")
            {
                map.CloseMap();
                Maps.Remove(0);
                MessageBox.Show("Invalid map!");
                return;
            }
            listScriptNames.Items.Clear();

            BinaryReader br = map.BR;
            br.BaseStream.Position = meta.offset + 436;
            ScriptOffsetInfo.ScriptStringTableStart = (int)br.ReadUInt32() - map.SecondaryMagic;

            br.BaseStream.Position = ScriptOffsetInfo.ScriptStringTableStart;
            string strtable = "";
            while (true)
            {
                Byte bt = br.ReadByte();
                if (bt == 205)
                {
                    break;
                }
                strtable += BitConverter.ToChar(new byte[] { bt, 0 }, 0);
            }
            stringTable = strtable.Split(new char[] { (char)0 });

            ScriptOffsetInfo.ScriptStringTableEnd = (int)(br.BaseStream.Position - 1);

            br.BaseStream.Position = meta.offset + 440;
            scripts = new Script[br.ReadInt32()];
            ScriptOffsetInfo.ScriptsOffset = br.ReadInt32() - map.SecondaryMagic;

            br.BaseStream.Position = meta.offset + 448;
            globals = new ScriptGlobal[br.ReadInt32()];
            ScriptOffsetInfo.GlobalsStart = br.ReadInt32() - map.SecondaryMagic;

            br.BaseStream.Position = meta.offset + 572;
            ScriptOffsetInfo.ScriptSyntaxDataStart = br.ReadInt32() - map.SecondaryMagic;

            try
            {
                br.BaseStream.Position = ScriptOffsetInfo.ScriptsOffset;

                for (int i = 0; i < scripts.Length; i++)
                {
                    var script = scripts[i] = new Script();

                    for (int j = 0; j < 32; j++)
                    {
                        var character = br.ReadChar();

                        if (character == '\0')
                        {
                            br.BaseStream.Position += (31 - j);
                            break; // null separator
                        }

                        script.Name += character;
                    }

                    script.Type = br.ReadInt16();
                    script.ReturnType = br.ReadInt16();
                    script.ChunkNumber = br.ReadInt16();
                    script.ExpressionIndex = br.ReadUInt16();

                    listScriptNames.Items.Add(script.Name + " - " + ScriptTypes.Names[script.Type]);
                }
            }
            catch
            {
                MessageBox.Show("Error! No scripts!");
                Text = "Script Tools";
                map.CloseMap();
                Maps.Remove(0);
            }
            try
            {
                br.BaseStream.Position = ScriptOffsetInfo.GlobalsStart;
                for (int i = 0; i < globals.Length; i++)
                {
                    globals[i] = new ScriptGlobal();
                    for (int j = 0; j < 32; j++)
                    {
                        Byte bt = br.ReadByte();
                        if (bt == 0)
                        {
                            br.BaseStream.Position += (31 - j);
                            break; //null separator
                        }
                        globals[i].Name += BitConverter.ToChar(new byte[] { bt, 0 }, 0);
                    }
                    globals[i].Type = br.ReadInt16();
                    br.ReadInt16();
                    globals[i].ChunkNumber = br.ReadInt16();
                    globals[i].ExpressionIndex = br.ReadUInt16();
                }
            }
            catch
            {
                //we don't care... just here to avoid errors
            }
        }

        private void closeMapToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (map == null)
                return;
            if (map.isOpen)
                map.CloseMap();
            Text = "Script Tools";
            Maps.map.RemoveAt(0);
            map = null;
            listScriptNames.Items.Clear();
        }

        private void listBoxRightClick(object sender, MouseEventArgs e)
        {
            if (e.Button != MouseButtons.Right)
                return;

            ListBox list = (ListBox)sender;
            if (list.SelectedIndex < 0)
                return;

            System.Windows.Forms.ContextMenuStrip cmenu = contextMenuStrip1;
            cmenu.Show((Control)sender, e.X, e.Y);

            Script sinfo = scripts[list.SelectedIndex];
            ToolStripMenuItem item = (ToolStripMenuItem)typeToolStripMenuItem.DropDownItems[sinfo.Type];
            item.Checked = true;
        }

        private void contextMenuStrip1_Closed(object sender, ToolStripDropDownClosedEventArgs e)
        {
            foreach (ToolStripMenuItem item in typeToolStripMenuItem.DropDownItems)
            {
                item.Checked = false;
            }
        }

        private void decompileSelected(object sender, EventArgs e)
        {
            if (listScriptNames.SelectedIndex < 0)
                return;

            ScriptEditorForm form = new ScriptEditorForm();
            form.Show();

            form.decompile(scripts[listScriptNames.SelectedIndex]);
            form.scriptNum = listScriptNames.SelectedIndex;
        }

        private void decompileAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (Maps.map.Count == 0) //open a map!
                return;

            DialogResult result = saveFileDialog1.ShowDialog();
            if (result != DialogResult.OK)
                return;

            StreamWriter sw = File.CreateText(saveFileDialog1.FileName);

            foreach (ScriptGlobal gbl in globals)
            {
                sw.WriteLine(ScriptEditorForm.DecompileGlobal(gbl));
            }
            sw.WriteLine("\r\n");
            foreach (Script sinfo in scripts)
            {
                sw.WriteLine(ScriptEditorForm.Decompile(sinfo));
                sw.Write("\r\n");
            }

            var dbFile = new FileInfo(Path.Combine(Application.StartupPath, "db.txt"));

            if (dbFile.Exists)
                dbFile.Delete();

            ScriptDatabase.Save(dbFile);
            ScriptDatabaseInfo.Save(dbFile);

            sw.Close();
        }

        private void List_Load(object sender, EventArgs e)
        {
            WebClient wc = new WebClient();
            StreamReader sr = null;
            string total;
            try
            {
                total = wc.DownloadString("http://jnatalie.no-ip.info/scripts/commands.php?app");
            }
           catch
            {
                sr = File.OpenText(Application.StartupPath + "\\db.txt");
                sr.BaseStream.Position = (long)0;
                total = sr.ReadToEnd();
            }
            int pos1 = 0;
            int pos2 = total.IndexOf('\n');
            string str;
            int identity;
            int returnType;
            string name;
            List<int> parameters;
            while (pos2 >= 0)
            {
                str = total.Substring(pos1, pos2 - pos1);
                string[] split = str.Split(new char[] { '\t' }, StringSplitOptions.RemoveEmptyEntries);
                if (split.Length < 3)
                {
                    break;
                }
                identity = int.Parse(split[0]);
                returnType = ScriptValueTypes.GetIndex(split[1]);
                name = split[2];
                parameters = new List<int>();
                for (int i = 3; i < split.Length; i++)
                {
                    parameters.Add(ScriptValueTypes.GetIndex(split[i]));
                }
                ScriptDatabaseInfo info = new ScriptDatabaseInfo();
                info.Identity = identity;
                info.Name = name;
                info.ReturnType = returnType;
                info.Parameters = parameters;
                ScriptDatabaseInfo.Add(info);
                pos1 = pos2 + 1;
                pos2 = total.IndexOf('\n', pos1);
            }
            if (sr != null)
                sr.Close();
            else
                ScriptDatabaseInfo.Save(new FileInfo(Path.Combine(Application.StartupPath, "db.txt"))); //output the downloaded commands
        }

        private void plaintextCompile(object sender, EventArgs e)
        {
            if (listScriptNames.SelectedIndex < 0)
                return;

            ScriptEditorForm form = new ScriptEditorForm();
            form.Show();

            form.scriptNum = listScriptNames.SelectedIndex;
        }

        private void compileAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (Maps.map.Count == 0) //open a map!
                return;

            DialogResult result = openScript.ShowDialog();
            if (result != DialogResult.OK)
                return;

            StreamReader sr = File.OpenText(openScript.FileName);
            string file = sr.ReadToEnd();
            sr.Close();
            int pos1 = 0;
            int pos2 = 0;
            int startchunk = 4;
            List<scriptInfo> sinfo = new List<scriptInfo>();
            List<scriptInfo> ginfo = new List<scriptInfo>();

            Script[] scriptbackup = scripts; //in-case...
            ScriptGlobal[] globalbackup = globals;

            while (true) //inf loop1, each script
            {
                string script = "";
                pos1 = file.IndexOf('(', pos2);
                if (pos1 == -1 || pos1 >= file.Length)
                    break;
                int pcount = 1;
                pos2 = pos1;
                while (true) //inf loop2, each chunk
                {
                    pos2 = file.IndexOfAny(new char[] { '(', ')' }, pos2 + 1);
                    if (file.Substring(pos2, 1) == "(") //add another pcount
                    {
                        pcount++;
                    }
                    else //take one away
                    {
                        pcount--;
                        if (pcount == 0)
                            break;
                    }
                }
                script = file.Substring(pos1, pos2 - pos1 + 1);
                scriptInfo info = ScriptEditorForm.comp(script, startchunk);
                if (info == null)
                {
                    scripts = scriptbackup;
                    globals = globalbackup;
                    return;
                }
                if (info.type == type.Script)
                {
                    if (sinfo.Count > scripts.Length)
                    {
                        MessageBox.Show("Add more script chunks!");
                        scripts = scriptbackup;
                        globals = globalbackup;
                        return;
                    }
                    scripts[sinfo.Count] = info.sinfo; //for static references
                    sinfo.Add(info);
                }
                else
                {
                    if (ginfo.Count > globals.Length)
                    {
                        MessageBox.Show("Add more global chunks!");
                        globals = globalbackup;
                        return;
                    }
                    globals[ginfo.Count] = info.ginfo; //for references later on
                    ginfo.Add(info);
                }
                startchunk += info.syntax.Count / 20 + 4;
                pos2++;
            }


            byte[] padding = new byte[] { 0, 0, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186, 186 };
            BinaryWriter bw = Maps.map[0].BW;
            int lastchunk = 0;
            for (int i = 0; i < ginfo.Count; i++)
            {
                int start = globals[i].ChunkNumber - 4;
                bw.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + start * 20;
                for (int x = 0; x < 4; x++)
                {
                    bw.Write(padding);
                }
                byte[] data = new byte[ginfo[i].syntax.Count];
                ginfo[i].syntax.CopyTo(data);
                bw.Write(data);
                int end = (int)(bw.BaseStream.Position - ScriptOffsetInfo.ScriptSyntaxDataStart) / 20;
                if (end > lastchunk)
                    lastchunk = end;

                bw.BaseStream.Position = ScriptOffsetInfo.GlobalsStart + i * 40;
                bw.Write(ginfo[i].header);
            }

            for (int i = 0; i < sinfo.Count; i++)
            {
                int start = scripts[i].ChunkNumber - 4;
                bw.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + start * 20;
                for (int x = 0; x < 4; x++)
                {
                    bw.Write(padding);
                }
                byte[] data = new byte[sinfo[i].syntax.Count];
                sinfo[i].syntax.CopyTo(data);
                bw.Write(data);
                int end = (int)(bw.BaseStream.Position - ScriptOffsetInfo.ScriptSyntaxDataStart) / 20;
                if (end > lastchunk)
                    lastchunk = end;

                bw.BaseStream.Position = ScriptOffsetInfo.ScriptsOffset + i * 40;
                bw.Write(sinfo[i].header);
            }

            BinaryReader br = Maps.map[0].BR;
            TagMeta meta = new TagMeta();
            meta.ReadMetaFromMap(3, 0, false);
            br.BaseStream.Position = meta.offset + 568;
            int syntaxCount = br.ReadInt32();
            for (int i = lastchunk; i < syntaxCount; i++)
            {
                bw.BaseStream.Position = ScriptOffsetInfo.ScriptSyntaxDataStart + i * 20;
                bw.Write(padding);
            }
        }

        private void constructorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (listScriptNames.SelectedIndex < 0)
                return;

            Constructor form = new Constructor();
            form.toDecompile = listScriptNames.SelectedIndex;
            form.Show();
        }

        private void typeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int x = 0;
            ToolStripMenuItem item = typeToolStripMenuItem;
            for (int i = 0; i < 6; i++)
            {
                if (item.DropDownItems[i] == sender)
                {
                    x = i;
                    break;
                }
            }

            ScriptListForm.scripts[listScriptNames.SelectedIndex].Type = x;
            BinaryWriter bw = Maps.map[0].BW;
            bw.BaseStream.Position = ScriptOffsetInfo.ScriptsOffset + listScriptNames.SelectedIndex * 40 + 32;
            bw.Write((byte)x);
            listScriptNames.Items[listScriptNames.SelectedIndex] = ScriptListForm.scripts[listScriptNames.SelectedIndex].Name + " - " + ScriptTypes.Names[x];
        }

        private void constructorToolStripMenuItem4_Click(object sender, EventArgs e)
        {
            if (listScriptNames.SelectedIndex < 0)
                return;

            Constructor form = new Constructor();
            form.Show();

            form.scriptNumber = listScriptNames.SelectedIndex;
        }

        private void globalsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (map == null)
                return;
            if (ScriptListForm.globals.Length <= 0)
            {
                MessageBox.Show("No globals!");
                return;
            }
            ScriptGlobalEditorForm form = new ScriptGlobalEditorForm();
            form.Show();
        }

        private void orderToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (Maps.map.Count <= 0)
                return;

            int x = 0;
            ToolStripMenuItem item = sortToolStripMenuItem;
            for (int i = 0; i < 4; i++)
            {
                if (item.DropDownItems[i] == sender)
                {
                    x = i;
                    ((ToolStripMenuItem)item.DropDownItems[i]).Checked = true;
                }
                else
                    ((ToolStripMenuItem)item.DropDownItems[i]).Checked = false;
            }
            SortType = x;
            List<Script> tmp = new List<Script>();
            for (int i = 0; i < scripts.Length; i++)
            {
                tmp.Add(scripts[i]);
            }
            tmp.Sort(new ScriptListComparer());
            tmp.CopyTo(scripts);
            reprintList();
        }

        public void reprintList()
        {
            listScriptNames.Items.Clear();

            for (int i = 0; i < scripts.Length; i++)
            {
                var script = scripts[i];

                listScriptNames.Items.Add(script.Name + " - " + ScriptValueTypes.Names[script.Type]);
            }
        }

        

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutBox1 about = new AboutBox1();
            about.ShowDialog();
        }

        private void addScriptChunkToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (Maps.map.Count <= 0)
                return;

            int x = 0;
            ToolStripMenuItem item = sortToolStripMenuItem;
            for (int i = 0; i < 4; i++)
            {
                if (item.DropDownItems[i] == orderToolStripMenuItem)
                {
                    x = i;
                    ((ToolStripMenuItem)item.DropDownItems[i]).Checked = true;
                }
                else
                    ((ToolStripMenuItem)item.DropDownItems[i]).Checked = false;
            }
            SortType = x;
            List<Script> tmp = new List<Script>();
            for (int i = 0; i < scripts.Length; i++)
            {
                tmp.Add(scripts[i]);
            }
            tmp.Sort(new ScriptListComparer());
            tmp.CopyTo(scripts);
            reprintList();

            int syntaxloc = scripts[scripts.Length - 1].ChunkNumber; ;
            if (globals[globals.Length - 1].ChunkNumber > syntaxloc)
                syntaxloc = globals[globals.Length - 1].ChunkNumber;
            ScriptChunkAdder.CreateScript(type.Script, "blank", syntaxloc);
        }
}
}