using System;
using System.Windows.Forms;
using System.ComponentModel;
using System.Collections;

namespace HaloMap
{
	/// <summary>
	/// Summary description for MetaFunctions.
	/// </summary>
	public class MetaFunctions
	{
		#region Constructor
		public MetaFunctions()
		{
			
		}
		#endregion
		#region Methods
		public RawDataContainer ReadRaw(int tagnumber,int mapnumber,bool dontreadraw)
		{
			RawDataContainer tempcontainer=new RawDataContainer();
			
			switch(CheckForRaw(Maps.map[mapnumber].MetaInfo.TagType[tagnumber]))
			{
				case RawDataContainerType.PRTM:
					tempcontainer=new PRTM();
					break;
				case RawDataContainerType.DECR:
					tempcontainer=new DECR();
					break;
				case RawDataContainerType.Animation:
					tempcontainer=new Animation();
					break;
				case RawDataContainerType.Model:
					tempcontainer=new Model();
					break;
				case RawDataContainerType.Bitmap:
					tempcontainer=new BitmapRaw();
					break;
				case RawDataContainerType.BSP:
					tempcontainer=new BSPRaw();
					break;

				case RawDataContainerType.Weather:
					tempcontainer=new Weather();
					break;
                case RawDataContainerType.Sound:
                    tempcontainer = new Sound();
                   break;
               case RawDataContainerType.CoconutsModel:
                   tempcontainer = new  UghRawContainer();
                   break;
               case RawDataContainerType.LightMap:
                   tempcontainer = new LightmapRaw();
                   break;
				default:
					break;
			}
			tempcontainer.Read(tagnumber,mapnumber,dontreadraw);
			return tempcontainer;
			
		}

		public int FindByNameAndTagType(string type,string name,int mapnumber)
		{
            for (int x = 0; x < Maps.map[mapnumber].IndexHeader.metaCount; x++)
			{
                if (type == Maps.map[mapnumber].MetaInfo.TagType[x] && name == Maps.map[mapnumber].FileNames.Name[x])
				{
					return x;
				}
			}
			return -1;
		}

		public int FindMetaByOffset(int offset,int mapnumber)
		{
            for (int x = 0; x < Maps.map[mapnumber].IndexHeader.metaCount; x++)
			{
                if (Maps.map[mapnumber].HaloVersion == Map.HaloVersionEnum.HaloCE) if (Maps.map[mapnumber].MetaInfo.external[x] == true) continue;
                if (offset >= Maps.map[mapnumber].MetaInfo.Offset[x] && offset < Maps.map[mapnumber].MetaInfo.Offset[x] + Maps.map[mapnumber].MetaInfo.Size[x])
				{
              
					return x;
				}
			}
			return -1;
		}

		public int FindMetaByID(int id,int mapnumber)
		{
            int xx = 0;
            object xxx = Maps.map[mapnumber].MetaInfo.identHT[id];
           if (xxx == null) { return -1; }
            xx =(int)xxx;
            return xx;
           
		
		}

		public RawDataContainerType CheckForRaw(string type)
		{
			switch (type)
			{
				case "PRTM":
					return RawDataContainerType.PRTM;
				case "DECR":
					return RawDataContainerType.DECR;
                case "ltmp":
                    return RawDataContainerType.LightMap;
                case "mod2":
                    return RawDataContainerType.Model;
				case "mode":
					return RawDataContainerType.Model;
				case "bitm":
					return RawDataContainerType.Bitmap;
				case "jmad":
					return RawDataContainerType.Animation;
				case "sbsp":
					return RawDataContainerType.BSP;
				case "weat":
					return RawDataContainerType.Weather;
                case "snd!":
                    return RawDataContainerType.Sound;
                case "ugh!":
                    return RawDataContainerType.CoconutsModel;
				default:
					return RawDataContainerType.Empty;
			}
		}
		
		#endregion
	}
}
