using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.ComponentModel;

namespace HaloMap
{

    public class TagInterface 
    {
        private TagBlock header;


        public TagBlock Header
        {
            get { return header;}
            set
            {

                header=value  ;
                value.ChunkCount = 1;
                value.MapOffset = Offset;
               // value.Translation.MapTranslation = Offset;
            }
        }
   
        private static int Ident;
        private static int TagNumber;
        private static string TagName;
        private static string TagType;
        private static int Offset;
        private static int MapNumber;
        private static int Magic;
        private static int BSPNumber=-1;
        public TagInterface(int tagnumber,int mapnumber)
        {
            Header = new TagBlock("Header",null,Maps.map[mapnumber].HaloVersion);
            SeekInMeta(0);
            MapNumber = mapnumber;
            TagNumber = tagnumber;
            Ident = Maps.map[MapNumber].MetaInfo.Ident[TagNumber];
            TagType = Maps.map[MapNumber].MetaInfo.TagType[TagNumber];
            TagName = Maps.map[MapNumber].FileNames.Name[TagNumber];
            Offset=Maps.map[MapNumber].MetaInfo.Offset[TagNumber];
            if (TagType == "sbsp")
            {
                BSPNumber = Maps.map[MapNumber].BSP.FindBSPNumberByBSPIdent(Ident);
                Magic=Maps.map[MapNumber].BSP.sbsp[BSPNumber].magic;
            }
            else
            {
                switch (Maps.map[MapNumber].HaloVersion)
                {
                    case Map.HaloVersionEnum.Halo2:
                        Magic = Maps.map[MapNumber].SecondaryMagic;
                        break;
                    case Map.HaloVersionEnum.Halo1:
                    case Map.HaloVersionEnum.HaloCE:
                        Magic = Maps.map[MapNumber].PrimaryMagic;
                        break;
                }
            }
        }

        public void SeekInMeta(int offset)
        {
            Maps.map[MapNumber].BR.BaseStream.Position=Offset+offset;
        }
        public void SeekInMap(int offset)
        {
            Maps.map[MapNumber].BR.BaseStream.Position=offset;
              
        }


        #region MetaObjects
        #region
        public enum MetaObjectType
        {
        }
        #endregion
        [TypeConverter(typeof(ExpandableObjectConverter))]
        public class BaseObject
        {
            internal TagBlock parent;
            internal string name;
            internal int offset;
            
            internal int mapOffset;
            internal int size;
            #region Properties
            [Category("Base Properties")]
            public string Name
            {
                get { return name; }
               // set { name = value; }
            }
            [Category("Base Properties")]
            public int Offset
            {
                get { return offset; }
               // set { offset = value; }
            }
            [Category("Base Properties")]
            public int MapOffset
            {
                get { return mapOffset; }
                set { mapOffset = value; }
            }
            [Category("Base Properties")]
            public int Size
            {
                get { return size; }
             //   set { size = value; }
            }
            [Category("Base Properties")]
            [TypeConverter(typeof(BaseObjectConverter))]
            public TagBlock Parent
            {
                get { return parent; }
            }

          
            #endregion
            //public int ParentIndex = -1;
            //public int SiblingIndex = -1;
            //public int ChildIndex = -1;
            public BaseObject(string name)
            {
                
                if (name != null) this.name = name;
            }

            public virtual void Read()
            {
            }
            public virtual void Write()
            {
            }
        }
        #region Bitmasks
        public class BitMaskFlags : BaseObject
        {
            public int BitCount;
            public BitMaskFlags(string name)
                : base(name)
            {
                
            }
        }

        public class ByteFlags : BitMaskFlags
        {
            public byte Value;
            public ByteFlags(string name)
                : base(name)
            {
                this.BitCount = 8;
                this.size = 1;
            }
            public override void Read()
            {
                Value = Maps.map[MapNumber].BR.ReadByte();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(Value);
            }
        }
        public class WordFlags : BitMaskFlags
        {
            public short Value;
            public WordFlags(string name)
                : base(name)
            {
                this.BitCount = 16;
                this.size = 2;
            }
            public override void Read()
            {
                Value = Maps.map[MapNumber].BR.ReadInt16();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(Value);
            }

        
        }
        #endregion
        public class RealFloat : BaseObject
        {
            public float Value;
            public RealFloat(string name)
                : base(name)
            {
                this.size = 4;
            }

            public override void Read()
            {
                Value = Maps.map[MapNumber].BR.ReadSingle();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(Value);
            }

             [Category("Values")]
            public float Float
            {
                get { return Value; }
                set { Value = value; }
            }
        }

        public class LongInteger : BaseObject
        {
            public int Value;
            public LongInteger(string name)
                : base(name)
            {
                this.size = 4;
            }

            public override void Read()
            {
                Value = Maps.map[MapNumber].BR.ReadInt32();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(Value);
            }

            [Category("Values")]
            public int Long
            {
                get { return Value; }
                set { Value = value; }
            }
        }
        public class ShortInteger : BaseObject
        {

            public short Value;
            public ShortInteger(string name)
                : base(name)
            {
                this.size = 2;
            }

            public override void Read()
            {
                Value = Maps.map[MapNumber].BR.ReadInt16();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(Value);
            }

            [Category("Values")]
            public short Short
            {
                get { return Value; }
                set { Value = value; }
            }
        }
        public class ByteInteger : BaseObject
        {

            public short Value;
            public ByteInteger(string name)
                : base(name)
            {
                this.size = 1;
            }

            public override void Read()
            {
                Value = Maps.map[MapNumber].BR.ReadByte();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(Value);
            }
        }

        #region Colors
        public class RealRGBColor : BaseObject
        {

            private float r;
            private float g;
            private float b;
            private System.Drawing.Color color;
            public RealRGBColor(string name)
                : base(name)
            {
                this.size = 12;
            }

            public override void Read()
            {
                r = Maps.map[MapNumber].BR.ReadSingle();
                g = Maps.map[MapNumber].BR.ReadSingle();
                b = Maps.map[MapNumber].BR.ReadSingle();
                UpdateColor();
            }
            public override void Write()
            {

                Maps.map[MapNumber].BW.Write(r);
                Maps.map[MapNumber].BW.Write(g);
                Maps.map[MapNumber].BW.Write(b);
            }
            #region Properties
            [Category("Values")]
            public float R
            {
                get { return r; }
                set { UpdateColor(); r = value; }
            }

            [Category("Values")]
            public float G
            {
                get { return g; }
                set { UpdateColor(); g = value; }
            }

            [Category("Values")]
            public float B
            {
                get { return b ; }
                set { UpdateColor(); b = value; }
            }

            [Category("Values")]
            public System.Drawing.Color Color
            {
                get { return color; }
                set { ConvertColor(); color = value; }
            }
            #endregion
            private void UpdateColor()
            {
                if (r > 1) r = 1;
                if (g > 1) g = 1;
                if (b > 1) b = 1;
                if (r < 0) r = 0;
                if (g < 0) g = 0;
                if (b < 0) b = 0;
                color = System.Drawing.Color.FromArgb(255, (int)(r * 255), (int)(g * 255), (int)(b * 255));
            }

            private void ConvertColor()
            {
                r = (float)color.R / (float)255;
                g = (float)color.G / (float)255;
                b = (float)color.B / (float)255;
            }
        }
        public class RealARGBColor : BaseObject
        {
            public float a;
            public float r;
            public float g;
            public float b;
            private System.Drawing.Color color;
            public RealARGBColor(string name)
                : base(name)
            {
                this.size = 16;
            }

            public override void Read()
            {
                a = Maps.map[MapNumber].BR.ReadSingle();
                r = Maps.map[MapNumber].BR.ReadSingle();
                g = Maps.map[MapNumber].BR.ReadSingle();
                b = Maps.map[MapNumber].BR.ReadSingle();
                UpdateColor();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(a);
                Maps.map[MapNumber].BW.Write(r);
                Maps.map[MapNumber].BW.Write(g);
                Maps.map[MapNumber].BW.Write(b);
            }
            #region Properties
            [Category("Values")]
            public float A
            {
                get { return a; }
                set { UpdateColor(); a = value; }
            }
            [Category("Values")]
            public float R
            {
                get { return r; }
                set { UpdateColor(); r = value; }
            }

            [Category("Values")]
            public float G
            {
                get { return g; }
                set { UpdateColor(); g = value; }
            }

            [Category("Values")]
            public float B
            {
                get { return b; }
                set { UpdateColor(); b = value; }
            }

            [Category("Values")]
            public System.Drawing.Color Color
            {
                get { return color; }
                set { ConvertColor(); color = value; }
            }
            #endregion
            private void UpdateColor()
            {
                if (r > 1) r = 1;
                if (g > 1) g = 1;
                if (b > 1) b = 1;
                if (a > 1) a = 1;
                if (r < 0) r = 0;
                if (g < 0) g = 0;
                if (b < 0) b = 0;
                if (a < 0) b = 0;
                color = System.Drawing.Color.FromArgb((int)(a * 255), (int)(r * 255), (int)(g * 255), (int)(b * 255));
            }

            private void ConvertColor()
            {
                a = (float)color.A / (float)255;
                r = (float)color.R / (float)255;
                g = (float)color.G / (float)255;
                b = (float)color.B / (float)255;
            }
        }
        #endregion

        public class RealPlane3D : BaseObject
        {
            private float i;
            private float j;
            private float k;
            private float d;
            public RealPlane3D(string name)
                : base(name)
            {
                this.size = 16;
            }


            public override void Read()
            {
                i = Maps.map[MapNumber].BR.ReadSingle();
                j = Maps.map[MapNumber].BR.ReadSingle();
                k = Maps.map[MapNumber].BR.ReadSingle();
                d = Maps.map[MapNumber].BR.ReadSingle();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(i);
                Maps.map[MapNumber].BW.Write(j);
                Maps.map[MapNumber].BW.Write(k);
                Maps.map[MapNumber].BW.Write(d);
            }

            [Category("Values")]
            public float I
            {
                get { return i; }
                set { i = value; }
            }

            [Category("Values")]
            public float J
            {
                get { return j; }
                set { j = value; }
            }

            [Category("Values")]
            public float K
            {
                get { return k; }
                set { k = value; }
            }

            [Category("Values")]
            public float D
            {
                get { return d; }
                set { d = value; }
            }
        }



        public class RealPoint3D : BaseObject
        {

            public float x;
            public float y;
            public float z;
            public RealPoint3D(string name)
                : base(name)
            {
                this.size = 12;
            }

            public override void Read()
            {
                x = Maps.map[MapNumber].BR.ReadSingle();
                y = Maps.map[MapNumber].BR.ReadSingle();
                z = Maps.map[MapNumber].BR.ReadSingle();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(x);
                Maps.map[MapNumber].BW.Write(y);
                Maps.map[MapNumber].BW.Write(z);
            }
        }

        public class Angle : BaseObject
        {
            public float Value;
            public Angle(string name)
                : base(name)
            {
                this.size = 4;
            }

            public float ConvertToRadian(float angle)
            {
                double dbl;

                if (angle != 0F)
                {
                    dbl = 0.017453292519943295 * ((double)(angle));
                    return ((float)(dbl));
                }

                return 0F;
            }

            public float ConvertFromRadian(float angle)
            {
                return (float)((angle * 180.0) / Math.PI);
            }

            public override void Read()
            {
                Value = Maps.map[MapNumber].BR.ReadSingle();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(ConvertToRadian(Value));
            }
        }

        public class Enumerator : BaseObject
        {

            public short Value;

            public Enumerator(string name)
                : base(name)
            {
                this.size = 2;
            }

            public override void Read()
            {
                Value = Maps.map[MapNumber].BR.ReadInt16();
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(Value);
            }
        }


        public class RealEulerAngle3D : BaseObject
        {

            public Angle y;
            public Angle p;
            public Angle r;
            public RealEulerAngle3D(string name)
                : base(name)
            {
                y = new Angle(name);
                p = new Angle(name);
                r = new Angle(name);
                this.size = 12;
            }


            public override void Read()
            {
                y.Read();
                p.Read();
                r.Read();
            }
            public override void Write()
            {
                y.Write();
                p.Write();
                r.Write();
            }
        }
        [TypeConverter(typeof(ExpandableObjectConverter))]
        public class Pointer : BaseObject
        {
            
            private int mapTranslation;
            private int metaTranslation;
         
            #region Properties
           
            public int MapTranslation
            {
                get { return mapTranslation; }
            }
            public int MetaTranslation
            {
                get { return metaTranslation; }
            }
            #endregion
            public Pointer(string name)
                : base(name)
            {
                
                this.size = 4;
            }

            public override void Read()
            {


                mapTranslation = Maps.map[MapNumber].BR.ReadInt32() - Magic;
                metaTranslation = MapTranslation - TagInterface.Offset;

            }
          //  public void Write()
         //   {
           //     int temp = metaoffset + MetaTranslation + Magic;
           //     Maps.map[MapNumber].BW.Write(temp);
            //}
            public override void Write()
            {
                int temp = MapTranslation + Magic;
                Maps.map[MapNumber].BW.Write(temp);
            }
        }
        public class H1TagReference : BaseObject
        {
            public HaloTagType TagType;
            public Pointer StringTranslation;
            public int Zero = 0;
            public TagReference Ident;
            public H1TagReference(string name)
                : base(name)
            {
                TagType = new HaloTagType(name);
                StringTranslation = new Pointer(name);
                this.size = 16;
            }

            public override void Read()
            {
                TagType.Read();
                StringTranslation.Read();
                Zero = Maps.map[MapNumber].BR.ReadInt32();
                Ident.Read();

            }
            public override void Write()
            {
                TagType.Write();
                StringTranslation.Write();
                Maps.map[MapNumber].BW.Write(Zero);
                Ident.Write();
            }
        }



        public class TagReference : BaseObject
        {
            public int Ident;
            public int TagNumber;
            public string TagName;
            public TagReference(string name)
                : base(name)
            {
                this.size = 4;
            }

            public override void Read()
            {
                Ident = Maps.map[MapNumber].BR.ReadInt32();
                TagNumber = Maps.map[MapNumber].Functions.Meta.FindMetaByID(Ident, MapNumber);
            }
            public override void Write()
            {
                Maps.map[MapNumber].BW.Write(Ident);
            }
        }

        public class HaloTagType : BaseObject
        {
            public string TagType;
            public HaloTagType(string name)
                : base(name)
            {
                this.size = 4;
            }

            public override void Read()
            {
                char[] temp = Maps.map[MapNumber].BR.ReadChars(4);
                Array.Reverse(temp);
                TagType = new string(temp);
            }
            public override void Write()
            {
                char[] temp = TagType.ToCharArray();
                Array.Reverse(temp);
                Maps.map[MapNumber].BW.Write(temp);
            }
        }

        [TypeConverter(typeof(ExpandableObjectConverter))]
        public class TagBlockLayout : List<BaseObject>
        {
            public int ChunkSize;
            public TagBlock Parent;
            public new BaseObject this[int indexer]
            {
            get
            {
                return base[indexer];
            }
            }
            public new void Add(BaseObject obj)
            {
                obj.offset = this.ChunkSize;
                obj.parent = Parent;
                base.Add(obj);
                this.ChunkSize += obj.Size;
            }

            public new void RemoveAt(int index)
            {
                this.ChunkSize -= base[index].Size;
                base.RemoveAt(index);

            }

            public void ReadFields()
            {
              
                foreach (BaseObject temp in this)
                {
                    temp.Read();
                    temp.parent = this.Parent;
                }
                foreach (BaseObject temp in this)
                {
                    if (temp is TagBlock) 
                        ((TagBlock)temp).ReadElements() ;
                }
            }
        }

        
        public class ChunkCollection : List<TagBlockLayout>
        {
            
            public new TagBlockLayout this[int indexer]
            {
                get
                {
                    return base[indexer];
                }
                set
                {
                     base[indexer]=value;
                }
            }
     
        }

        [DefaultPropertyAttribute("Name")]
       
        public class TagBlock : BaseObject
        {
            private int chunknumber;
            private int chunkCount;
            TagBlockLayout Layout=new TagBlockLayout();
            private Pointer translation = new Pointer("Translation");
            Map.HaloVersionEnum HaloVersion;
            private ChunkCollection chunks = new ChunkCollection();
            #region Properties
          
          
            [Category("Chunk Info")]
            public int ChunkNumber
            {
                get { return chunknumber; }
                set { chunknumber = value; }
            }
            [Category("Chunk Info")]
            [TypeConverter(typeof(BaseObjectConverter))]
            public TagBlockLayout ChunkInfo
            {
                get { return chunks[chunknumber]; }
            }
            [Category("Chunk Info")]
            [TypeConverter(typeof(BaseObjectConverter))]
            public Pointer Translation
            {
                get { return translation; }
            }
            [Category("Chunk Info")]
            public int ChunkCount
            {
                get { return chunkCount; }
                set { chunkCount = value; }
            }

            #endregion 
            public TagBlock(string name, TagBlockLayout tagblocklayout, Map.HaloVersionEnum haloversion)
                : base(name)
            {
                if (tagblocklayout!=null) Layout = tagblocklayout;
                Layout.Parent = this;
                this.size=8;
                if (haloversion != Map.HaloVersionEnum.Halo2) this.size += 4;
                HaloVersion = haloversion;
            }

            

            public override void Read()
            {
                chunkCount = Maps.map[MapNumber].BR.ReadInt32();
                translation.Read();
                if (HaloVersion != Map.HaloVersionEnum.Halo2)
                    Maps.map[MapNumber].BR.ReadInt32();
            }
       
   
       

            public void ReadElements()
            {
                
                for (int x = 0; x < ChunkCount; x++)
                {
                    chunks.Add(new TagBlockLayout());
                  //  elements[x] = new TagBlockLayout();
                    chunks[x] = Layout;
                    chunks[x].Parent = this;
                }

                Maps.map[MapNumber].BR.BaseStream.Position = TagInterface.Offset+Translation.MetaTranslation ;
                for (int x = 0; x < ChunkCount; x++)

                    chunks[x].ReadFields();


            }

            





        }

        public class PaddingX : BaseObject
        {
        

            public PaddingX(int size, string name)
                : base(name)
            {
                this.size = size;

            }
            public override void Read()
            {
                Maps.map[MapNumber].BR.BaseStream.Position += this.Size;
            }
            public override void Write()
            {
                byte zero = 0;
                for (int x = 0; x < this.Size; x++)
                    Maps.map[MapNumber].BW.Write(zero);
            }
        }
        #endregion


  



        public class BaseObjectConverter : ExpandableObjectConverter
        {
            public override object ConvertTo(ITypeDescriptorContext context,
                                   System.Globalization.CultureInfo culture,
                                   object value, Type destType)
            {

                if (destType == typeof(string))
                {
                    if (value is Pointer emp)
                    {
                        return "Offsets";
                    }
                    else if (value is TagBlockLayout layout)
                    {
                        return layout.Parent.Name;
                    }
                    else if (value is TagBlock tag)
                    {
                        return tag.Name;
                    }
                }
                return "";
            }
        }
    }

    

}
