using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Xml;
using HaloMap.MetaContainers;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
namespace HaloMap
{

    #region Raw Data Chunks
    public enum RawDataType
    {
        bitm,
        bsp1,
        bsp2,
        bsp3,
        bsp4,
        DECR,
        jmad,
        ltmp,
        mode1,
        mode2,
        PRTM,
        snd1,
        snd2,
        weat,
        HaloCEVertices,
        HaloCEIndices
    }
    public class RawDataChunk
	{
		#region Fields
        public int shadernumber;
		public RawDataType rawDataType;
		public int offset;
		public int size;
		public int pointerMetaOffset;
		public MapTypes rawLocation;
		public MemoryStream MS;
		#endregion
		#region Constructor
		public RawDataChunk()
		{
		
		}
		#endregion

	}

	public class BitmapRawDataChunk:RawDataChunk
	{
		#region Fields
		public int inchunk;
		public int num;
		#endregion
		#region Constructor
		public BitmapRawDataChunk()
		{
		
		}
		#endregion
    }

    #endregion

    public enum RawDataContainerType
	{
        
        Bitmap,
        Header,
		Model,
		Animation,
		Sound,
        CoconutsModel,
		PRTM,
		DECR,
		BSP,
		LightMap,
        Empty,
		Weather,
		BSPMeta,
		UnicodeNamesIndex,
		UnicodeNames,
		StringsIndex,
		Strings1,
		Strings2,
		FileNamesIndex,
		FileNames,
		Crazy,
		MetaIndex,
        MetaData
    
        
	}

    //raw data containers

	public class BitmapRaw:RawDataContainer
	{
		#region Fields
		
		#endregion
		#region Constructor
		public BitmapRaw()
		{
			containerType=RawDataContainerType.Bitmap;
		}
		#endregion
		#region Methods
        public override void Read(int tagnumber, int mapnumber, bool dontreadraw)
		{
            switch (Maps.map[mapnumber].HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
                    ReadH2BitmapRaw(tagnumber, mapnumber, dontreadraw);
                    break;
                case Map.HaloVersionEnum.HaloCE:
                case Map.HaloVersionEnum.Halo1:
                    ReadCEBitmapRaw(tagnumber, mapnumber, dontreadraw);
                    break;
            }
        }
        public void ReadCEBitmapRaw(int tagnumber, int mapnumber, bool dontreadraw)
        {
            if (Maps.map[mapnumber].MetaInfo.external[tagnumber] == true)
            {

          
                Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + 84;
                int rawcount = Maps.map[mapnumber].BR.ReadInt32();
                Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + 100;
                int translation = Maps.map[mapnumber].BR.ReadInt32();
                for (int x = 0; x < rawcount; x++)
                {
                    BitmapRawDataChunk Raw = new BitmapRawDataChunk();
                    Raw.inchunk = x;
                    Raw.num = 0;
                    Raw.rawDataType = RawDataType.bitm;
                    Raw.pointerMetaOffset = (x*48)+ translation+24;
                    Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+Raw.pointerMetaOffset;
                    Raw.offset = Maps.map[mapnumber].BR.ReadInt32();
                    Raw.size = Maps.map[mapnumber].BR.ReadInt32();
                    Raw.rawLocation = MapTypes.Bitmaps;

                    try
                    {
                        if (dontreadraw == false)
                        {
                            //Maps.map[mapnumber].OpenMap(Raw.rawLocation);
                            Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                            Raw.MS = new MemoryStream(Raw.size);
                            Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                            //  Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                        }
                        this.rawChunks.Add(Raw);
                    }
                   
                        catch
                    {}
                }
                // m.CloseMap();
            }
            else
            {
                Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+96;
                int tempc=Maps.map[mapnumber].BR.ReadInt32();
                int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                for (int x=0;x<tempc;x++)
                {
                             BitmapRawDataChunk Raw = new BitmapRawDataChunk();
                Raw.inchunk = x;
                Raw.num = 0;
                Raw.pointerMetaOffset = tempr + (x * 48) + 24;
                Maps.map[mapnumber].BR.BaseStream.Position = Raw.pointerMetaOffset + Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                Raw.rawDataType = RawDataType.bitm;
                Raw.offset = Maps.map[mapnumber].BR.ReadInt32();
                Raw.size = Maps.map[mapnumber].BR.ReadInt32();
                Raw.rawLocation = MapTypes.Internal;

                if (dontreadraw == false)
                {
                    //Maps.map[mapnumber].OpenMap(Raw.rawLocation);
                    Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                    //  Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                }
                this.rawChunks.Add(Raw);
            }
            }

        }
        public void ReadH2BitmapRaw(int tagnumber, int mapnumber, bool dontreadraw)
        {
            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+68;
            int tempc = Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].SecondaryMagic;
			
			for (int x=0;x<tempc;x++)
			{
				for (int xx=0;xx<6;xx++)
				{
					
					BitmapRawDataChunk Raw=new BitmapRawDataChunk();
					Raw.inchunk=x;
					Raw.num=xx;
					Raw.pointerMetaOffset=tempr+(x*116)+28+(xx*4)-Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                    Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + Raw.pointerMetaOffset;
					Raw.rawDataType=RawDataType.bitm;
                    Raw.offset = Maps.map[mapnumber].BR.ReadInt32();
					if (Raw.offset==-1){break;}
                    Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + Raw.pointerMetaOffset + 24;
                    Raw.size = Maps.map[mapnumber].BR.ReadInt32();
             
					Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset,ref Raw.rawLocation);
					if (dontreadraw==false)
					{
						Maps.map[mapnumber].OpenMap(Raw.rawLocation);
						Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
						Raw.MS=new MemoryStream(Raw.size);
						Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size),0,Raw.size);
                        Maps.map[mapnumber].OpenMap(MapTypes.Internal);
					}
					this.rawChunks.Add(Raw);
					
				}
			}
			
		}
	#endregion
	}

	public class Model:RawDataContainer
    {
        #region Fields 

        public int IntermediatePointerOffset; // for halo 1
		#endregion
		#region Constructor
		public Model()
		{
			containerType=RawDataContainerType.Model;
		}
		#endregion
		#region Methods
        public override void Read(int tagnumber, int mapnumber, bool dontreadraw)
		{
            switch (Maps.map[mapnumber].HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
                    ReadH2ModelRaw(tagnumber, mapnumber, dontreadraw);
                    break;
                case Map.HaloVersionEnum.HaloCE:
                     ReadCEModelRaw(tagnumber, mapnumber, dontreadraw);
                             break;
                case Map.HaloVersionEnum.Halo1:
                    ReadH1ModelRaw(tagnumber, mapnumber, dontreadraw);
                    break;
            }
        }
         public void ReadH1ModelRaw(int tagnumber, int mapnumber, bool dontreadraw)
        {
            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + 208;
            int tempc = Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic;
            for (int y = 0; y < tempc; y++)
            {
                Maps.map[mapnumber].BR.BaseStream.Position = tempr+(y * 48)+36;
                int tempcc = Maps.map[mapnumber].BR.ReadInt32();
                int temprc = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic;
                for (int x = 0; x < tempcc; x++)
                {
                    RawDataChunk Raw = new RawDataChunk();



                    Maps.map[mapnumber].BR.BaseStream.Position = temprc + (x * 104) + 4;
                    Raw.shadernumber = Maps.map[mapnumber].BR.ReadInt16();
                    Raw.pointerMetaOffset = temprc + (x * 104) + 76 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];

                    Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + Raw.pointerMetaOffset-4;
                    Raw.rawDataType = RawDataType.HaloCEIndices;
                    Raw.size = (Maps.map[mapnumber].BR.ReadInt32()+2)*2;
                    Raw.offset = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic; ;
                    
              
                    
                    if (dontreadraw == false)
                    {
                        Maps.map[mapnumber].OpenMap(Raw.rawLocation);
                        Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                        Raw.MS = new MemoryStream(Raw.size);
                        Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                        Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                    }
                    this.rawChunks.Add(Raw);

                    Raw = new RawDataChunk();
                    Raw.rawDataType = RawDataType.HaloCEVertices;
                    Raw.pointerMetaOffset = temprc + (x * 104) + 100 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                    Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + Raw.pointerMetaOffset-12;
                    Raw.size = Maps.map[mapnumber].BR.ReadInt32();
                     Raw.size *=   32;
                    
                    Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + Raw.pointerMetaOffset;

                    this.IntermediatePointerOffset=Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic; ;
                   Maps.map[mapnumber].BR.BaseStream.Position=this.IntermediatePointerOffset+4;
                   Raw.offset = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic; 
                  

                    if (dontreadraw == false)
                    {
                        Maps.map[mapnumber].OpenMap(Raw.rawLocation);
                        Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                        Raw.MS = new MemoryStream(Raw.size);
                        Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                        Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                    }
                    this.rawChunks.Add(Raw);
                }
            }
        }
        public void ReadCEModelRaw(int tagnumber, int mapnumber, bool dontreadraw)
        {
            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + 208;
            int tempc = Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic;
            for (int y = 0; y < tempc; y++)
            {
                Maps.map[mapnumber].BR.BaseStream.Position = tempr+(y * 48)+36;
                int tempcc = Maps.map[mapnumber].BR.ReadInt32();
                int temprc = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic;
                for (int x = 0; x < tempcc; x++)
                {
                    RawDataChunk Raw = new RawDataChunk();



                    Maps.map[mapnumber].BR.BaseStream.Position = temprc + (x * 132) + 4;
                    Raw.shadernumber = Maps.map[mapnumber].BR.ReadInt16();
                    Raw.pointerMetaOffset = temprc + (x * 132) + 76 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];

                    Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + Raw.pointerMetaOffset-4;
                    Raw.rawDataType = RawDataType.HaloCEIndices;
                    Raw.size = (Maps.map[mapnumber].BR.ReadInt32()+2)*2;
                    Raw.offset = Maps.map[mapnumber].IndexHeader.ModelRawDataOffset + Maps.map[mapnumber].IndexHeader.ModelIndicesOffset + Maps.map[mapnumber].BR.ReadInt32();
              
                    
                    if (dontreadraw == false)
                    {
                        Maps.map[mapnumber].OpenMap(Raw.rawLocation);
                        Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                        Raw.MS = new MemoryStream(Raw.size);
                        Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                        Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                    }
                    this.rawChunks.Add(Raw);

                    Raw = new RawDataChunk();
                    Raw.rawDataType = RawDataType.HaloCEVertices;
                    Raw.pointerMetaOffset = temprc + (x * 132) + 100 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                    Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + Raw.pointerMetaOffset-12;
                    Raw.size = Maps.map[mapnumber].BR.ReadInt32();
                     Raw.size *=   68;
                    
                    Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + Raw.pointerMetaOffset;
                    
                    
                    Raw.offset = Maps.map[mapnumber].IndexHeader.ModelRawDataOffset  + Maps.map[mapnumber].BR.ReadInt32();
                  

                    if (dontreadraw == false)
                    {
                        Maps.map[mapnumber].OpenMap(Raw.rawLocation);
                        Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                        Raw.MS = new MemoryStream(Raw.size);
                        Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                        Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                    }
                    this.rawChunks.Add(Raw);
                }
            }
        }
        public void ReadH2ModelRaw(int tagnumber, int mapnumber, bool dontreadraw)
        {
            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+36;
            int tempc = Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].SecondaryMagic;
			for (int x=0;x<tempc;x++)
			{
				RawDataChunk Raw=new RawDataChunk();
                Raw.pointerMetaOffset = tempr + (x * 92) + 56 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+Raw.pointerMetaOffset;
				Raw.rawDataType=RawDataType.mode1;
				Raw.offset=Maps.map[mapnumber].BR.ReadInt32();
				Raw.size=Maps.map[mapnumber].BR.ReadInt32();
				Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset,ref Raw.rawLocation);
				if (dontreadraw==false)
				{
					Maps.map[mapnumber].OpenMap(Raw.rawLocation);
					Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
					Raw.MS=new MemoryStream(Raw.size);
					Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size),0,Raw.size);
					Maps.map[mapnumber].OpenMap(MapTypes.Internal);
				}
				this.rawChunks.Add(Raw);
			}

			
			Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+116;
			tempc=Maps.map[mapnumber].BR.ReadInt32();
            tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].SecondaryMagic;
			for (int x=0;x<tempc;x++)
			{
				RawDataChunk Raw=new RawDataChunk();
                Raw.pointerMetaOffset = tempr + (x * 88) + 52 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+Raw.pointerMetaOffset;
				Raw.rawDataType=RawDataType.mode2;
                Raw.offset = Maps.map[mapnumber].BR.ReadInt32();
                Raw.size = Maps.map[mapnumber].BR.ReadInt32();
				Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset,ref Raw.rawLocation);
				if (dontreadraw==false)
				{
					Maps.map[mapnumber].OpenMap(Raw.rawLocation);
					Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
					Raw.MS=new MemoryStream(Raw.size);
					Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size),0,Raw.size);						
					Maps.map[mapnumber].OpenMap(MapTypes.Internal);
				}
				this.rawChunks.Add(Raw);
			}
			
		}
		#endregion
	}
    
    public class Sound : RawDataContainer
    {
        #region Fields
        public int index;
        public int count;
        public ugh_.SoundPermutationChunk[] Permutations;
        
        #endregion
        #region Constructor
        public Sound()
        {
            containerType = RawDataContainerType.Sound;
        }
        #endregion
        #region Methods
        public override void Read(int tagnumber, int mapnumber, bool dontreadraw)
        {
            switch (Maps.map[mapnumber].HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
                    ReadH2SoundData(tagnumber, mapnumber, dontreadraw);
                    break;
            //    case Map.HaloVersionEnum.HaloCE:
                case Map.HaloVersionEnum.Halo1:
                    ReadH1SoundData(tagnumber, mapnumber, dontreadraw);
                    break;
            }
        }
        public void ReadH1SoundData(int tagnumber, int mapnumber, bool dontreadraw)
        {
            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + 152;
            int tempc = Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic;
            Maps.map[mapnumber].BR.BaseStream.Position = tempr + 60;
            tempc = Maps.map[mapnumber].BR.ReadInt32();
            tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic;
            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.pointerMetaOffset = tempr + (x * 124) + 64 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + Raw.pointerMetaOffset;
                Raw.rawDataType = RawDataType.snd1;
                 Raw.size = Maps.map[mapnumber].BR.ReadInt32();
                 Maps.map[mapnumber].BR.ReadInt32();
                 Raw.offset = Maps.map[mapnumber].BR.ReadInt32();
               
                if (dontreadraw == false)
                {
           
                    Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                
                }
                this.rawChunks.Add(Raw);
            }
		
        }
        public void ReadH2SoundData(int tagnumber, int mapnumber, bool dontreadraw)
         {
            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber] + 8;
            
            index = Maps.map[mapnumber].BR.ReadInt16();
            count = 1;// Maps.map[mapnumber].BR.ReadInt16();
            Permutations = new ugh_.SoundPermutationChunk[count];
            for (int x = 0; x < count; x++)
            {
                int currentindex = index + x;
                Permutations[x] = Maps.map[mapnumber].ugh.Permutations[currentindex];
                for (int xx = 0; xx < Maps.map[mapnumber].ugh.Permutations[currentindex].choicecount; xx++)
                {
                    int choiceindexx = Maps.map[mapnumber].ugh.Permutations[currentindex].choiceindex + xx;
                    Permutations[x].Choices.Add(Maps.map[mapnumber].ugh.Choices[choiceindexx]);
                    for (int xxx = 0; xxx < Maps.map[mapnumber].ugh.Choices[choiceindexx].soundcount; xxx++)
                    {
                        int soundindex = Maps.map[mapnumber].ugh.Choices[choiceindexx].soundindex + xxx;
                        Permutations[x].Choices[xx].SoundChunks1.Add(Maps.map[mapnumber].ugh.SoundChunks1[soundindex]);

                        RawDataChunk Raw = new RawDataChunk();
                        
                        
                        Raw.rawDataType = RawDataType.snd1;
                        Raw.offset = Maps.map[mapnumber].ugh.SoundChunks1[soundindex].offset;
                        Raw.size = Maps.map[mapnumber].ugh.SoundChunks1[soundindex].size&0x3FFFFFFF;
                        Raw.rawLocation = Maps.map[mapnumber].ugh.SoundChunks1[soundindex].rawLocation;
                        if (dontreadraw == false)
                        {
                            Maps.map[mapnumber].OpenMap(Raw.rawLocation);
                            Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                            Raw.MS = new MemoryStream(Raw.size);
                            Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                        }
                        this.rawChunks.Add(Raw);
                    }
                }

                
            }

        }
        #endregion
    }

	public class Weather:RawDataContainer
	{
		#region Fields
		#endregion
		#region Constructor
		public Weather()
		{
			containerType=RawDataContainerType.Weather;
		}
		#endregion
		#region Methods
        public override void Read(int tagnumber, int mapnumber, bool dontreadraw)
		{

            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
			int tempc=Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].SecondaryMagic;
			for (int x=0;x<tempc;x++)
			{
				RawDataChunk Raw=new RawDataChunk();
                Raw.pointerMetaOffset = tempr + (x * 140) + 64 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
				Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+Raw.pointerMetaOffset;
				Raw.rawDataType=RawDataType.weat;
				Raw.offset=Maps.map[mapnumber].BR.ReadInt32();
				Raw.size=Maps.map[mapnumber].BR.ReadInt32();
				Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset,ref Raw.rawLocation);
				if (dontreadraw==false)
				{
					Maps.map[mapnumber].OpenMap(Raw.rawLocation);
					Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
					Raw.MS=new MemoryStream(Raw.size);
					Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size),0,Raw.size);
					Maps.map[mapnumber].OpenMap(MapTypes.Internal);
				}
				this.rawChunks.Add(Raw);
			}
		
			
		}
		#endregion
	}

	public class Animation:RawDataContainer
	{
		#region Fields
		#endregion
		#region Constructor
		public Animation()
		{
			containerType=RawDataContainerType.Animation;
		}
		#endregion
		#region Methods
        public override void Read(int tagnumber, int mapnumber, bool dontreadraw)
		{
			
			Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+172;
			int tempc=Maps.map[mapnumber].BR.ReadInt32();
			int tempr=Maps.map[mapnumber].BR.ReadInt32()-Maps.map[mapnumber].SecondaryMagic;
           	for (int x=0;x<tempc;x++)
			{
				RawDataChunk Raw=new RawDataChunk();
				Raw.rawDataType=RawDataType.jmad;
                Raw.pointerMetaOffset = tempr + (x * 20) + 8 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
				Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+Raw.pointerMetaOffset-4;
				Raw.size=Maps.map[mapnumber].BR.ReadInt32();
				Raw.offset=Maps.map[mapnumber].BR.ReadInt32();
				
				Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset,ref Raw.rawLocation);
				if (dontreadraw==false)
				{
					Maps.map[mapnumber].OpenMap(Raw.rawLocation);
					Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
					Raw.MS=new MemoryStream(Raw.size);
					Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size),0,Raw.size);
					Maps.map[mapnumber].OpenMap(MapTypes.Internal);
				}
				this.rawChunks.Add(Raw);
			}
			
		}
		#endregion
	}

	public class PRTM:RawDataContainer
	{
		#region Fields
		#endregion
		#region Constructor
		public PRTM()
		{
			containerType=RawDataContainerType.PRTM;
		}

		#endregion
		#region Methods
		public override void Read(int tagnumber,int mapnumber,bool dontreadraw)
		{
			
			RawDataChunk Raw=new RawDataChunk();
			Raw.rawDataType=RawDataType.PRTM;
			Raw.pointerMetaOffset=160;
			Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+160;
			Raw.offset=Maps.map[mapnumber].BR.ReadInt32();
			Raw.size=Maps.map[mapnumber].BR.ReadInt32();
			Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset,ref Raw.rawLocation);
			if (dontreadraw==false)
			{
				Maps.map[mapnumber].OpenMap(Raw.rawLocation);
				Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
				Raw.MS=new MemoryStream(Raw.size);
				Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size),0,Raw.size);
				Maps.map[mapnumber].OpenMap(MapTypes.Internal);
			}
			this.rawChunks.Add(Raw);
		}
		#endregion
	}

    public class UghRawContainer : RawDataContainer
    {
        #region Fields
        #endregion
        #region Constructor
        public UghRawContainer()
        {
            containerType = RawDataContainerType.CoconutsModel;
        }

        #endregion
        #region Methods
        public override void Read(int tagnumber, int mapnumber, bool dontreadraw)
        {

            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+80;
            int tempc = Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].SecondaryMagic;
            for (int x = 0; x < tempc; x++)
            {
                RawDataChunk Raw = new RawDataChunk();
                Raw.rawDataType = RawDataType.snd2;
                Raw.pointerMetaOffset = tempr + (x * 44) + 8 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+Raw.pointerMetaOffset ;
             
                Raw.offset = Maps.map[mapnumber].BR.ReadInt32();
                Raw.size = Maps.map[mapnumber].BR.ReadInt32();
                Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                if (dontreadraw == false)
                {
                    Maps.map[mapnumber].OpenMap(Raw.rawLocation);
                    Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                    Raw.MS = new MemoryStream(Raw.size);
                    Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                    Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                }
                this.rawChunks.Add(Raw);
            }

        }
        #endregion
    }



	public class DECR:RawDataContainer
	{
		#region Fields
		#endregion
		#region Constructor
		public DECR()
		{
			containerType=RawDataContainerType.DECR;
		}

		#endregion
		#region Methods
		public override void Read(int tagnumber,int mapnumber,bool dontreadraw)
		{
			
			RawDataChunk Raw=new RawDataChunk();
			Raw.rawDataType=RawDataType.DECR;
			Raw.pointerMetaOffset=56;
			Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+56;
			Raw.offset=Maps.map[mapnumber].BR.ReadInt32();
			Raw.size=Maps.map[mapnumber].BR.ReadInt32();
			Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset,ref Raw.rawLocation);
			if (dontreadraw==false)
			{
				Maps.map[mapnumber].OpenMap(Raw.rawLocation);
				Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
				Raw.MS=new MemoryStream(Raw.size);
				Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size),0,Raw.size);
				Maps.map[mapnumber].OpenMap(MapTypes.Internal);
			}
			this.rawChunks.Add(Raw);
		}
		#endregion

	}

	public class BSPRaw:RawDataContainer
	{
		#region Fields
		#endregion
		#region Constructor
		public BSPRaw()
		{
			containerType=RawDataContainerType.BSP;
		}

		#endregion
		#region Methods
		public override void Read(int tagnumber,int mapnumber,bool dontreadraw)
		{
            switch (Maps.map[mapnumber].HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
                    H2Read(tagnumber, mapnumber, dontreadraw);
                    break;
                case Map.HaloVersionEnum.HaloCE:
                    break;
            }
        }
        public void H2Read(int tagnumber, int mapnumber, bool dontreadraw)
        {
            int w = Maps.map[mapnumber].BSP.FindBSPNumberByBSPIdent(Maps.map[mapnumber].MetaInfo.Ident[tagnumber]);
			
			//bsp model 1
			Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+172;
			int tempc=Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].BSP.sbsp[w].magic;
			for (int x=0;x<tempc;x++)
			{
				RawDataChunk Raw=new RawDataChunk();
				Raw.rawDataType=RawDataType.bsp1;
                Raw.pointerMetaOffset = tempr + (x * 176) + 40 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
				Maps.map[mapnumber].BR.BaseStream.Position=tempr+(x*176)+40;
				Raw.offset=Maps.map[mapnumber].BR.ReadInt32();
			
				Raw.size=Maps.map[mapnumber].BR.ReadInt32();
                if (Raw.offset == -1) { this.rawChunks.Add(Raw); continue; }
                Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
				if (dontreadraw==false)
				{
                    Maps.map[mapnumber].OpenMap(Raw.rawLocation);
					Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
					Raw.MS=new MemoryStream(Raw.size);
                    Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                    Maps.map[mapnumber].OpenMap(MapTypes.Internal);
				}
				this.rawChunks.Add(Raw);
			}
			

			//bsp model 2
			Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+328;
			tempc=Maps.map[mapnumber].BR.ReadInt32();
			tempr=Maps.map[mapnumber].BR.ReadInt32()-Maps.map[mapnumber].BSP.sbsp[w].magic;
			for (int x=0;x<tempc;x++)
			{
				RawDataChunk Raw=new RawDataChunk();
				Raw.rawDataType=RawDataType.bsp2;
                Raw.pointerMetaOffset = tempr + (x * 200) + 40 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
				Maps.map[mapnumber].BR.BaseStream.Position=tempr+(x*200)+40;
				Raw.offset=Maps.map[mapnumber].BR.ReadInt32();
				
				Raw.size=Maps.map[mapnumber].BR.ReadInt32();
                if (Raw.offset == -1) { this.rawChunks.Add(Raw); continue; }
				Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset,ref Raw.rawLocation);
				if (dontreadraw==false)
				{
					Maps.map[mapnumber].OpenMap(Raw.rawLocation);
					Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
					Raw.MS=new MemoryStream(Raw.size);
					Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size),0,Raw.size);
					Maps.map[mapnumber].OpenMap(MapTypes.Internal);
				}
				this.rawChunks.Add(Raw);
			}

			//light map
            if (Maps.map[mapnumber].BSP.sbsp[w].lightmapident != -1)
            {
                Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+8;
                tempr = Maps.map[mapnumber].BR.ReadInt32();
                if (tempr == 0) { goto skiplightmap; }
               tempr -= Maps.map[mapnumber].BSP.sbsp[w].magic;
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + 128;
                tempc = Maps.map[mapnumber].BR.ReadInt32();
                tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].BSP.sbsp[w].magic;
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + 64;
                tempc = Maps.map[mapnumber].BR.ReadInt32();
                tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].BSP.sbsp[w].magic ;
                for (int x = 0; x < tempc; x++)
                {
                    RawDataChunk Raw = new RawDataChunk();
                    Raw.rawDataType = RawDataType.ltmp;
                    Raw.pointerMetaOffset = tempr + (x * 56) + 12 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                    Maps.map[mapnumber].BR.BaseStream.Position =  tempr + (x * 56) + 12;
                    Raw.offset = Maps.map[mapnumber].BR.ReadInt32();
                  
                    Raw.size = Maps.map[mapnumber].BR.ReadInt32();
                    if (Raw.offset == -1) { this.rawChunks.Add(Raw); continue; }
                    Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                    if (dontreadraw == false)
                    {
                        Maps.map[mapnumber].OpenMap(Raw.rawLocation);
                        Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                        Raw.MS = new MemoryStream(Raw.size);
                        Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                        Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                    }
                    this.rawChunks.Add(Raw);
                }

            }
            skiplightmap:
			//bsp model 3
			Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+580;
			tempc=Maps.map[mapnumber].BR.ReadInt32();
			tempr=Maps.map[mapnumber].BR.ReadInt32()-Maps.map[mapnumber].BSP.sbsp[w].magic;
			//if (tempc==0){return;}
			Maps.map[mapnumber].BR.BaseStream.Position=tempr+16;
			tempc=Maps.map[mapnumber].BR.ReadInt32();
			tempr=Maps.map[mapnumber].BR.ReadInt32()-Maps.map[mapnumber].BSP.sbsp[w].magic;
			for (int x=0;x<tempc;x++)
			{
				RawDataChunk Raw=new RawDataChunk();
				Raw.rawDataType=RawDataType.bsp3;
                Raw.pointerMetaOffset = tempr + (x * 44) - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
				Maps.map[mapnumber].BR.BaseStream.Position=tempr+(x*44);
				Raw.offset=Maps.map[mapnumber].BR.ReadInt32();
				Raw.size = Maps.map[mapnumber].BR.ReadInt32();
                if (Raw.offset == -1) { this.rawChunks.Add(Raw); continue; }
				Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset,ref Raw.rawLocation);
				if (dontreadraw==false)
				{
					Maps.map[mapnumber].OpenMap(Raw.rawLocation);
					Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
					Raw.MS=new MemoryStream(Raw.size);
					Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size),0,Raw.size);
                    Maps.map[mapnumber].OpenMap(MapTypes.Internal);
				}
				this.rawChunks.Add(Raw);
			}

			//bsp model 4
			Maps.map[mapnumber].BR.BaseStream.Position=Maps.map[mapnumber].MetaInfo.Offset[tagnumber]+548;
			tempc=Maps.map[mapnumber].BR.ReadInt32();
			tempr=Maps.map[mapnumber].BR.ReadInt32()-Maps.map[mapnumber].BSP.sbsp[w].magic;
		
			for (int x=0;x<tempc;x++)
			{
				RawDataChunk Raw=new RawDataChunk();
				Raw.rawDataType=RawDataType.bsp4;
                Raw.pointerMetaOffset = tempr + (x * 172) + 16 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
				Maps.map[mapnumber].BR.BaseStream.Position=tempr+(x*172)+16;
				Raw.offset=Maps.map[mapnumber].BR.ReadInt32();
				if (Raw.offset==-1) {continue;}
				Raw.size=Maps.map[mapnumber].BR.ReadInt32();
				Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset,ref Raw.rawLocation);
				if (dontreadraw==false)
				{
					Maps.map[mapnumber].OpenMap(Raw.rawLocation);
					Maps.map[mapnumber].BR.BaseStream.Position=Raw.offset;
					Raw.MS=new MemoryStream(Raw.size);
					Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size),0,Raw.size);
					Maps.map[mapnumber].OpenMap(MapTypes.Internal);
				}
				this.rawChunks.Add(Raw);
			}
		}
		#endregion
		
	}

    public class LightmapRaw : RawDataContainer
    {
        #region Fields
        #endregion
        #region Constructor
        public LightmapRaw()
        {
            containerType = RawDataContainerType.LightMap;
        }

        #endregion
        #region Methods
        public override void Read(int tagnumber, int mapnumber, bool dontreadraw)
        {
            switch (Maps.map[mapnumber].HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
                    H2Read(tagnumber, mapnumber, dontreadraw);
                    break;
                case Map.HaloVersionEnum.HaloCE:
                    break;
            }
        }
        public void H2Read(int tagnumber, int mapnumber, bool dontreadraw)
        {
            int w = Maps.map[mapnumber].BSP.FindBSPNumberByLightMapIdent(Maps.map[mapnumber].MetaInfo.Ident[tagnumber]);

            

            //light map
            if (Maps.map[mapnumber].BSP.sbsp[w].lightmapident != -1)
            {
              
                int tempr = Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + 128;
                int tempc = Maps.map[mapnumber].BR.ReadInt32();
                tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].BSP.sbsp[w].magic;
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + 64;
                tempc = Maps.map[mapnumber].BR.ReadInt32();
                tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].BSP.sbsp[w].magic;
                for (int x = 0; x < tempc; x++)
                {
                    RawDataChunk Raw = new RawDataChunk();
                    Raw.rawDataType = RawDataType.ltmp;
                    Raw.pointerMetaOffset = tempr + (x * 56) + 12 - Maps.map[mapnumber].MetaInfo.Offset[tagnumber];
                    Maps.map[mapnumber].BR.BaseStream.Position = tempr + (x * 56) + 12;
                    Raw.offset = Maps.map[mapnumber].BR.ReadInt32();

                    Raw.size = Maps.map[mapnumber].BR.ReadInt32();
                    if (Raw.offset == -1) { this.rawChunks.Add(Raw); continue; }
                    Maps.map[mapnumber].Functions.ParsePointer(ref Raw.offset, ref Raw.rawLocation);
                    if (dontreadraw == false)
                    {
                        Maps.map[mapnumber].OpenMap(Raw.rawLocation);
                        Maps.map[mapnumber].BR.BaseStream.Position = Raw.offset;
                        Raw.MS = new MemoryStream(Raw.size);
                        Raw.MS.Write(Maps.map[mapnumber].BR.ReadBytes(Raw.size), 0, Raw.size);
                        Maps.map[mapnumber].OpenMap(MapTypes.Internal);
                    }
                    this.rawChunks.Add(Raw);
                }

            }
    
        }
        #endregion

    }

	public class RawDataContainer
	{
		#region Fields
		public RawDataContainerType containerType;
        public List<RawDataChunk> rawChunks = new List<RawDataChunk>();
		#endregion
		#region Constructor
		public RawDataContainer()
		{
			
		}
		#endregion
		#region Methods
        public virtual void Read(int tagnumber, int mapnumber, bool dontreadraw)
        {
        }
		public void SaveRawToFile(string outputFilePath,TagMeta meta)
		{
			int x=outputFilePath.LastIndexOf('.');
			string temp=outputFilePath.Substring(0,x+1)+meta.type+"raw";

            if (meta.type == "snd!"&&Maps.map[meta.mapnumber].HaloVersion== Map.HaloVersionEnum.Halo2)
            {
               
                Stream s = File.Open(temp+"layout",FileMode.Create);
                BinaryFormatter b = new BinaryFormatter();
                Sound temps=(Sound) meta.raw;
                b.Serialize(s, temps.Permutations);
                s.Close();

            }


			XmlTextWriter xtw=new XmlTextWriter(temp+"xml",System.Text.Encoding.Default);
			xtw.Formatting=Formatting.Indented;
			xtw.WriteStartElement("RawData");
			xtw.WriteAttributeString("TagType",meta.type);
			xtw.WriteAttributeString("TagName",meta.name);
			xtw.WriteAttributeString("RawType",meta.raw.containerType.ToString());
			xtw.WriteAttributeString("RawChunkCount",this.rawChunks.Count.ToString());
			xtw.WriteAttributeString("Date",DateTime.Today.ToShortDateString());
			xtw.WriteAttributeString("Time",DateTime.Now.ToShortTimeString());
			xtw.WriteAttributeString("EntityVersion","0.1");

			FileStream FS=new FileStream(temp,FileMode.Create);
			BinaryWriter BW=new BinaryWriter(FS);
			int loc=0;
			for (x=0;x<this.rawChunks.Count;x++)
			{

				RawDataChunk r=(RawDataChunk)this.rawChunks[x];
                if (r.offset==-1){continue;}
				xtw.WriteStartElement("RawChunk");
				xtw.WriteAttributeString("RawDataType",r.rawDataType.ToString());
				xtw.WriteAttributeString("PointerMetaOffset",r.pointerMetaOffset.ToString());
				xtw.WriteAttributeString("ChunkSize",r.size.ToString());
				xtw.WriteAttributeString("RawDataOffset",loc.ToString());
				xtw.WriteAttributeString("PointsToOffset",r.offset.ToString());
				xtw.WriteAttributeString("RawLocation",r.rawLocation.ToString());
				xtw.WriteEndElement();
				BW.BaseStream.Write(r.MS.ToArray(),0,r.size);
				loc+=r.size;
			}
			BW.Close();
			FS.Close();
			xtw.WriteEndElement();
			xtw.Close();
		}
		public RawDataContainer LoadRawFromFile(string inputFilePath,TagMeta meta)
		{
			RawDataContainer raw=new RawDataContainer();
			int x=inputFilePath.LastIndexOf('.');
			string temp=inputFilePath.Substring(0,x+1)+meta.type+"raw";





			XmlTextReader xtr=new XmlTextReader(temp+"xml");
			xtr.WhitespaceHandling=WhitespaceHandling.None;

            

			FileStream FS=new FileStream(temp,FileMode.Open);
			BinaryReader BR=new BinaryReader(FS);
			

			while (xtr.Read())
			{
				switch (xtr.NodeType)
				{
					case XmlNodeType.Element:
						if (xtr.Name=="RawData")
						{
							string oi=xtr.GetAttribute("RawType");
                            
							switch (oi)
							{
								case "Model":
									raw=new Model();
									break;
								case "Bitmap":
									raw=new BitmapRaw();
									break;
								case "Animation":
									raw=new Animation();
									break;
								case "DECR":
									raw=new DECR();
									break;
								case "PRTM":
									raw=new PRTM();
									break;
                                case "Weather":
                                    raw = new Weather();
                                    break;
                                case "Sound":
                                    raw = new Sound();
                                    break;
                                case "BSP":
                                    raw = new BSPRaw();
                                    break;
							}
                           
						}
						else if (xtr.Name=="RawChunk")
						{
							RawDataChunk r=new RawDataChunk();
                            string temps = xtr.GetAttribute("RawDataType");
							switch (temps)
							{
								case "bitm":
									r.rawDataType=RawDataType.bitm;
									break;
								case "bsp1":
									r.rawDataType=RawDataType.bsp1;
									break;
								case "bsp2":
									r.rawDataType=RawDataType.bsp2;
									break;
								case "bsp3":
									r.rawDataType=RawDataType.bsp3;
									break;
								case "bsp4":
									r.rawDataType=RawDataType.bsp4;
									break;
								case "DECR":
									r.rawDataType=RawDataType.DECR;
									break;
								case "jmad":
									r.rawDataType=RawDataType.jmad;
									break;
								case "ltmp":
									r.rawDataType=RawDataType.ltmp;
									break;
								case "mode1":
									r.rawDataType=RawDataType.mode1;
									break;
								case "mode2":
									r.rawDataType=RawDataType.mode2;
									break;
								case "PRTM":
									r.rawDataType=RawDataType.PRTM;
									break;
								case "snd1":
									r.rawDataType=RawDataType.snd1;
									break;
								case "snd2":
									r.rawDataType=RawDataType.snd2;
									break;


							}

                            r.offset = Convert.ToInt32(xtr.GetAttribute("PointsToOffset"));
							r.pointerMetaOffset=Convert.ToInt32(xtr.GetAttribute("PointerMetaOffset"));
							r.size=Convert.ToInt32(xtr.GetAttribute("ChunkSize"));
							int rawdataspot=Convert.ToInt32(xtr.GetAttribute("RawDataOffset"));
							BR.BaseStream.Position=rawdataspot;
							r.MS=new MemoryStream(r.size);
							r.MS.Write(BR.ReadBytes(r.size),0,r.size);

							raw.rawChunks.Add(r);
                           
						}
						break;
				}

            
			}
           
			BR.Close();
			FS.Close();
			xtr.Close();

            if (meta.type == "snd!")
            {
             
                Stream s = File.Open(temp + "layout", FileMode.Open);
                BinaryFormatter b = new BinaryFormatter();
                Sound temps = (Sound)raw;
                temps.Permutations = (ugh_.SoundPermutationChunk[])b.Deserialize(s);
                meta.raw = temps;
                s.Close();


            }
		
			return raw;
		}
		#endregion		
	}


}
