using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace HaloMap
{
    public class UnicodeTableReader
    {
        #region Fields
        private string filePath;
        private int matgOffset;
        
        public UnicodeTable[] ut;
        #endregion
        #region Constructor
        public UnicodeTableReader(string File, int MatgOffset)
        {
            this.filePath = File;
            this.matgOffset = MatgOffset;
            this.GetOffsetsAndSizes(false);
        }
        #endregion
        #region Methods/Functions
        private void GetOffsetsAndSizes(bool scan)
        {
            ut = new UnicodeTable[9];
            FileStream fs = new FileStream(filePath, FileMode.Open, FileAccess.Read);
            BinaryReader br = new BinaryReader(fs);
            for (int counter = 0; counter < 9; counter++)
            {
                br.BaseStream.Position = matgOffset + 400 + 28 * counter;
                ut[counter] = new UnicodeTable(false, counter, br.ReadInt32(), br.ReadInt32(), (int)br.BaseStream.Position - matgOffset, br.ReadInt32(), (int)br.BaseStream.Position - matgOffset, br.ReadInt32(), filePath);
            }
            br.Close();
            fs.Close();
        }
        #endregion
        #region Enums
        public enum Language
        {
            English,
            Japanese,
            German,
            French,
            Spanish,
            Italian,
            Korean,
            Chinese,
            Portuguese
        }
        #endregion
        #region Classes
        public class UnicodeTable
        {
            #region Fields
            private string filePath;
            public bool scanned;
            public Language language;
            public int count;
            public int tableSize;
            public int indexOffset;
            public int indexPointerOffset;
            public int tableOffset;
            public int tablePointerOffset;
            public Sid[] SIDs;
            public UnicodeString[] US;
            #endregion
            #region Constructor
            public UnicodeTable(bool iscanned, int iLanguage, int icount, int itablesize, int iindexPointerOffset, int iindexOffset, int itablePointerOffset, int itableOffset, string ifilePath)
            {
                this.scanned = iscanned;
                this.language = (Language)iLanguage;
                this.count = icount;
                this.tableSize = itablesize;
                this.indexPointerOffset = iindexPointerOffset;
                this.indexOffset = iindexOffset;
                this.tablePointerOffset = itablePointerOffset;
                this.tableOffset = itableOffset;
                this.filePath = ifilePath;
            }
            #endregion
            #region Methods
            public void Read()
            {
                SIDs = new Sid[count];
                US = new UnicodeString[count];
                ReadIndex();
                ReadTable();
            }
            private void ReadIndex()
            {
                FileStream fs = new FileStream(filePath, FileMode.Open, FileAccess.Read);
                BinaryReader br = new BinaryReader(fs);
                br.BaseStream.Position = this.indexOffset;
                for (int counter = 0; counter < count; counter++)
                {
                    SIDs[counter] = new Sid(br.ReadInt16(), br.ReadByte(), br.ReadByte());
                    US[counter] = new UnicodeString(br.ReadInt32());
                }
                br.Close();
                fs.Close();
                for (int counter = 0; counter < count - 1; counter++)
                {
                    US[counter].size = US[counter + 1].offset - US[counter].offset - 1;
                }
                US[count - 1].size = tableSize - US[count - 1].offset;
            }
            private void ReadTable()
            {
                FileStream fs = new FileStream(filePath, FileMode.Open, FileAccess.Read);
                BinaryReader br = new BinaryReader(fs);
                Encoding decode = Encoding.UTF8;
                for (int counter = 0; counter < count; counter++)
                {
                    br.BaseStream.Position = this.tableOffset + US[counter].offset;
                    byte[] temp = br.ReadBytes(US[counter].size);
                    US[counter].uString = decode.GetString(temp);
                }
            }
            #endregion

        }
        public class Sid
        {
            #region Fields
            public int id;
            public byte size;
            #endregion
            #region Constructor
            public Sid(int iid, byte uselessByte, byte isize)
            {
                this.size = isize;
                this.id = iid;
            }
            #endregion
        }
        public class UnicodeString
        {
            #region Fields
            public int offset;
            public int size;
            public string uString;
            #endregion
            #region Constructor
            public UnicodeString(int ioffset)
            {
                this.offset = ioffset;
            }
            #endregion
        }
        #endregion
    }
}
