﻿using System.Collections.Generic;
using System.IO;

namespace ScriptTools
{
    public static class ScriptDatabase
    {
        public static List<string> Strings { get; } = new List<string>();
        public static List<int> Types { get; } = new List<int>();
        public static List<int> Identities { get; } = new List<int>();
        public static List<int[]> Values { get; } = new List<int[]>();

        public static void AddValue(string name, int type, int identity, int[] value)
        {
            if (Types.Count == 0)
            {
                Types.Add(type);
                Strings.Add(name);
                Identities.Add(identity);
                Values.Add(value);
                return;
            }

            int start = Types.IndexOf(type);
            int end = Types.LastIndexOf(type);

            if (start == -1 || end == -1)
            {
                for (int i = 0; i < Types.Count; i++)
                {
                    if (Types[i] < type)
                    {
                        Types.Insert(i, type);
                        Values.Insert(i, value);
                        Identities.Add(identity);
                        Strings.Insert(i, name);
                        return;
                    }
                }

                Types.Add(type);
                Values.Add(value);
                Identities.Add(identity);
                Strings.Add(name);
                return;
            }

            if (name == null)
                name = "!";

            for (int i = start; i < end; i++)
            {
                char[] strchar = Strings[i].ToCharArray();
                char[] strname = name.ToCharArray();
                int length = strchar.Length > strname.Length ? strname.Length : strchar.Length;

                for (int j = 0; j < length; j++)
                {
                    if (strname[j] > strchar[j])
                    {
                        Strings.Insert(i, name);
                        Types.Insert(i, type);
                        Identities.Insert(i, identity);
                        Values.Insert(i, value);
                        return;
                    }

                    if (strchar[j] > strname[j])
                        break;
                }
            }

            Strings.Insert(end, name);
            Types.Insert(end, type);
            Identities.Insert(end, identity);
            Values.Insert(end, value);
        }

        public static void Save(FileInfo file)
        {
            using (var writer = file.Exists ? new StreamWriter(file.OpenWrite()) : file.CreateText())
            {
                for (int i = 0; i < Strings.Count; i++)
                {
                    string value = "";

                    for (int j = 0; j < 4; j++)
                        value += Values[i][j].ToString("X") + " ";

                    writer.WriteLine(Types[i].ToString() + "\t" + Strings[i] + "\t" + Identities[i].ToString() + "\t" + value);
                }
            }
        }
    }
}