﻿using System.Collections.Generic;
using System.IO;

namespace ScriptTools
{
    public class ScriptDatabaseInfo
    {
        public static List<ScriptDatabaseInfo> Entries { get; } = new List<ScriptDatabaseInfo>();

        public int Identity;
        public int ReturnType;
        public string Name;
        public List<int> Parameters;

        public ScriptDatabaseInfo()
        {
            Parameters = new List<int>();
        }

        public static void Add(ScriptDatabaseInfo info)
        {
            for (int i = 0; i < Entries.Count; i++)
            {
                if (Entries[i].Identity == info.Identity &&
                    Entries[i].ReturnType == info.ReturnType &&
                    Entries[i].Parameters.Count == info.Parameters.Count)
                {
                    var good = true;

                    for (int j = 0; j < info.Parameters.Count; j++)
                    {
                        if (Entries[i].Parameters[j] != info.Parameters[j])
                        {
                            good = false;
                            break;
                        }
                    }

                    if (good)
                        return;
                }
            }

            Entries.Add(info);
        }

        public static void Save(FileInfo file)
        {
            Entries.Sort(new ScriptDatabaseInfoComparer());

            using (var writer = file.Exists ? new StreamWriter(file.OpenWrite()) : file.CreateText())
            {
                for (int i = 0; i < Entries.Count; i++)
                {
                    var currentInfo = Entries[i];

                    var paramString = "";

                    for (int j = 0; j < currentInfo.Parameters.Count; j++)
                    {
                        var currentParam = currentInfo.Parameters[j];

                        if (currentParam < 0 || currentParam >= ScriptValueTypes.Names.Length)
                            continue;

                        paramString += "\t" + ScriptValueTypes.Names[Entries[i].Parameters[j]];
                    }

                    writer.WriteLine(Entries[i].Identity.ToString() + "\t" +
                        ScriptValueTypes.Names[Entries[i].ReturnType] + "\t" +
                        Entries[i].Name + "\t" + paramString);
                }

                writer.WriteLine("");
            }
        }
    }
}